;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                        3d Table Analysis Functions                     *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;

;;; Copyright (c) 1990 - 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.

;;; For more information write to clasp-support@cs.umass.edu

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;; 06-16-95 Created (schmill)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

(in-package :clasp)

;;; ***************************************************************************
;;; table constructors

(defun make-nway-contingency-table (&rest gvs)
  "Makes an array-representation of an n-way contingency table"
  (let* ((gns (mapcar #'remove-duplicates gvs))
         (ct (make-array (mapcar #'length gns) :initial-element 0)))
    (labels ((indexize (gvvs gvns)
                       (if (null gvvs) nil
		           (cons (position (car gvvs) (car gvns)) (indexize (cdr gvvs) (cdr gvns))))))
      (labels ((crunch (gvas)
                       (if (null (car gvas)) NIL
                           (progn
                             (incf (apply #'aref ct (indexize (mapcar #'car gvas) gns)))
                             (crunch (mapcar #'cdr gvas))))))
        (crunch gvs)
        (values ct gns)))))

(defun make-3d-contingency-table (v1 v2 v3)
  "Builds an array-represented 3d contingency table"
  (let* ((v1s (extract-unique-values v1))
         (v2s (extract-unique-values v2))
         (v3s (extract-unique-values v3))
         (table (make-array (list (length v1s) (length v2s) (length v3s)) :initial-element 0)))
    (loop for n1 in v1
          for n2 in v2
	  for n3 in v3 do
	  (incf (aref table (position n1 v1s) (position n2 v2s) (position n3 v3s))))
    (values table v1s v2s v3s)))

(defun make-3d-contingency-table-set (v1 v2 v3)
  "Builds arrays of a 3d contingency table and the margin tables"
  (let* ((v1s (extract-unique-values v1))
         (v2s (extract-unique-values v2))
         (v3s (extract-unique-values v3))
         (table (make-array (list (length v1s) (length v2s) (length v3s)) :initial-element 0))
         (m1d (make-array (list (length v2s) (length v3s)) :initial-element 0))
         (m2d (make-array (list (length v1s) (length v3s)) :initial-element 0))
         (m3d (make-array (list (length v1s) (length v2s)) :initial-element 0)))
    (loop for n1 in v1
          for n2 in v2
	  for n3 in v3 do
	  (incf (aref table (position n1 v1s) (position n2 v2s) (position n3 v3s)))
	  (incf (aref m1d   (position n2 v2s) (position n3 v3s)))
	  (incf (aref m2d   (position n1 v1s) (position n3 v3s)))
	  (incf (aref m3d   (position n1 v1s) (position n2 v2s))))
    (values table m1d m2d m3d v1s v2s v3s)))

;;; ---------------------------------------------------------------------------
;;; analyses

(defun 3d-complete-independence (ct m23 m13 m12 L)
  "Tests for complete independence of the factors in a 3way contingency table. Returns the G statistic
and its chi-square significance.
params = (3way-ctable 2way-23-margins 2way-13-margins 2way-12-margins data-length)"
  (let* ((ts (* L L))
	 (dim1 (1- (array-dimension ct 0)))
	 (dim2 (1- (array-dimension ct 1)))
	 (dim3 (1- (array-dimension ct 2)))
	 (DFci (- (* (1+ dim1) (1+ dim2) (1+ dim3)) (1+ dim1) (1+ dim2) (1+ dim3) -2))
	 #+NO (etable (make-array (list (1+ dim1) (1+ dim2) (1+ dim3))))
	 Gci)
      ;; G statistic
      (setq Gci (* 2
                 (loop for i1 from 0 to dim1 sum
                       (loop for i2 from 0 to dim2 sum
	                     (loop for i3 from 0 to dim3 sum
			           (* (aref ct i1 i2 i3)
			              (log (/ (aref ct i1 i2 i3)
					      ;; multiply expected frequency of i1, i2, i3, divide by ts
					      (/ (* (loop for i2m from 0 to dim2 sum
							  (aref m12 i1 i2m))
						    (loop for i3m from 0 to dim3 sum
							  (aref m23 i2 i3m))
						    (loop for i1m from 0 to dim1 sum
							  (aref m13 i1m i3))) ts)))))))))
      (values (list Gci (chi-square-significance Gci DFci)))))

;;; ---------------------------------------------------------------------------

(defun 3d-one-factor-independence (ct m23 m13 m12 ts)
  "Performs the three possible one-factor independence tests on a 3 factor table. Returns G1 p2 G2 p2 G3 p3.
params = (3way-ctable 2way-23-margins 2way-13-margins 2way-12-margins data-length)"
  (let* ((dim1 (1- (array-dimension ct 0)))
	 (dim2 (1- (array-dimension ct 1)))
	 (dim3 (1- (array-dimension ct 2)))
	 (DF1 (* (1- (* (1+ dim3) (1+ dim2))) dim1))
         (DF2 (* (1- (* (1+ dim1) (1+ dim3))) dim2))
	 (DF3 (* (1- (* (1+ dim1) (1+ dim2))) dim3))
	 G1 G2 G3 P1 P2 P3)
    ;; factor 1 independence
    (setq G1 (* 2 (loop for i1 from 0 to dim1 sum
			;; frequency of factor 1 at current level
			(let ((f1 (loop for i2 from 0 to dim2 sum (aref m12 i1 i2))))
			  ;; vary the "free" terms and compute the G contribution
			     (loop for i2 from 0 to dim2 sum
				   (loop for i3 from 0 to dim3 sum
					 ;; G
					 (* (aref ct i1 i2 i3)
					    (log (/ (aref ct i1 i2 i3)
						    ;; expected frequency
						    (/ (* (aref m23 i2 i3) ;; margin value varying factor 1
							  f1) ts))))))))))
    ;; factor 2 independence
    (setq G2 (* 2 (loop for i2 from 0 to dim2 sum
			;; frequency of factor 2 at current level
			(let ((f2 (loop for i3 from 0 to dim3 sum (aref m23 i2 i3))))
			  ;; vary the "free" terms and compute the G contribution
			     (loop for i1 from 0 to dim1 sum
				   (loop for i3 from 0 to dim3 sum
					 ;; G
					 (* (aref ct i1 i2 i3)
					    (log (/ (aref ct i1 i2 i3)
						    ;; expected frequency
						    (/ (* (aref m13 i1 i3) ;; margin value varying factor 1
							  f2) ts))))))))))
    ;; factor 3 independence
    (setq G3 (* 2 (loop for i3 from 0 to dim3 sum
			;; frequency of factor 3 at current level
			(let ((f3 (loop for i2 from 0 to dim2 sum (aref m23 i2 i3))))
			  ;; vary the "free" terms and compute the G contribution
			     (loop for i1 from 0 to dim1 sum
				   (loop for i2 from 0 to dim2 sum
					 ;; G
					 (* (aref ct i1 i2 i3)
					    (log (/ (aref ct i1 i2 i3)
						    ;; expected frequency
						    (/ (* (aref m12 i1 i2) ;; margin value varying factor 1
							  f3) ts))))))))))
    ;; P values (chi square)
    (setq p1 (chi-square-significance G1 DF1) p2 (chi-square-significance G2 DF2) p3 (chi-square-significance G3 DF3))
    (values (list (list G1 p1) (list G2 p2) (list G3 p3)))))

;;; ---------------------------------------------------------------------------

(defun 3d-conditional-independence (ct m23 m13 m12)
  "Performs the three possible conditional independence analyses. Returns g23|1 p23|1 g13|2 p13|2 g12|3 p12|3.
params = (3way-ctable 2way-23-margins 2way-13-margins 2way-12-margins data-length)"
  (let* ((dim1 (1- (array-dimension ct 0)))
	 (dim2 (1- (array-dimension ct 1)))
	 (dim3 (1- (array-dimension ct 2)))
	 (DF1ci (* (1+ dim1) dim2 dim3)) (DF2ci (* dim1 (1+ dim2) dim3)) (DF3ci (* dim1 dim2 (1+ dim3)))
	 G1 G2 G3 P1 P2 P3)
    ;; 1 X 2 given 3
    ;; iterate over "slices" of the conditional
       (setq G3
	     (* 2 (loop for i3 from 0 to dim3 sum
			(let ((f3 (loop for i1 from 0 to dim1 sum (aref m13 i1 i3))))
			  ;; calculate G for each slice
			  (loop for i1 from 0 to dim1 sum
				(loop for i2 from 0 to dim2 sum
				      (* (aref ct i1 i2 i3)
					 (log (/ (aref ct i1 i2 i3)
						 ;; expected frequency
						 (/ (* 1.0 (aref m13 i1 i3)
						       (aref m23 i2 i3))
						    f3)))))))))
	     ;; 1 X 3 given 2
	     G2
	     (* 2 (loop for i2 from 0 to dim2 sum
			(let ((f2 (loop for i1 from 0 to dim1 sum (aref m12 i1 i2))))
			  (loop for i1 from 0 to dim1 sum
				(loop for i3 from 0 to dim3 sum
				      (* (aref ct i1 i2 i3)
					 (log (/ (aref ct i1 i2 i3)
						 ;; expected frequency
						 (/ (* 1.0 (aref m12 i1 i2)
						       (aref m23 i2 i3))
						    f2)))))))))
	     ;; 2 X 3 given 1
	     G1
	     (* 2 (loop for i1 from 0 to dim1 sum
			(let ((f1 (loop for i2 from 0 to dim2 sum (aref m12 i1 i2))))
			  (loop for i3 from 0 to dim3 sum
				(loop for i2 from 0 to dim2 sum
				      (* (aref ct i1 i2 i3)
					 (log (/ (aref ct i1 i2 i3)
						 ;; expected frequency
						 (/ (* 1.0 (aref m12 i1 i2)
						       (aref m13 i1 i3))
						    f1))))))))))
       (setq P1 (chi-square-significance G1 DF1ci) P2 (chi-square-significance G2 DF2ci)
	     P3 (chi-square-significance G3 DF3ci))
       (values (list (list G1 P1) (list G2 P2) (list G3 P3)))))

;;; ---------------------------------------------------------------------------

(defparameter wick #3A(((7 5 3) (15 20 2)) ((22 49 140) (4 3 5))))

(defun 3d-homogenous-association (ct &key (reps 4) (max-error nil))
  (let* ((is (make-array (array-dimensions ct) :initial-element 1))
	 (cm (loop for i from 0 to (1- (array-dimension ct 0)) collect
		   (loop for i2 from 0 to (1- (array-dimension ct 1)) collect
			 (loop for i3 from 0 to (1- (array-dimension ct 2)) sum
			       (aref ct i i2 i3)))))
	 (bm (loop for i3 from 0 to (1- (array-dimension ct 2)) collect
		   (loop for i from 0 to (1- (array-dimension ct 0)) collect
			 (loop for i2 from 0 to (1- (array-dimension ct 1)) sum
			       (aref ct i i2 i3)))))
	 (rm (loop for i2 from 0 to (1- (array-dimension ct 1)) collect
		   (loop for i3 from 0 to (1- (array-dimension ct 2)) collect
			 (loop for i from 0 to (1- (array-dimension ct 0)) sum
			       (aref ct i i2 i3)))))
         DF G P)
    (loop for x from 0 to (1- reps) do
	  (progn
	    ;; do cols
	    (loop for ctl in cm
		  for col from 0 do
		  (loop for cti in ctl
			for coli from 0 do
			(let ((mul (float (/ cti (loop for i3 from 0 to (1- (array-dimension is 2)) sum
						       (aref is col coli i3))))))
			  (loop for i3 from 0 to (1- (array-dimension is 2)) do
				(setf (aref is col coli i3) (* (aref is col coli i3) mul))))))
	    ;; do beams
	    (loop for btl in bm
		  for bol from 0 do
		  (loop for bti in btl
			for boli from 0 do
			(let ((mul (float (/ bti (loop for i2 from 0 to (1- (array-dimension is 1)) sum
						       (aref is boli i2 bol))))))
			  (loop for i2 from 0 to (1- (array-dimension is 1)) do
				(setf (aref is boli i2 bol) (* (aref is boli i2 bol) mul))))))
	    ;; do rows
	    (loop for rtl in rm
		  for rol from 0 do
		  (loop for rti in rtl
			for roli from 0 do
			(let ((mul (float (/ rti (loop for i from 0 to (1- (array-dimension is 0)) sum
						       (aref is i rol roli))))))
			  (loop for i from 0 to (1- (array-dimension is 0)) do
				(setf (aref is i rol roli) (* (aref is i rol roli) mul))))))
            #+DEBUG
            (format t "~S~%~S~%~S~%"
		    (loop for i from 0 to (1- (array-dimension ct 0)) collect
			  (loop for i2 from 0 to (1- (array-dimension ct 1)) collect
				(loop for i3 from 0 to (1- (array-dimension ct 2)) sum
				      (aref is i i2 i3))))
		    (loop for i3 from 0 to (1- (array-dimension ct 2)) collect
			  (loop for i from 0 to (1- (array-dimension ct 0)) collect
				(loop for i2 from 0 to (1- (array-dimension ct 1)) sum
				      (aref is i i2 i3))))
		    (loop for i2 from 0 to (1- (array-dimension ct 1)) collect
			  (loop for i3 from 0 to (1- (array-dimension ct 2)) collect
				(loop for i from 0 to (1- (array-dimension ct 0)) sum
				      (aref is i i2 i3)))))
            #+DEBUG
            (ptt is)))
    (setq DF (reduce #'* (mapcar #'1- (array-dimensions is))))
    (setq G (* 2.0 (loop for i from 0 to (1- (array-dimension is 0)) sum
                         (loop for i2 from 0 to (1- (array-dimension is 1)) sum
                               (loop for i3 from 0 to (1- (array-dimension is 2)) sum
                                     (* (aref ct i i2 i3)
				        (log (/ (aref ct i i2 i3)
					        (aref is i i2 i3)))))))))
    (values (list G (chi-square-significance G DF)))
    ))

(defun analyze-3way-relationship (v1 v2 v3 &key (cutoff-p 0.25) (complete? t) (one-way? t) (conditional? t) (homogenous? 3))
  "The driver for 3way ct analysis. Creates the set of 3way tables and performs the requested
analyses. 'cuttoff-p' is a p-value, that when exceeded, will cut off execution."
  (let ((cutoff? NIL)
        (L (length v1)))
    (multiple-value-bind (ct m23 m13 m12 v1s v2s v3s) (make-3d-contingency-table-set v1 v2 v3)
      (declare (ignore v1s v2s v3s))
      (append
       (when complete?
         (multiple-value-list (3d-complete-independence    ct m23 m13 m12 L)))
       (when (and one-way? (not cutoff?))
         (multiple-value-list (3d-one-factor-independence  ct m23 m13 m12 L)))
       (when (and conditional? (not cutoff?))
         (multiple-value-list (3d-conditional-independence ct m23 m13 m12)))
       (when homogenous?
         (multiple-value-list (3d-homogenous-association ct)))))))
