;;;; -*- Mode: Common-Lisp; Package: CLASP; Base: 10 -*-
;;;; *-* File: /usr/users/eksl/systems/clasp/clasp/dev/dataset-fns.lisp *-*
;;;; *-* Last-Edit: Monday, October 2, 1995  17:05:03; Edited-By: Westy *-*
;;;; *-* Machine: Mesmer (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                           Data Set Functions                           *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Adam Carlson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  06-24-92 File Created.  (Carlson)
;;;  02-13-93 Added uses of *dataset-class* and *variable-class*.  (Westy)
;;;  06-21-94 Began changes to remove RTM. (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :clasp)

;;; --*--
;;; ***************************************************************************

;;; ***************************************************************************
;;;; Dataset manipulation functions
;;; ***************************************************************************

;;; The following functions are used to maintain the datasets in a session.  At
;;; the dataset level, they include creation, destruction, activatation and
;;; deactivation of datasets.  At the variable level they include adding and
;;; deleting of variables (columns), adding and deleting of tuples (rows), and
;;; movement of variables between datasets.

;;; ***************************************************************************
;;;; Dataset level
;;; ***************************************************************************

(defun make-dataset (&rest args &key (name "Dataset") (description nil) (create-row-number-variable t))
  "Creates a new dataset and fills the `name' and `description' slots.
Make-dataset also initializes a row-number variable for the dataset."
  (let* ((new-dataset  (apply #'make-instance *dataset-class*
			      :name (make-clasp-name name :dataset)
			      :variables nil
			      :variable-count (make-hash-table :test #'equal)
			      :description description
			      :rows 0
			      ; This should not be necessary (I think) but
			      ; MCL blows up if we do not include it.
			      :allow-other-keys t
			      args)))
    #+OLD-VARS
    (setf (dataset-write-data new-dataset)
	  (intern (make-clasp-name "RTM-TABLE" :data)))
    #+OLD-VARS
    (setf (dataset-read-data new-dataset) (dataset-write-data new-dataset))
    #+OLD-VARS
    (unless (create-table (dataset-write-data new-dataset))
      (error 'rtm-table-creation-error))
    (when create-row-number-variable
      (push (make-variable
             :class *row-number-variable-class*
	     :name 'row-number
	     #+OLD-VARS
	     :attribute #+OLD-VARS '(row-number :number-or-symbol :index)
	     :dataset new-dataset)
	    (dataset-variables new-dataset)))
    (setf (dataset-where-clause new-dataset) nil)
    (push new-dataset *clasp-datasets*)
    (when *preferences-activate-automatically*
      (push new-dataset *clasp-active-datasets*))
    (values new-dataset)))


;; If this is changed remember that there is an optimized version used when loading
;; datasets that should also be changed.
(defmethod make-dataset-from-rows
    (name data column-names &rest args &key description report-stream push-rows)
  "Creates a new data set and fills the `name' and `description' slots.  Then
it creates variables for the dataset using `column-names' as the names of the
variables and `data' as the data.  `Data' should be in row-major order."
  (check-type name (or symbol string))
  (check-type data sequence)
  (check-type column-names sequence)
  (check-type description (or string null))
  
  (if (not (= (length column-names) (length (car data))))
      (warn 'row-length-mismatch))
  (let ((new-dataset (apply #'make-dataset :name name :description description
			    :allow-other-keys t
			    args)))
    (dotimes (n (length column-names))
      (add-variable-to-dataset new-dataset (list (nth n (car data)))
			       (nth n column-names)))
    (add-rows-to-dataset new-dataset (cdr data)
			 (remove-if #'row-number-p
				    (dataset-variables new-dataset))
                         report-stream
                         push-rows)
    (values new-dataset)))

(defmethod make-dataset-from-columns
    (name data column-names &rest args &key description)
  "Creates a new data set and fills the `name' and `description' slots.  Then
it creates variables for the dataset using `column-names' as the names of the
variables and `data' as the data.  `Data' should be in column-major order."
  (check-type name (or symbol string))
  (check-type data sequence)
  (check-type column-names sequence)
  (check-type description (or string null))

  (when (not (= (length data) (length column-names)))
    (warn 'row-length-mismatch))
  (let ((new-dataset (apply #'make-dataset :name (string name)
			    :description description
			    :allow-other-keys t
			    args)))
    (loop for column-name in column-names
	for column-data in data
	do (add-variable-to-dataset new-dataset column-data column-name))
    new-dataset))

(defparameter *missing-data-value* nil)

(defmethod merge-datasets (datasets)
  (setf datasets (mapcar #'get-dataset datasets))
  (let* ((new-dataset (make-dataset :name
				    (format nil "Merge of ~{~a~^ ~}"
					    (mapcar #'name datasets))))
	 ;; Record all the unique variable names.
         (old-variable-names (loop for dataset in datasets
                                   append
                                   (loop for variable in (dataset-variables dataset)
                                         unless (or (row-number-p variable)
                                                    (member variable variable-names :test #'(lambda (variable name)
                                                                                              (string= (name variable) name))))
				         collect (name variable))
                                   into variable-names
				   finally (return variable-names)))
	 (source-variable-name (name (add-variable-to-dataset new-dataset '() "SOURCE")))
	 ;; Add all the variables to the new-dataset and save their new name.
         ;; We do this in case their are name collisions between any of these and the `source'
         ;; and `row-number' variables. If there are collisions some will be renamed
         ;; when the variable is created.
         (new-variable-names (loop for variable-name in old-variable-names
                                   collect (name (add-variable-to-dataset new-dataset '() variable-name))))
         
         )


    (flet ((extract-padded-rows (dataset)
             (loop for row-number from 0 to (1- (dataset-rows dataset))
	           collect
                   (loop for old-variable-name in old-variable-names
                         for old-variable = (get-variable old-variable-name dataset)
		         collect
		         (if old-variable
		             (variable-value old-variable :where row-number)
		             *missing-data-value*)))))

      ;; Go through all the datasets and extract rows from them padding to match the number of columns in the
      ;; new dataset.
      (dolist (dataset datasets)
        (add-rows-to-dataset
         new-dataset
         ;; Extract the rows from this dataset
         (mapcar #'(lambda (row)
                     ;; Add the source
		     (cons (name dataset) row))
                 (extract-padded-rows dataset))
         
         (cons source-variable-name new-variable-names))))

    new-dataset))

(defun rename-dataset (dataset new-name)
  "rename-dataset changes the name of `dataset to `new-name', insuring the
new name is unique.  `dataset' can either be a dataset or the name of a
dataset, and `new-name' must be a string."
  (setf dataset (get-dataset dataset))
  (setf (name dataset) (make-clasp-name new-name :dataset))
  (values))

(defun delete-dataset (dataset)
  "Destroys a dataset.  `dataset' can either be a dataset or the name of a
dataset."
  (setf dataset (get-dataset dataset)) ;If this line is removed, add checking of dataset existance
  (dolist (variable (dataset-variables dataset))
    (delete-variable variable))
  (setf (name dataset) nil
	;;(slot-value dataset 'name) nil
	(dataset-variable-count dataset) nil
        #+OLD-VARS
	(dataset-write-data dataset) #+OLD-VARS nil
	#+OLD-VARS
	(dataset-read-data dataset) #+OLD-VARS nil
	(dataset-where-clause dataset) nil
	(dataset-variables dataset) nil
	(dataset-variable-count dataset) nil
	(dataset-rows dataset) nil)
  (setf *clasp-datasets* (remove dataset *clasp-datasets*))
  (setf *clasp-active-datasets* (remove dataset *clasp-active-datasets*))
  #+ignore
  (drop-table (dataset-read-data dataset))
  #+ignore
  (drop-table (dataset-write-data dataset))
  (kill dataset)
  (values))


(defun activate-dataset (dataset)
  "Puts a dataset on the clasp active-dataset list."
  (setf dataset (get-dataset dataset nil)) ;If this line is removed, add checking of dataset existance
  (unless (find dataset *clasp-active-datasets*)
    (push dataset *clasp-active-datasets*)))

(defun deactivate-dataset (dataset)
  "Takes a dataset off of the clasp active-datasets list."
  (setf *clasp-active-datasets* (remove (get-dataset dataset)
					*clasp-active-datasets*)))

(defmethod get-dataset ((the-dataset dataset)
			&optional (active-datasets-only-p t)
				  (not-found-error-p t))
  "Given a name or id, finds the dataset.  If `active-datasets-only-p' is nil,
will search through all current clasp datasets, otherwise, only the active
datasets are searched."
  (declare (ignore active-datasets-only-p not-found-error-p))
  the-dataset)

(defmethod get-dataset ((name string)
			&optional (active-datasets-only-p t)
				  (not-found-error-p t))
  "Given a name or id, finds the dataset.  If `active-datasets-only-p' is nil,
will search through all current clasp datasets, otherwise, only the active
datasets are searched."
  (or (find name (if active-datasets-only-p
		      *clasp-active-datasets*
		    *clasp-datasets*)
	    :key #'name
	    :test #'string-equal)
      (when not-found-error-p (error 'dataset-not-found #+ANSI-CONDITIONS :dataset #+ANSI-CONDITIONS name))))

(defmethod get-dataset ((name symbol)
			&optional (active-datasets-only-p t)
				  (not-found-error-p t))
  "Given a name or id, finds the dataset.  If `active-datasets-only-p' is nil,
will search through all current clasp datasets, otherwise, only the active
datasets are searched."
  (or (find name (if active-datasets-only-p
		     *clasp-active-datasets*
		   *clasp-datasets*)
	    :key #'name)
      (find (format nil "~a" name) (if active-datasets-only-p
				       *clasp-active-datasets*
				     *clasp-datasets*)
	    :key #'name :test #'string-equal)
      (when not-found-error-p (error 'dataset-not-found #+ANSI-CONDITIONS :dataset #+ANSI-CONDITIONS name))))

;;; ***************************************************************************
;;;; Variable level
;;; ***************************************************************************

(defun make-variable (&key (class *variable-class*) name attribute value description dataset rtm-table (unique-name t))
  "make-variable creates a new variable.  `name' is a string.  `value'
is the data the variable should be initialized with.  `description' is
a description of the variable.  `dataset' is the dataset to which the
variable belongs."
  (declare (ignore rtm-table attribute)) 
  (let* (new-variable)
    (setf new-variable (make-instance class
			 :name (if unique-name (make-clasp-name name :variable dataset) name)
			 :description description
			 :dataset dataset))
    #+OLD-VARS
    (setf (variable-attribute new-variable)
      (or attribute (list (id new-variable) :number-or-symbol)))
    #+OLD-VARS
    (create-attribute (or rtm-table (dataset-write-data dataset))
		      (variable-attribute new-variable))
    (when value
      (setf (variable-value new-variable) value))
    new-variable
    )
  )


(defmethod add-variable-to-dataset (dataset data variable-name
				    &optional (description nil))
  "add-variable-to-dataset adds the `data', which must be a list of values, to
`dataset' using `variable-name' as the name of the new variable and
`description' as a description."
  (check-type dataset dataset)
  (check-type data sequence)
  (check-type variable-name (or symbol string))
  (check-type description (or string null))

  (if (< (length data) (dataset-rows dataset))
      (warn 'column-length-mismatch))
  (let ((new-variable (make-variable
			:name variable-name
			;; :type (car data)
			:description description
			:dataset dataset)))
    (push-end new-variable (dataset-variables dataset))
    (setf (variable-value new-variable) data)
    new-variable
    )
  )

(defmethod add-variables-to-dataset (dataset data-list variable-names
				     &optional (descriptions nil))
  "add-variables-to-dataset adds a number of variables to `dataset'.  For each
variable, i, its value (which must be a list) is the ith element of
`data-list', its name is the ith element of `variable-names', and its
description is the ith element of descriptions."
  (check-type dataset dataset)
  (check-type data-list sequence)
  (check-type variable-names sequence)
  (check-type descriptions (or sequence null))

  (dotimes (variable-number (length data-list))
    (let* ((variable-name (nth variable-number variable-names))
	   (description (nth variable-number descriptions))
	   (data (nth variable-number data-list))
	   (new-variable (make-variable
			   :name variable-name
			   ;; :type (car data)
			   :description description
			   :dataset dataset)))
      (push-end new-variable (dataset-variables dataset))
      (setf (variable-value new-variable) data)
      )
    )
  )

(defmethod rename-variable (variable new-name)
  "rename-variable changes the name of `variable' to `new-name', insure the
new name is unique.  `variable' can either be a variable or the name of a
variable, and `new-name' must be a string."
  (setf variable (get-variable variable)) ;If this line is removed, add checking of variable existance
  (if (row-number-p variable)
      (error 'rename-row-number)
    (setf (name variable) (make-clasp-name new-name :variable
					     (variable-dataset variable))))
  (values))


(defmethod delete-variable (variable)
  "Destroys a variable.  `variable' can either be a variable or the name of a
variable."
  (setf variable (get-variable variable)) ;If this line is removed, add checking of variable existance
  (let ((dataset (variable-dataset variable)))
    #+ignore
    (drop-attribute (dataset-write-data dataset) (name variable))
    (setf (dataset-variables dataset)
      (remove variable (dataset-variables dataset))))
  (setf ;;(slot-value variable 'name) nil
	(name variable) nil
	(description variable) nil
	(slot-value variable 'value) nil
	(slot-value variable 'order-by-array-cache) nil
	(slot-value variable 'order-by-list-cache) nil
	(variable-dataset variable) nil
	#+OLD-VARS
	(variable-attribute variable) #+OLD-VARS nil)
  (kill variable)
  (values))

(defgeneric get-variable (the-variable &optional dataset active-datasets-only-p not-found-error-p)
  (:documentation "Given a name or id, finds the variable.  If `dataset' is specified, that is
the dataset which will be searched for `variable'.  Otherwise, a search will be
made across datasets.  If this is the case, then `active-datasets-only-p' will
determine whether all datasets currently loaded are searched, or just the
active ones.  Note, since variable names must only be unique within a dataset,
not across datasets, care must be taken to request the correct variable.  It is
advisable to call get-variable with the dataset optional variable set."))


(defmethod get-variable ((the-variable variable)
			 &optional dataset (active-datasets-only-p t)
				   (not-found-error-p t))
  (declare (ignore dataset active-datasets-only-p not-found-error-p))
  the-variable)

(defmethod get-variable ((variable-name symbol)
			 &optional dataset (active-datasets-only-p t)
				   (not-found-error-p t))
  (get-variable (symbol-name variable-name) dataset active-datasets-only-p not-found-error-p))

(defmethod get-variable ((variable-name string)
			 &optional dataset (active-datasets-only-p t)
				   (not-found-error-p t))
  (flet ((find-variable (name dataset)
           (find name (dataset-variables dataset)
		 :key #'name :test #'string-equal)))
    (cond (dataset
	   (unless (typep dataset 'dataset)
	     (error 'dataset-not-found #+ANSI-CONDITIONS :dataset #+ANSI-CONDITIONS dataset))
	   (find-variable variable-name dataset))
	  (t
	   (loop for dataset in (if active-datasets-only-p
				    *clasp-active-datasets*
				  *clasp-datasets*)
	       for found = (find variable-name (dataset-variables dataset)
				 :test #'(lambda (name variable)
					   (or (string-equal name (name variable))
					       (string-equal name (id variable)))))
	       until found
	       finally (return (cond (found)
				     (not-found-error-p 
				      (error 'variable-not-found #+ANSI-CONDITIONS 
					     :variable #+ANSI-CONDITIONS variable-name))
				     (t nil))))))))
  
;;; ***************************************************************************
;;;; Row level
;;; ***************************************************************************

(defmethod add-row-to-dataset (dataset data variables push-on-front &optional defer-updating-rows)
  "This function will add `data' to the dataset `dataset' as a new row.
`variables' is a list of variables and specifies what order the new row is in
with respect to the variables in the dataset.  Therefore, `data' must match
`variables' in both length and the types of its values.

Note: Whenever this function is called, clear-cache should be called on the
same dataset."
  (check-type data sequence)
  (check-type variables sequence)
  
  (setf dataset (get-dataset dataset))
  (let ((row-number (incf (dataset-rows dataset)))
	(list-length (length variables))
	#+OLD-VARS
	pos
	#+OLD-VARS
	addition)
    (if (not (= list-length (length data)))
	(warn 'row-length-mismatch))
    #+OLD-VARS
    (dotimes (n list-length)
      (setf pos (- list-length n 1))
      (push (nth pos data) addition)
      #+OLD-VARS
      (push (car (variable-attribute
		  (get-variable (nth pos variables) dataset))) addition))
    #+OLD-VARS
    (push row-number addition)
    #+OLD-VARS
    (push 'row-number addition)
    #+OLD-VARS
    (insert-into-table (dataset-write-data dataset) addition)

    (loop for variable in variables
	  for data-point in data do
	  (add-data-point (get-variable variable dataset) data-point push-on-front))
    (unless defer-updating-rows
      (update-row-number-for-changed-variable-value dataset row-number)))
  (values))


;; If this is changed remember that there is an optimized version used when loading
;; datasets that should also be changed.
(defmethod add-rows-to-dataset (dataset data-list variables &optional report-stream push-rows)
  "This function will add `data-list' to the dataset `dataset'.  `data-list' is
a list of lists representing rows of data.  The rows must be of equal length.
`variables' is a list of variables and specifies what order the new rows are in
with respect to the variables in the dataset.  Therefore, each element of
`data-list' must match variables in both length and the types of its elements.
If `push-rows' is non-nil the data will be added to the front of each column."
  
  (check-type data-list sequence)
  (check-type variables sequence)
  
  (clear-cache dataset)
  (with-progress-thermometer (report-stream 100 10000 :start 1 :initial-string "Processing")
    (loop for data in data-list
          for row-number from (1+ (dataset-rows dataset))
          do
          (add-row-to-dataset dataset data variables push-rows t)
          (note-progress)
          finally
          (update-row-number-for-changed-variable-value dataset row-number))))

#+BROKEN
(defmethod delete-row-from-dataset (dataset row-number)
  (let ((dataset (get-dataset dataset)))
    (loop for variable in (dataset-variables dataset) do
          (delete-data-point variable row-number))
        
    (decf (dataset-rows dataset))
    (update-row-number-for-changed-variable-value dataset row-number))
  (values))

;;;

(defmacro selecting-rows ((values dataset variables where where-variables &optional from-end)
			&body body)
  "Map over the rows in `dataset' that satisfy the `where' predicate applied
to the `where-variables'. `values' will be bound to a list of the values of
`variables' in the row."
  (with-variables (row-number where-vars)
    `(loop with variables = (mapcar #'(lambda (v) (get-variable v ,dataset)) ,variables)
      ,@(if from-end
	    `(for ,row-number from (1- (dataset-rows ,dataset)) downto 0)
            `(for ,row-number from 0 to (1- (dataset-rows ,dataset))))
      do
      (let ((,where-vars
	     (loop for wvar in ,where-variables
		   collect
		   (variable-value wvar :where ,row-number))))
	(when (apply ,where ,where-vars)
	  (let ((,values
		 (loop for var in variables
		       collect
		       (variable-value var :where ,row-number))))
	    (progn . ,body)))))))

(defun map-dataset (dataset function &rest args)
  (loop for row-number from 0 to (1- (dataset-rows dataset)) do
        (let ((processed-args
	       (loop for arg in args
		     collect
                     (if (typep arg 'variable)
		       (variable-value arg :where row-number)
                       args))))
	  (apply function processed-args))))

(defun sum-distinct-values (dataset sum-variable where &rest where-args)
  (let ((assoc-table nil))
    (apply #'map-dataset dataset #'(lambda (sum &rest where-args)
                             (when (apply where where-args)
                               (let ((old-value (assoc sum assoc-table :test #'equal)))
                                 (if old-value
                                   (setf (rest old-value) (1+ (rest old-value)))
                                   (setf assoc-table (acons sum 1 assoc-table))))))
                 (cons sum-variable where-args))
    (values assoc-table)))

;;; ***************************************************************************
;;; EOF
