;;;; -*- Mode:Common-Lisp; Package:CLASP; Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/clasp/clasp/development/manipulation-fns.lisp *-*
;;;; *-* Last-Edit: Wednesday, December 14, 1994  13:57:02; Edited-By: westy *-*
;;;; *-* Machine: Earhart (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                       Data Manipulation Functions                      *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Copyright (c) 1990 - 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.

;;; For more information write to clasp-support@cs.umass.edu

;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  02-25-93 File Created.  (Carlson)
;;;  11-28-94 Fixed a variable lookup bug in the friends of 'partition-on'.
;;;           (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :clasp)

;;; --*--
;;; ***************************************************************************


(defun create-new-column-from-expression (dataset expression)
   "`expression' must be a valid lisp expression, it may contain the
names of variables in `dataset'.  The rows of `dataset' are mapped
over, substituting the values of the variables in `expression', after
substitution, `expression' is evaluated.  This produces a new column
of data which is added as a variable to `dataset'."
  (check-type dataset dataset)
  (check-type expression list)
  
  (add-variable-to-dataset dataset
                           (apply-expression-to-variables dataset expression)
                           (princ-to-string expression)))

(defun apply-expression-to-variables (dataset expression)
  (let* ((new-expression (replace-variables expression dataset))
	 (variable-list (get-variable-list new-expression
					   (dataset-variables dataset)))
	 data-list
	 lambda-expression)
    (setf data-list (dataset-to-columns dataset :columns variable-list))
    (setf lambda-expression
      (maybe-compile-lambda (list 'lambda (mapcar #'id variable-list) new-expression)))
    (apply #'mapcar lambda-expression data-list)))


(defun create-new-column-from-function (dataset new-name function &rest arg-variables)
  "`function' must be a valid lisp function, while `arg-variables' should be a
list of variables that are used to create the arguments to the function.  The
rows of `dataset' are mapped over, calling `function' with values extracted
from each of the `arg-variables' which are then. This produces a new column of
data which is added as a variable to `dataset'."
  (check-type dataset dataset)
  
  (add-variable-to-dataset dataset
			   (apply-function-to-variables dataset function arg-variables)
			   new-name))

(defun apply-function-to-variables (dataset function arg-variables)
  (apply #'mapcar function (dataset-to-columns dataset :columns arg-variables)))

(defun recode (value old-values new-values)
  
  "Maps `value' from it's position in `original-values' to the value at the equivalent
position in `new-values'.  For instance, (recode 2 '(1 2 3) '(a b c)) would return 'b.  
This function can be used with create-new-column-from-function to recode a categorical
variable. From the program interface, it is recommended that ``recode-list''
be used instead."
  
  (check-type value atom)
  (check-type old-values list)
  (check-type new-values list)
  
  (aif (position value old-values :test #'string-equal)
    (nth it new-values)
    (error "value, ~s, is not in old-values, ~s." value old-values)))


(defun recode-list (list old-values new-values)
  "Maps each element of `list' from it's position in `old-values' to
the value at the equivalent position in `new-values'."
  (map nil
       #'(lambda (new old)
	   (setf list (substitute new old list :test #'string-equal)))
       new-values
       old-values)
  list)

(defun get-variable-list (expression legal-variables)
  (unless (null expression)
    (let (variable-list
          variable)
      (map nil
	#'(lambda (x)
	    (cond
	     ((listp x)
	      (map nil #'(lambda (x)
			   (pushnew x variable-list))
		   (get-variable-list x legal-variables)))
	     ((numberp x))
	     ((setf variable (find x legal-variables :key #'name :test #'string-equal))
	      (pushnew variable  variable-list))
             ((setf variable (find x legal-variables :key #'id))
              (pushnew variable variable-list))))
	expression)
      (remove-duplicates variable-list))))

(defun replace-variables (expression dataset)
  (setf dataset (get-dataset dataset))
  (let ((sublis-alist (mapcar #'(lambda (var) (cons (name var) (id var)))
			      (clasp::dataset-variables dataset))))
    (sublis sublis-alist expression :key #'(lambda (x) (format nil "~a" x))
	    :test #'string-equal)))

;;; ***************************************************************************
;;;; Partitioning
;;; ***************************************************************************

;;; Obsolete notes: 7/15/94  (Westy)
;;; Note: the partitioning code is going to have to make actual copies
;;; of the variable objects in order to make proper use of the
;;; variable-value method.  As it is there is no way for variable value
;;; to know whether a variable belongs to a "source" dataset or a
;;; partitioned view of a source dataset, therefore it always gives all
;;; the data.

#|

Often it is desirable to operate on a portion of a dataset, or to
separate out different parts of a dataset according to the values of
some of the variables.  When a dataset is broken up this way, it is
called partitioning, and the parts of the dataset created are called
partitions.  partition-dataset takes a dataset and a partition-clause
which describes how to split the dataset up, and returns a list of
datasets, each one containing one partition.  (Note, this does not
destroy the original dataset.)

Partition clauses are made up of boolean operators, and the
partitioning operator, ``.on.''.  The boolean operators act as
expected and are explained in detail in the reference manual.  The
.on. operator takes any number of variables as its operands, and
separates the dataset into different partitions for each unique set of
values of its operands.  For example, if a dataset consisted of a
boolean variable ``key1'', a categorical variable ``key2'', with
values 'a, 'b and 'c, and a data variable ``y'', and was partitioned
with a partitioning clause of (.on. key1 key2), then at most six new
datasets would be created, one containing all the rows of the original
dataset where key1 = true and key2 = 'a, one with all the rows where
key1 = false and key2 = 'a, one with all the rows where key1 = true
and key2 = 'b, and so on.  If a partition would be empty (i.e. if
there were no rows where key1 = false and key2 = 'c), then no dataset
is created for that partition, which is why AT MOST six new datasets
would be created in the above example.

The .on. operator can be used in conjunction with the boolean
operators.  When this happens, the .on. operator can be thought of as
a macro which causes a set of partitioning clauses to be created, each
exactly like the original, except that .on. is replaces with an .==.
operator and one of the unique values of the key.  If there are
multiple keys in the .on., it is replaced by (.and. (.==. key1 value11)
(.==. key2 value21), etc.).  In the above example, a partition clause
of (.and. (.==. key1 true) (.on. key2)) would cause up to three
datasets to be created.  Each would contain rows where key1 was true,
and one would be created for key2 = 'a, one for key2 = 'b and one for
key2 = 'c."

|#


(defun partition-dataset (old-dataset where where-variables &optional variables expression)
  (setf old-dataset (get-dataset old-dataset))
  (if variables
      (setf variables (mapcar #'(lambda (x) (get-variable x old-dataset))
			      variables))
    (setf variables (dataset-variables old-dataset)))
  (setf where-variables
	(mapcar #'(lambda (x) (get-variable x old-dataset))
		where-variables))

  (multiple-value-bind (new-dataset new-variables)
      (build-new-dataset old-dataset variables (make-clasp-name
                                                (if expression
                                                    (concatenate 'simple-string
                                                                 (name old-dataset)
                                                                 "-WHERE"
							         (infix-string expression))
						    (name old-dataset))
                                                :dataset
                                                nil
                                                nil) nil where)
    ;; Add the data from the old dataset to the new one
    (selecting-rows (row old-dataset variables where where-variables t)
		    (add-row-to-dataset new-dataset row new-variables t))
    (values new-dataset))
   )

(defvar *more-partitions-than-you-really-want* 50)

(defun partition-on (old-dataset variables &optional include-variables suppress-warnings)
  (if (null (cdr variables))
    (partition-on-single-variable old-dataset (first variables) include-variables suppress-warnings)
    (partition-on-multiple-variables old-dataset variables include-variables suppress-warnings)))

(defun partition-on-single-variable (old-dataset variable variables &optional suppress-warnings)
  (setf old-dataset (get-dataset old-dataset))
  (setf variable (get-variable variable old-dataset))
  (if variables
      (setf variables (mapcar #'(lambda (x) (get-variable x old-dataset))
			      variables))
    (setf variables (dataset-variables old-dataset)))
  (let ((distinct-values (variable-value variable :distinct t))
        (dataset-assoc-list nil))
    ;; Wanted to use handlers and restarts to suppress this, but can not get it to work...
    (when (and (not suppress-warnings)
               (> (length distinct-values) *more-partitions-than-you-really-want*))
        (cerror "Partition it anyway." 'more-partitions-than-you-really-want :variable variable :number-of-partitions (length distinct-values))
        )
      
    (loop for new-value in distinct-values
          do
          (multiple-value-bind (new-dataset new-variables)
              (build-new-dataset old-dataset variables
                                 (make-clasp-name (format nil "~a-where-~a=~a"
                                                          (name old-dataset)
                                                          (name variable)
                                                          new-value)
                                                  :dataset))
            (setf dataset-assoc-list (acons new-value (list new-dataset new-variables) dataset-assoc-list))))
    
    ;; We might want to consider mapping over these in reverse order so we can push the values on the new dataset.
    ;; Currently 'add-row-to-dataset' nconc's things on the end.
    (loop for row-number from (1- (dataset-rows old-dataset)) downto 0 do
	  (destructuring-bind (new-dataset new-variables) (rest (assoc 
							   (variable-value variable :where row-number)
							   dataset-assoc-list
							   :test #'equalp ; someday values might be non-atomic
							   ))
	    (let ((new-row
		   (loop for var in variables
		         collect
		         (variable-value var :where row-number))))
              
              (add-row-to-dataset new-dataset new-row new-variables t))))

    (values (mapcar #'second dataset-assoc-list))))

(defun partition-on-multiple-variables (old-dataset variables include-variables &optional suppress-warnings)
  (setf old-dataset (get-dataset old-dataset))
  (setf variables (mapcar #'(lambda (x) (get-variable x old-dataset))
			  variables))
  (if include-variables
      (setf include-variables (mapcar #'(lambda (x) (get-variable x old-dataset))
			      include-variables))
    (setf include-variables (dataset-variables old-dataset)))
  (let ((distinct-values (power-set (mapcar #'(lambda (var)
                                                (variable-value var :distinct t))
                                            variables)))
        (dataset-assoc-list nil))

    (when (and (not suppress-warnings)
               (> (length distinct-values) *more-partitions-than-you-really-want*))
      (cerror "Partition it anyway." 
              'more-partitions-than-you-really-want 
              :variable variables :number-of-partitions (length distinct-values)))

    ;; Build all the new datasets that we need and put them in an assoc table whose key is the distinct values.
    (loop for new-value-set in distinct-values
          do
          (multiple-value-bind (new-dataset new-variables)
              (build-new-dataset old-dataset include-variables
                                 (make-clasp-name (format nil "~a-where-~{~a~^,~}=~{~a~^,~}"
                                                          (name old-dataset)
                                                          (mapcar #'name variables)
                                                          new-value-set)
                                                  :dataset))
            (setf dataset-assoc-list (acons new-value-set (list new-dataset new-variables) dataset-assoc-list))))
    
    ;; Go through the original dataset by rows and place each row in the appropriate new dataset
    ;; We might want to consider mapping over these in reverse order so we can push the values on the new dataset.
    ;; Currently 'add-row-to-dataset' nconc's things on the end.
    (loop for row-number from (1- (dataset-rows old-dataset)) downto 0 do
	  (destructuring-bind (new-dataset new-variables) (rest
                                                           (assoc 
							    (mapcar #'(lambda (var)
                                                                        (variable-value var :where row-number))
                                                                    variables)
							    dataset-assoc-list
							    :test #'equalp))
	    (let ((new-row
		   (loop for var in include-variables
		         collect
		         (variable-value var :where row-number))))
              
              (add-row-to-dataset new-dataset new-row new-variables t))))

    (values (mapcar #'second dataset-assoc-list))))
    
(defun build-new-dataset (old-dataset variables new-dataset-name &optional variable-postfix where)
  (assert (or new-dataset-name where))
  (let ((new-dataset (make-dataset
                      :create-row-number-variable nil
                      ;; the name has already been made unique so we need to tell `make-dataset' that fact
		      :name (or new-dataset-name
				(format nil "~a where ~a"
					(name old-dataset) where))
		      :description (format nil "~a where ~a"
					   (description old-dataset) where)))
	  new-variables
	  #+OLD-VARS
	  (row 0))
    ;; Build new variables
    (dolist (old-variable variables)
      (push
       (make-instance *variable-class*
		      :name (if (row-number-p old-variable)
				(make-clasp-name 
				 (format nil "~a-ROW-NUMBER" (name old-dataset))
				 :dataset)
			      (make-clasp-name
			       (format nil "~a~@[-~a~]" (name old-variable)
				       variable-postfix)
			       :variable new-dataset))
		      ;; :id (id old-variable)
		      #+OLD-VARS
		      :attribute #+OLD-VARS (variable-attribute old-variable)
		      :dataset new-dataset
		      :source (name new-dataset))
       new-variables))

    ;(spy (dataset-variables new-dataset))
    ;; Add new variables to the dataset
    (setf (dataset-variables new-dataset) (setf new-variables (nreverse new-variables)))
    ;; Add a row number variable

    ;(spy new-variables)
    ;(spy (dataset-variables new-dataset))
        
    (push
     (make-variable :class *row-number-variable-class*
                    :name "ROW-NUMBER"
                    :unique-name nil
		    ;;:attribute (list (make-clasp-name
		    ;;(format nil "~a-ROW-NUMBER" (name new-dataset))))
		    ;;:number-or-symbol)
		    :dataset new-dataset
		    #+OLD-VARS :rtm-table #+OLD-VARS (dataset-write-data new-dataset))
     (dataset-variables new-dataset))
    ;(spy (dataset-variables new-dataset))
    (values new-dataset new-variables)))


  


#+OLD-VARS
(defun partition-dataset (old-dataset partition-clause &optional variables
				      &key (unique-variable-names-p t))

  (setf old-dataset (get-dataset old-dataset))
  (if variables
      (setf variables (mapcar #'(lambda (x) (get-variable x old-dataset))
			      variables))
    (setf variables (dataset-variables old-dataset)))
  (multiple-value-bind
   (partition-clause on-variables)
   (replace-on-clauses partition-clause old-dataset)
   (if on-variables
       (let ((counter 0)
	     (partition-call-symbol (gensym "PARTITION-CALL")))
	 (eval
	  (subst (build-partition-call
		  on-variables partition-clause
		  old-dataset unique-variable-names-p (incf counter))
		 partition-call-symbol
		 (build-partition-loop on-variables partition-call-symbol))))
     (make-dataset-from-rtm-boolean
      old-dataset partition-clause variables
      :new-dataset-name (infix partition-clause)))))


#+OLD-VARS
(defun build-partition-loop (on-variables partition-call-symbol)
  (if on-variables
      `(loop for ,(cdadr (car on-variables)) in
	     ',(variable-value (caar on-variables) :distinct t)
	     ,(if (cdr on-variables)
		  'do
		'collect)
	     ,(build-partition-loop (cdr on-variables)
				    partition-call-symbol))
    partition-call-symbol))

#+OLD-VARS
(defun build-partition-call (on-variables partition-clause
                             old-dataset unique-variable-names-p counter)
  (let ((sublist (mapcar #'second on-variables))
	(partition-gensym (gensym "PARTITION-CLAUSE")))
    (setf sublist (cons 'list (loop for (sym . val) in sublist
				    collect `(list* (quote ,sym) ,val))))
    `(let ((,partition-gensym (sublis ,sublist ',partition-clause)))
       (make-dataset-from-rtm-boolean
	,old-dataset ,partition-gensym ',(dataset-variables old-dataset)
	:new-dataset-name (infix ,partition-gensym)
	:variable-postfix (and ,unique-variable-names-p ,counter)))))

#+OLD-VARS
(defun replace-on-clauses (partition-clause dataset)
  (let ((operator (car partition-clause))
	new-partition-clause
	on-variables)
    (cond
     ((member operator *rtm-logical-connectives*)
      (dolist (clause (cdr partition-clause))
	(multiple-value-bind (new-clause new-on-variables)
	   (replace-on-clauses clause dataset)
	  (push new-clause new-partition-clause)
	  (setf on-variables (append new-on-variables on-variables))))
      (values (cons operator (reverse new-partition-clause))
	      (remove-duplicates on-variables :key #'second)))
     ((eq operator '.on.)
      (if (> (length partition-clause) 2)
	  (values
	   `(.and.
	     ,@(mapcar #'(lambda (var)
			   (let ((sublist-entry
				  (get-sublist-entry var dataset)))
			     (push sublist-entry on-variables)
			     `(.==. ,(name (car sublist-entry))
				    ,(second sublist-entry))))
		       (cdr partition-clause)))
	   (remove-duplicates on-variables :key #'car))
	(let ((sublist-entry (get-sublist-entry (second partition-clause) dataset)))
	  (push sublist-entry on-variables)
	  (values
	   `(.==. ,(name (first sublist-entry)) ,(car (second sublist-entry)))
	   (remove-duplicates on-variables :key #'car))))) 
     (t (values partition-clause on-variables)))))

#+OLD-VARS
(defun get-sublist-entry (variable dataset)
  (setf variable (get-variable variable dataset))
  (list variable
	(cons
	 (gensym (format nil "~a-SYMBOL" (symbol-name (id variable))))
	 (gensym (format nil "~a-VALUE" (symbol-name (id variable)))))))

  
#+OLD-VARS
(defconstant *where-clause-infix-translators*
    '((.and. . and) (.or. . or) (.not. . not)
      (.==. . =) (.<=. . <=) (.>=. . >=) (./=. . /=) (.>>. . >) (.<<. . <)
      (.=><=. => <=) (.=><<. => <) (.>><=. > <=) (.>><<. > <)
      (.<==>. <= =>) (.<=>>. <= >) (.<<=>. < =>) (.<<>>. < >)
      (.min. . min) (.max. . max)
      (.pred. . pred) (.succ. . succ) (.floor. . floor) (.ceiling. . ceiling)))

#+OLD-VARS
(defun infix (where-clause)
  (cond
   ((member (car where-clause) *rtm-logical-connectives*)
    (format nil "~a~{-~a~}"
	    (cdr (assoc (first where-clause) *where-clause-infix-translators*))
	    (mapcar #'infix (cdr where-clause))))
   ((member (car where-clause) *rtm-two-place-comparison-operators*)
    (format nil "~a-~a-~a"
	    (second where-clause)
	    (cdr (assoc (first where-clause) *where-clause-infix-translators*))
	    (third where-clause)))
   ((member (car where-clause) *rtm-three-place-comparison-operators*)
    (format nil "~a-~a-~a-~a-~a"
	    (third where-clause)
	    (car (cdr (assoc (first where-clause)
			     *where-clause-infix-translators*)))
	    (second where-clause)
	    (cadr (cdr (assoc (first where-clause)
			      *where-clause-infix-translators*)))
	    (fourth where-clause)))
   ((member (car where-clause) *rtm-one-place-comparison-operators*)
    (format nil "~a-~a"
	    (cdr (assoc (first where-clause) *where-clause-infix-translators*))
	    (second where-clause)))))

(defun infix (where-clause stream)
  (cond ((atom where-clause)
         (write-char #\. stream)
         (princ where-clause stream))
        ((atom (first where-clause))
         (map nil #'(lambda (clause)
	              (infix clause stream)
                      (unless (eq clause (first (last where-clause)))
		        (write-char #\. stream)
                        (princ (first where-clause) stream)))
              (rest where-clause)))
        (t nil)))

(defun infix-string (where-clause)
  (with-output-to-string (stream)
    (infix where-clause stream)))
         
#+OLD-VARS
(defun make-dataset-from-rtm-boolean (old-dataset where &optional variables
				      &key new-dataset-name variable-postfix)
  (let ((new-dataset (make-dataset
		      :name (or new-dataset-name
				 (format nil "~a where ~a"
					 (name old-dataset) where))
		      :description (format nil "~a where ~a"
					   (description old-dataset) where)))
	new-variable
	(row 0))
    (dolist (new-variable (get-variable-list where
					 (dataset-variables old-dataset)))
      (unless (rtm-index-p (dataset-write-data old-dataset)
			   (car (variable-attribute new-variable)))
	(create-index (dataset-write-data old-dataset)
		      (car (variable-attribute new-variable)))))
    (setf (dataset-read-data new-dataset) (intern (make-clasp-name "VIEW" :data)))
    (setf (dataset-write-data new-dataset) (dataset-write-data old-dataset))
    (if (dataset-where-clause old-dataset)
	(setf (dataset-where-clause new-dataset)
	  (list '.and.
		(dataset-where-clause old-dataset)
		(process-where-clause where old-dataset)))
      (setf (dataset-where-clause new-dataset)
	(process-where-clause where old-dataset)))
    (create-view (dataset-read-data new-dataset)
		 :from (dataset-write-data new-dataset)
		 :where (dataset-where-clause new-dataset))
    (dolist (old-variable variables)
      (unless (row-number-p old-variable)
	(setf new-variable
	  (make-instance *variable-class*
	    :name (make-clasp-name
		    (format nil "~a~@[-~a~]" (name old-variable)
			    variable-postfix)
		    :variable new-dataset)
	    ;; :id (id old-variable)
	    :attribute (variable-attribute old-variable)
	    :dataset new-dataset
	    :source (name new-dataset)))
	(push-end new-variable (dataset-variables new-dataset))))
    (setf new-variable
      (make-variable :name "ROW-NUMBER"
		     ;;:attribute (list (make-clasp-name
		     ;;(format nil "~a-ROW-NUMBER" (name new-dataset))))
		     ;;:number-or-symbol)
		     :dataset new-dataset
		     :rtm-table (dataset-write-data new-dataset)))
    (push new-variable (dataset-variables new-dataset))
    (do-selection  (:from (dataset-write-data new-dataset)
			  :where (dataset-where-clause new-dataset)
			  :order-by 'row-number)
      (incf row)
      (setf (attr-value (car (variable-attribute new-variable))) row))
    (setf (dataset-rows new-dataset) row)
    new-dataset))

;;; ***************************************************************************
;;; EOF
