;;;; -*- Mode:Common-Lisp; Package:CLASP; Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/clasp/clasp/dev/sequence-fns.lisp *-*
;;;; *-* Last-edit: Monday, July 19, 1993  14:32:16; Edited-By: carlson *-* 
;;;; *-* Machine: Lester (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                           Sequence Functions                           *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Adam Carlson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; Copyright (c) 1990 1991 1992 University of Massachusetts
;;; Department of Computer Science
;;; Created by the Experimental Knowledge Systems Laboratory
;;;    under the direction of Professor Paul R. Cohen.
;;; For more information write to Dhart@cs.umass.edu (Internet)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  07-08-93 File Created.  (carlson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :CLASP)

;;; --*--
;;; ***************************************************************************

;;; ***************************************************************************
;;; Building sequences of numbers
;;; ***************************************************************************

(defun integer-sequence (&optional (start 0) (end 0) &key (by 1))
  "Returns a sequence of integers from `start' to `end', steping by `by'"
  (if (> end start)
      (loop for counter from start to end by by
	    collect counter)
      (loop for counter from start downto end by by
	    collect counter)))

(defun real-sequence (&optional start end
		      &key number-of-points (interval 1))
  "Returns a sequence of real numbers from `start' to `end'.  If
`number-of-points' is present, that many points will be evenly spaced in the
range, otherwise, if `interval' is present, it will be used as the interval to
step by, defaulting to 1 (or -1)."
  (when number-of-points
    (setf interval (/ (- (max end start) (min end start)) number-of-points)))
  (if (> end start)
      (loop for counter from start to end by interval
	    collect counter)
      (loop for counter from start downto end by interval
	    collect counter)))

;;; ***************************************************************************
;;; Sequence iteration
;;; ***************************************************************************

(defmacro do-sequence ((element-variable sequence
					 &rest args &key index filter predicate
					 &allow-other-keys)
		       &body body)
  
  "Iterates through the elements of `sequence', binding them to
`element-variable'.

If `index' is present, it determines the order that the elements of sequence
will be iterated over.  `index' is list of integers which give the positions in
`sequence' which will be considered for iteration.  If an element of `index' is
a list of two numbers, they delineate a range of values to iterate over.

`filter' must be a sequence of the same length as `sequence'.  Any elements of
`sequence' in the same position as a value of nil in `filter' will be skipped.

`predicate' is a function and anything in the lambda list which is not `index'
`filter' or `predicate' wil be used as arguments to `predicate'.  Arguments to
`predicate' must either be atoms or sequences of the same length as `sequence'.
For each position in `sequence', the corresponding element of each of
`predicate's arguments are referenced, and `predicate' is evaluated for those
values.  If the value of `predicate' is nil, that element of `sequence'will be
skipped.

For each element of `sequence' which is selected by `index' and passes through
`filter' and `predicate', `element-variable' is bound to that element, and
`body' is evaluated.  The return-value of do-sequence is the return value of
the last evaluation of `body'."
  (with-variables (n i j k element position return-value)
    (remf args :index)
    (remf args :filter)
    (remf args :predicate)
    `(let (,element-variable
	   ,return-value)
       (flet ((execute (,position)
		(setf ,element-variable (elt ,sequence ,position))
		,@body))
	 (if (null ,index)
	     (dotimes (,element (length ,sequence) ,return-value)
	       (when (pass-filter-and-predicate-p
		       ,element ,filter ,predicate ,@args)
		 (setf ,return-value (execute ,element))))
	     (let (,element)
	       (dotimes (,n (length ,index) ,return-value)
		 (setf ,element (elt ,index ,n))
		 (if (listp ,element)
		     (let ((,i (or (first ,element) 0))
			   (,j (or (second ,element) 0))
			   (,k (or (third ,element) 1)))
		       (if (> ,j ,i)
			   (loop for ,element from ,i to (1- ,j) by ,k
				 do
				 (when (pass-filter-and-predicate-p
					 ,element ,filter ,predicate ,@args)
				   (setf ,return-value (execute ,element))))
			   (loop for ,element from (1- ,i) downto ,j by ,k
				 do
				 (when (pass-filter-and-predicate-p
					 ,element ,filter ,predicate ,@args)
				   (setf ,return-value (execute ,element))))))
		     #+ignore
		     (let (,i ,j)
		       (setf ,i (apply #'min ,element)
			     ,j (apply #'max ,element))
		       (dotimes (,element (- ,j ,i) ,return-value)
			 (when (pass-filter-and-predicate-p
				 (+ ,element ,i) ,filter ,predicate ,@args)
			   (setf ,return-value (execute (+ ,element ,i))))))
		     (when (pass-filter-and-predicate-p
			     ,element ,filter ,predicate ,@args)
		       (setf ,return-value (execute ,element)))))))))))

(defun pass-filter-and-predicate-p
    (position filter predicate &rest predicate-args)
  (cond
   ((and filter predicate)
    (and (pass-filter-p position filter)
	 (apply #'pass-predicate-p predicate predicate-args)))
   (filter
    (pass-filter-p position filter))
   (predicate
    (apply #'pass-predicate-p predicate predicate-args))
   (t t)))

(defun pass-filter-p (position filter)
  (elt filter position))

(defun pass-predicate-p (position predicate &rest args)
  (apply predicate
	 (mapcar #'(lambda (arg)
		     (cond
                       ((typep arg 'sequence)
                        (elt arg position))
                       (t
                        arg)))
                 args)))

#|
(setf a '(1 2 3))
(do-sequence (element a)
  (print element))				; 3 lines are printed
1
2
3
=> 3

(do-sequence (element #(1 a 2 b 3 c) :index '(4 2 0))
  (print element))				; 3 lines are printed
3
2
1
=> 1

(do-sequence (element #(1 a 2 b 3 c) :index '(0 (2 5) 1))
  (print element))				; 5 lines are printed
1
2
B
3
A
=> A

(do-sequence (element '(1 a 2 b 3 c) :filter #(nil t nil t nil t))
  (print element))				; 3 lines are printed
A
B
C
=> C

(setf a '(1 a 2 b 3 c))
(do-sequence (element a :predicate #'numberp a)
  (print element))				; 3 lines are printed
1
2
3
=> 3

(setf a '(1 a 2 b 3 c))
(do-sequence (element a :index '(5 4 3 2 1 0) :filter '(nil nil nil t t t))
  (print element))				; 3 lines are printed
C
3
B
=> B

(setf a '(1 a 2 b 3 c))
(setf b '(a 7 6 5 4 b))
(do-sequence (element a
		      :index '(5 4 3 2 1 0)
		      :filter '(nil t nil t t t)
		      :predicate #'(lambda (x) (and (numberp x)
						    (> x 4)))
		      b)
  (print element))				; 2 lines are printed
B
A
=> A

(setf a '(4 2 1654 23 253 256 1 5))
(setf b '(5 4 0 2 3 1 6 7))
(do-sequence (a-elt a)
  (print a-elt))
(do-sequence (a-elt a :index b)
  (print a-elt))
(do-sequence (a-elt a :filter '(t t t nil nil nil t nil))
  (print a-elt))
(do-sequence (a-elt a :predicate #'> a b)
  (print a-elt))
(do-sequence (a-elt a :index (coerce b 'vector))
  (print a-elt))
(do-sequence (a-elt a :filter #(t t t nil nil nil t nil))
  (print a-elt))
(do-sequence (a-elt a :predicate #'> (coerce a 'vector) (coerce b 'vector))
  (print a-elt))
(do-sequence (a-elt a :predicate #'> a 5)
  (print a-elt))
|#

;;; ***************************************************************************
;;; Sequence accessors
;;; ***************************************************************************

(defun elts (sequence &rest args &key index filter predicate &allow-other-keys)
  "Returns a subsequence of `sequence'.

If `index' is present, it determines the order that the elements of sequence
will be returned.  `index' is list of integers which give the positions in
`sequence' which will be returned.  If an element of `index' is a list of two
numbers, they delineate a range of values from `sequence' to return.

`filter' must be a sequence of the same length as `sequence'.  Any elements of
`sequence' in the same position as a value of nil in `filter' will be skipped.

`predicate' is a function and anything in the lambda list which is not `index'
`filter' or `predicate' wil be used as arguments to `predicate'.  Arguments to
`predicate' must either be atoms or sequences of the same length as `sequence'.
For each position in `sequence', the corresponding element of each of
`predicate's arguments are referenced, and `predicate' is evaluated for those
values.  If the value of `predicate' is nil, that element of `sequence'will be
skipped.

`elts' returns a sequence of the same type as `sequence', consisting of
elements of `sequence' which are selected by `index' and pass through `filter'
and `predicate'."
  (remf args :index)
  (remf args :filter)
  (remf args :predicate)
  (if (or filter predicate index)
      (let ((new-vector (make-array '(0) :adjustable t :fill-pointer t))
	    element-position
            element)
	(if index
	    (dotimes (n (length index))
	      (setf element-position (elt index n))
	      (if (listp element-position)
		  (let ((i (or (first element-position) 0))
			(j (or (second element-position) 0))
			(k (or (third element-position) 1)))
		    (if (> j i)
			(loop for element-position from i to (1- j) by k
			      do
                              (setq element (elt sequence element-position))
			      (when (apply #'pass-filter-and-predicate-p
					   element-position filter predicate args)
				(vector-push-extend element new-vector)))
			(loop for element from (1- i) downto j by k
			      do
			      (when (apply #'pass-filter-and-predicate-p
				     element-position filter predicate args)
				(vector-push-extend element new-vector)))))
		(progn
                  (setq element (elt sequence element-position))
                  (when (apply #'pass-filter-and-predicate-p
                               element-position filter
			       predicate args)
		    (vector-push-extend element new-vector)))))
	    (dotimes (element-position (length sequence))
              (setq element (elt sequence element-position))
	      (when (apply #'pass-filter-and-predicate-p
			   element-position filter predicate args)
		(vector-push-extend element new-vector))))
	#+explorer
	(if (listp sequence)
	    (coerce new-vector 'list)
	    (coerce new-vector (type-of sequence)))
	#-explorer
	(sequence-coerce new-vector (type-of sequence)))
      sequence))


(defun sequence-coerce (sequence sequence-type)
  (cond
   ((symbolp sequence-type)
    (if (eq sequence-type 'cons)
        (coerce sequence 'list)
      (coerce sequence sequence-type)))
   (t
    (coerce sequence (car sequence-type)))))

#|
(setf a '(1 2 3))
(elts a) ;=> (1 2 3)

(elts #(1 a 2 b 3 c) :index '(4 2 0)) ;=> #(3 2 1)

(elts #(1 a 2 b 3 c) :index '(0 (2 5) 1)) ;=> #(1 2 B 3 A)

(elts '(1 a 2 b 3 c) :filter #(nil t nil t nil t)) ;=> (A B C)

(setf a '(1 a 2 b 3 c))
(elts a :predicate #'numberp a) ;=> (1 2 3)

(setf a '(1 a 2 b 3 c))
(elts a :index '(5 4 3 2 1 0) :filter '(nil nil nil t t t)) ;=> (C 3 B)

(setf a '(1 a 2 b 3 c))
(setf b '(a 7 6 5 4 b))
(elts a :index '(5 4 3 2 1 0)
        :filter '(nil t nil t t t)
	:predicate #'(lambda (x) (and (numberp x)
				      (> x 4)))
	b) ;=> (B A)

(setf a '(4 2 1654 23 253 256 1 5))
(setf b '(5 4 0 2 3 1 6 7))
(elts a)
(elts a :index b)
(elts a :filter '(t t t nil nil nil t nil))
(elts a :predicate #'> a b)
(elts a :index (coerce b 'vector))
(elts a :filter #(t t t nil nil nil t nil))
(elts a :predicate #'> (coerce a 'vector) (coerce b 'vector))
(elts a :predicate #'> a 5)
|#

(defsetf elts (sequence &rest args
			&key index filter predicate
			#-lispworks &allow-other-keys) (new-values)
  
  "Destructively modifies elements of `sequence'.
  
If `index' is present, then `new-values' must be an atom or a sequence of the
same length as `index'.  In the former case, any position that is overwritten
is given the value of the atom, and in the latter, elements of `sequence' are
over written with successive elements of `new-values'.  If `index'
is :end, `filter' and `predicate' are ignored, and `new-values' are added to
the end of `sequence'.

`filter' must be a sequence of the same length as `sequence'.  Any elements of
`sequence' in the same position as a value of nil in `filter' will not be
changed.

`predicate' is a function and anything in the lambda list which is not `index'
`filter' or `predicate' wil be used as arguments to `predicate'.  Arguments to
`predicate' must either be atoms or sequences of the same length as `sequence'.
For each position in `sequence', the corresponding element of each of
`predicate's arguments are referenced, and `predicate' is evaluated for those
values.  If the value of `predicate' is nil, that element of `sequence' will
not be changed."
  `(setf-elts-helper ,new-values ,sequence ,@args))

(defun setf-elts-helper (new-values sequence
			 &rest args
			 &key index filter predicate &allow-other-keys)
  "Destructively modifies elements of `sequence' to the corresponding elements
of `new-values' for those elements which are selected by `index', `filter' and
`predicate'.  For a more detailed description, see do-sequence, elts and (setf
elts)."
  ;; remove recognized keywords from args, leaving only arguments to predicate
  (remf args :index)
  (remf args :filter)
  (remf args :predicate)
  (labels (;; Put value at the end of sequence
	   (extend (value)
	     (cond
	       ((typep sequence 'vector)
		(vector-push-extend value sequence))
	       ((typep sequence 'list)
		(push-end value sequence))
	       (t (error "Can't extend ~a" sequence))))
	   ;; change the positionth element of sequence to value
	   (change (value position)
	     (if (or (eq position :end)
		     (>= position (length sequence)))
		 (extend value)
		 (setf (elt sequence position) value)))
	   (vectorize (value position)
	     (if (typep value 'sequence) (elt value position) value)))
    (cond
      ((null index)
       ;; Map across sequence, checking the filter and predicate at each
       ;; element and changing elements which pass through
       (dotimes (element (length sequence))
	 (when (apply #'pass-filter-and-predicate-p element
		      filter predicate args)
	   (change (vectorize new-values element) element))))
      ((eq index :end)
       (if (typep new-values 'sequence)
	   (dotimes (n (length new-values))
	     (change (elt new-values n) :end))
	   (change new-values :end)))
      (t
       ;; There is an index
       (let (element (offset 0))
	 ;; For each element of index
	 (dotimes (n (length index))
	   (setf element (elt index n))
	   ;; If the element is a list
	   (if (listp element)
	       (let ((i (or (first element) 0))
		     (j (or (second element) 0))
		     (k (or (third element) 1)))
		 (if (> j i)
		     (loop for element from i to (1- j) by k
			   do
			   (when (apply #'pass-filter-and-predicate-p
					element filter predicate args)
			     (change (vectorize new-values (+ n offset))
						element))
			   (incf offset))
		     (loop for element from (1- i) downto j by k
			   do
			   (when (apply #'pass-filter-and-predicate-p
					element filter predicate args)
			     (change (vectorize new-values (+ n offset))
						element))
			   (incf offset))))
	       #+ignore
	       ;; Use the max and min of it to delineate a range
	       (let (i j)
		 (setf i (apply #'min element) j (apply #'max element))
		 ;; For each element in the range
		 (dotimes (range (- j i))
		   ;; Check the corresponding filter and predicate
		   (when (apply #'pass-filter-and-predicate-p (+ range i)
				filter predicate args)
		     ;; and if it passes, change the elt of sequence to the
		     ;; next value of new-values
		     (change (vectorize new-values (+ n range)) (+ range i)))))
	       ;; if the element of index is a number, check the filter and
	       ;; predicate and change the corresponding value of sequence
	       (when (apply #'pass-filter-and-predicate-p element
			    filter predicate args)
		 (change (vectorize new-values (+ n offset)) element)))))))
    sequence))

#|
(setf a '(1 2 3))
(setf (elts a) 'z) => (Z Z Z)
(setf (elts a) #(1 2 3 4)) => '(1 2 3)

(setf a #(1 a 2 b 3 c))
(setf (elts a :index '(4 2 0)) '(x y z)) => #(Z A Y B X C)

(setf a #(1 a 2 b 3 c))
(setf (elts a :index '(0 (2 5) 1)) '(0 1 2 3 4 5)) => #(0 5 1 2 3 C)

(setf a #(1 a 2 b 3 c))
(setf (elts a :filter #(nil t nil t nil t)) 'z) => (1 Z 2 Z 3 Z)

(setf a '(1 a 2 b 3 c))
(setf (elts a :predicate #'numberp a) 'z) => (Z A Z B Z C)

(setf a '(1 a 2 b 3 c))
(setf (elts a :index '(5 4 3 2 1 0) :filter '(nil nil nil t t t)) 'z) => (1 A 2 Z Z Z)

(setf a '(1 a 2 b 3 c))
(setf b '(a 7 6 5 4 b))
(setf (elts a :index '(5 4 3 2 1 0)
	    :filter '(nil t nil t t t)
	    :predicate #'(lambda (x) (and (numberp x)
					  (> x 4)))
	    b)
      '(h i j k l m)) => (1 L 2 J 3 C)
|#


;;; ***************************************************************************
;;; EOF
