;;;
;;; clasp-command.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines classes used in the MCL interface. In this interface the user
manipulates the following kinds of objects, via the specified actions:

clasp::dataset	- Save, rename, statistical operations
clasp::variable	- Rename, statistical operations, delete
clasp-command	- Re-run?, log, open (display LISP). (Generated by issuing a command.)
graph::graph	- Open (display graphic), log, delete


In general most of these can be a "result" of issuing a clasp command,
and thus would be displayed in the history-view. In addition, any LISP
object (e.g. typep T) can be a result as well since results are
literally the result of issuing a command. Any result can be logged;
specialize the log-item method to get correct behavior.


================================================================
Status =========================================================
================================================================
In-progress.

FIX: It will be much cleaner to use clasp-commands by simply creating
the command with its sexp then calling run-clasp-command, which would
check for a clasp-window, wrap with-clasp-status around evaluating the
sexp, save the l-results in the command, put the command and results
into the history (as add-command&result-cw does), and optionally update
dataset-view.



================================================================
Change history =================================================
================================================================
27-May-94 mc	Created.
19-Jul-94 ms    made clasp-commands openable
                added result-functions for more informative output
15-Sep-94 ms    *most-recent-vresult* buffers most rec. variable

|#


(in-package "CLASP-INTERFACE")

(export '(with-clasp-status
          clasp-command
          sexp-clasp-command
          l-result-clasp-command
          make-clasp-command
          run-clasp-command
          *most-recent-vresult*))

(defvar *most-recent-vresult* '()
  "the most recent variable result.")

;;;
;;; Define support functions and macros.
;;;


;;; FIX: BUG: with-clasp-status sometimes doesn't update the textual status.

(defmacro with-clasp-status (str-message &body body)
  "Wraps a status message around evaluation of body, which is wrapped
by progn. Returns the values from body's last form."
  ;;
  `(with-cursor *watch-cursor*
     (let ((clasp-window (clasp-window-front))
           l-return-values)
       (when clasp-window
         (show-status-view-string clasp-window ,str-message))
       ;;
       (setf l-return-values
             (multiple-value-list (progn ,@body)))
       ;;
       (when clasp-window
         (show-status-view-string clasp-window
                                  (concatenate 'string ,str-message " Done")))
       (apply #'values l-return-values))))

;;; CLASP ERROR TRAPPING

(defvar *use-native-debugger* nil)
(defun get-nasty () (setq *use-native-debugger* t))

(defmacro handle-errors (standard-message &body body)
  `(catch 'recover
     (if *use-native-debugger*
       (progn . ,body)
       ;; Otherwise set up our nice error handlers...
       (handler-bind ((clasp::more-partitions-than-you-really-want
                       #'(lambda (condition)
                           (if (y-or-n-dialog "You are about to get lots of partitions. Keep going ?" :yes-text "of course"
                                              :no-text "no way" :cancel-text nil)
                             (continue condition)
                             (throw 'recover nil))))
                      (clasp::enormous-contingency-table
                       #'(lambda (condition)
                           (if (y-or-n-dialog "CLASP is about to generate a huge contingency table. Proceed ?" :yes-text "yes"
                                              :no-text "no way" :cancel-text nil)
                             (continue condition)
                             (throw 'recover nil))))
                      (plotter::plotter-error
                       #'(lambda (condition)
                           (gui-warn (format nil "Plotter exception ~S" (type-of condition)))
                           (throw 'recover nil)))
                      (clasp::clasp-error #'(lambda (condition)
                                              (gui-error condition "CLASP: ")
                                              (throw 'recover NIL)))
                      (error #'(lambda (condition)
                                 (gui-error condition "ERROR: " ,standard-message)
                                 (throw 'recover NIL)))
                      (warning #'(lambda (condition) (gui-error condition "WARNING: ")
                                  (muffle-warning condition))))
         ,@body))))

;;;
;;; Define clasp-command and related functions.
;;;

(defclass clasp-command ()
  ((sexp
    :initform ()
    :initarg :sexp
    :type T
    :accessor sexp-clasp-command
    :documentation "The s-expression that was evaluated when the
user chose the command.")
   (standard-message
    :initform ()
    :initarg :standard-message
    :accessor standard-message
    :documentation "A generic message to print if the command fails.")
   (sexp-describe
    :initform ()
    :initarg :sexp-describe
    :type LIST
    :accessor sexp-describe-clasp-command
    :documentation "A description of the command and its args. Use the format
((tag item1 item2..) (tag item1 item2...)...)")
   ;; FIX: Should we have a description for both the sexp and the result?
   (l-result
    :initform nil
    :initarg :l-result
    :type LIST
    :accessor l-result-clasp-command
    :documentation "The value returned by the command.")
   (format-result-function
    :initform nil
    :initarg :format-function
    :type function
    :accessor format-function
    :documentation "A function which will prepare the output of a clasp
function for viewing in the history view.")
   )
  (:documentation "Created when the user issues a clasp command.
Represents the s-expression that is funcalled to make the command
happen."))


(defmethod print-object ((object clasp-command) stream)
  "Prints object's sexp-clasp-command and result."
  ;;
  (print-unreadable-object (object stream :type t :identity t)
    (when (slot-boundp object 'sexp)
      (format stream " ~S" (sexp-clasp-command object)))
    (when (slot-boundp object 'l-result)
      (format stream " ~S" (l-result-clasp-command object)))))


(defmethod make-clasp-command ((sexp t) &key (format-function nil)
                               (description nil) (standard-error NIL))
  "The top-level function for creating clasp-commands. Returns a a
clasp-command instance initialized by sexp . Note: This does
not evaluate the command; use run-clasp-command to do that."
  ;;
  (make-instance 'clasp-command :sexp sexp :format-function format-function
                 :sexp-describe description :standard-message standard-error))

(defmethod prepare-results ((clasp-command clasp-command))
  "make the results list of a clasp command more informative using
the format-function slot."
  (with-slots (l-result format-result-function) clasp-command
    (if (null format-result-function)
      l-result
      (funcall format-result-function clasp-command))))

(defmethod get-descript-tag ((clasp-command clasp-command)
                                (tag SYMBOL))
  (let ((list (sexp-describe-clasp-command clasp-command))
        (result nil))
    (if list
      (progn
        (loop for x in list do
              (if (equal (car x) tag)
                (setq result (append result (cdr x)))))
        result))))

(defmethod f-can-open-clasp-object ((clasp-command clasp-command))
  t)
(defmethod open-clasp-object ((clasp-command clasp-command))
  (run-clasp-command clasp-command))

(defmethod open-clasp-object :around ((object t))
  (handle-errors NIL (call-next-method)))

;;; FIX: Default str-message to "prettified" (first sexp) :

(defvar *remove-derelict-datasets* t
  "Someitmes datasets are intermediate results and get leftover after an error.")

(defmethod run-clasp-command ((clasp-command clasp-command)
                               &key (str-message "Computing clasp command")
                                    (f-update-dataset-view nil))
  "The top-level function for running a clasp-command. Clasp-command must
have its sexp-clasp-command set to an expression that, when evaluated,
computes/does the function. Eval results are saved via
multiple-value-list in clasp-command's l-result-clasp-command . If
there's a clasp-window-front then str-message is passed to
with-clasp-status before eval, and clasp-command is passed to
add-command&result-hv after eval. And, if f-update-dataset-view is
non-nil, then update-dataset-view is called."
  ;;
  (let* ((clasp-window (clasp-window-front))
         (sexp (sexp-clasp-command clasp-command))
         (dset-image clasp::*clasp-datasets*)
         (l-values-result (with-clasp-status str-message
                            ;; FIX: Eval-enqueue !?
                            (handle-errors (standard-message clasp-command)
                                           (multiple-value-list (eval sexp))))))
    #+DEMO
    (when (or (< (slot-value cds-record 'aenfastt) (get-universal-time))
              (< *weird-time-interval* (- (slot-value cds-record 'aenfastt) (get-universal-time))))
      (demo-times-up)
      (return-from run-clasp-command))
    (setf (l-result-clasp-command clasp-command) l-values-result)
    ;; 15-Sep buffer most recent result
    (if (typep (car (l-result-clasp-command clasp-command)) 'clasp::variable)
      (setf *most-recent-vresult* (ignore-errors (car (l-result-clasp-command clasp-command)))))
    (when clasp-window
      (if (car (l-result-clasp-command clasp-command))
        ;; output results
        (progn
          (setf (l-result-clasp-command clasp-command) (prepare-results clasp-command))
          (add-command&result-hv clasp-command
                                 (view-named 'history-view clasp-window)))
        ;; no result -- maybe an error, check for derelict datasets
        (if (and *remove-derelict-datasets* (not (equalp dset-image clasp::*clasp-datasets*)))
          (loop for x in clasp::*clasp-datasets* do
                (if (not (member x dset-image))
                  (clasp::delete-dataset x)))))
      (when f-update-dataset-view
        (update-dataset-view clasp-window)))))


;;;
;;; Define save/restore-related functions. (Don't you wish you had
;;;  persistent CLOS !?)
;;;

(defgeneric l-recreate-form-object (object)
  (:documentation "Returns a form that, when evaluated, returns an
instance equivalent to object. Subclasses should add their slot value
specializations to the usual via :around methods."))


(defmethod l-recreate-form-object ((object standard-object))
  "The root method for standard-objects. Creates the basic
make-instance form with no slot initializations."
  ;;
  `(make-instance ',(type-of object)))


(defmethod l-recreate-form-object ((object list))
  "The root method for lists. Recurses on elements."
  ;;
  (loop for item in object
        collect (l-recreate-form-object item)))


(defmethod l-recreate-form-object ((object t))
  "The root method for all other types of objects. Returns object ."
  ;;
  object)


(defmethod l-recreate-form-object ((object clasp::dataset))
  "The root method for datasets. Returns an expression getting the name."
  ;;
  ;; FIX: Is read-from-string right/needed?
  ;;
  `(clasp::get-dataset ',(read-from-string (clasp::name object))))


(defmethod l-recreate-form-object ((object clasp::variable))
  "The root method for variables. Returns an expression getting the name."
  ;;
  ;; FIX: Is read-from-string right/needed?
  ;;
  `(clasp::get-variable ',(read-from-string (clasp::name object))))


(defmethod l-recreate-form-object :around ((clasp-command clasp-command))
  "Adds clasp-command's slot values to the usual. Slots include just
sexp-clasp-command, since l-result-clasp-command is added when the
command is run."
  ;;
  (let ((l-recreate-form-parent (call-next-method))
        (l-recreate-form-sexp (l-recreate-form-object
                               (sexp-clasp-command clasp-command))))
    (append l-recreate-form-parent
            ;;
            ;; FIX: Recurse on elements:
            ;;  list      -> recurse on elements
            ;;  variables -> (clasp::get-variable ,(clasp::name variable))
            ;;  datasets  -> (clasp::get-dataset ,(clasp::name dataset))
            ;;  others    -> themselves
            `(:sexp ',l-recreate-form-sexp))))

;;

(defun gui-error (condition &optional (prefix "") (standard-message nil))
  (with-cursor *arrow-cursor*
    (gui-warn (format nil "~A~A" prefix 
                      (if standard-message 
                        standard-message
                        condition)))))

