;;;
;;; clasp-window.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines the clasp window frame in which users interact.


================================================================
Status =========================================================
================================================================
In-progress.


================================================================
Change history =================================================
================================================================
22-May-94 mc	Created.
6-Jul-94  ms    added f-some-variables-selected-cw as a more general method
20-Jul-94 ms    Beta finish.

|#


(in-package "CLASP-INTERFACE")

(export '(clasp-window
          clasp-window-front
          show-status-view-string
          clear-status-view
          update-dataset-view
          clear-history-view
          l-data-selected-cw
          l-history-selected-cw
          f-one-data-item-selected-cw
          f-one-dataset-selected-cw
          f-one-variable-selected-cw
          f-some-variables-selected-cw
          f-selection-openable-cw))


;;;

(defclass clasp-window (window)
  ()
  (:documentation "The top-level CLASP 'frame' in which users interact.
The window contains the following three subviews positioned as
described. The DATASET-VIEW is a scrolling vertical
sequence-dialog-item that fills the entire left third of the window
area, i.e. it's left, top, and bottom borders are 0, and its right
border is 1/3 of the window's width. It lists each CLASP dataset with
its variables listed and indented under it. Selection is :disjoint . The
HISTORY-VIEW is a scrolling vertical sequence-dialog-item that
fills most of the window's remainder, leaving room below it for a
status area. Its left border is flush against the DATASET-VIEW's
right border, its top border is 0, its right border is flush against
the window's right border, and its bottom is inline with the top of
the window's grow box. The history-view shows pairs of user
commands and command results, with a heavy line separating pairs.
Selection is :disjoint . Selected commands and results can be 'logged'
to the logging program for editing and reporting. The STATUS-VIEW
fills the space immediately below the history-view, and is a
general-purpose message area.

Abbreviated 'cw'.")
  (:DEFAULT-INITARGS
    :WINDOW-TYPE :DOCUMENT-WITH-ZOOM
    :window-title "CLASP"
    :VIEW-POSITION (make-point (- *screen-width* 560) 50)
    :VIEW-SIZE #@(550 300)
    #-TEST
    :view-subviews
    #-TEST
    (list (make-instance 'dataset-view :VIEW-NICK-NAME 'dataset-view)
          (make-instance 'history-view :VIEW-NICK-NAME 'history-view)
          (make-instance 'status-view :VIEW-NICK-NAME 'status-view)
          (make-instance 'comment-entry-item :view-nick-name 'comment-view
                         :destination 'history-view)
          (make-instance 'comment-activator :view-nick-name 'ca-view
                         :activate-who-tag 'comment-view))))


(defun clasp-window-front ()
  "Returns the frontmost clasp-window or nil if there are none."
  ;;
  (first (windows :class 'clasp-window)))

#-TEST
(defmethod window-size-parts :after ((clasp-window clasp-window))
  "Sets the sizes and positions of clasp-window's 'dataset-view,
'history-view, and 'status-view according to the constraints
identified in clasp-window's documentation."
  ;;
  ;; 4-28-95 Added this test of subviews. It looks like window size parts is getting
  ;; called earlier in the initialization process than it was in MCL 2.0.  (Westy)
  (unless (or (null (view-named 'dataset-view clasp-window))
              (null (view-named 'history-view clasp-window))
              (null (view-named 'status-view clasp-window))
              (null (view-named 'comment-view clasp-window))
              (null (view-named 'ca-view clasp-window)))
    ;(print (view-subviews clasp-window))
    (let* ((dataset-view (view-named 'dataset-view clasp-window))
           (history-view (view-named 'history-view clasp-window))
           (status-view (view-named 'status-view clasp-window))
           (comment-view (view-named 'comment-view clasp-window))
           (ca-view (view-named 'ca-view clasp-window))
           ;;
           (int-window-width (point-h (view-size clasp-window)))
           (int-window-height (point-v (view-size clasp-window)))
           (int-dv-right (round int-window-width 3))
           (int-grow-icon-left nil)       ;set in next binding (tricky)
           (int-grow-icon-top
            (multiple-value-bind (pt-top-left   #|pt-bottom-right|#  )
                                 (ccl::grow-icon-corners clasp-window)    ;FIX: undocumented
              (setf int-grow-icon-left (point-h pt-top-left))     ;tricky
              (point-v pt-top-left)))
           )
      #-TEST
      (progn
      ;;
      ;; DATASET-VIEW fills the entire left third of the window
      ;;  area, i.e. it's left, top, and bottom borders are 0, and its right
      ;;  border is 1/3 of the window's width.
      ;;
      (set-view-position dataset-view 0 0)
      (set-view-size dataset-view int-dv-right int-window-height)
      ;;
      ;; COMMENT-VIEW takes up one text line below history and above status
      ;;
      (set-view-size comment-view (- int-window-width int-dv-right
                                     (- int-window-width int-grow-icon-left)
                                     5)
                     (view-font-height comment-view))
      (set-view-position comment-view (+ int-dv-right 3)
                         (- int-grow-icon-top
                            (point-v (view-size comment-view))
                            2))
      ;; COMMENT-ACTIVATOR-VIEW goes right above the grow-icon
      (set-view-position ca-view (1+ int-grow-icon-left)
                         (point-v (view-position comment-view)))
      ;;
      ;; HISTORY-VIEW fills most of the window's remainder, leaving room below
      ;;  it for a status area. Its left border is flush against the DATASET-VIEW's
      ;;  right border, its top border is 0, its right border is flush against
      ;;  the window's right border, and its bottom is inline with the top of
      ;;  the window's grow box.
      ;;
      (set-view-position history-view (+ int-dv-right 1) 0)
      (set-view-size history-view (- int-window-width int-dv-right 1)
                     (- int-grow-icon-top
                        (point-v (view-size comment-view))
                        5))
      ;;
      ;; STATUS-VIEW fills the space immediately below the history-view
      ;;
      (set-view-position status-view (+ int-dv-right 1) (+ int-grow-icon-top 1))
      (set-view-size status-view (- int-grow-icon-left int-dv-right 1)
                     (- int-window-height int-grow-icon-top))))))


;;;
;;; Specialize methods for handling a clasp-window's status-view from
;;;  the window.
;;;

(defmethod show-status-view-string ((clasp-window clasp-window)
                                    (string string))
  ;  "Recurses on clasp-window's 'status-view ."
  ;;
  #-TEST
  (show-status-view-string (view-named 'status-view clasp-window)
                           string))


(defmethod clear-status-view ((clasp-window clasp-window))
;  "Recurses on clasp-window's 'status-view ."
  ;;
  (clear-status-view (view-named 'status-view clasp-window)))


;;;
;;; Specialize methods for handling a clasp-window's dataset-view from
;;;  the window.
;;;

(defmethod update-dataset-view ((clasp-window clasp-window))
;  "Recurses on clasp-window's 'dataset-view ."
  ;;
  (update-dataset-view (view-named 'dataset-view clasp-window)))


;;;
;;; Specialize methods for handling a clasp-window's history-view from
;;;  the window.
;;;

(defmethod clear-history-view ((clasp-window clasp-window))
;  "Recurses on clasp-window's 'history-view then clears clasp-window's
;status-view."
  ;;
  (clear-history-view (view-named 'history-view clasp-window))
  (clear-status-view clasp-window))


;;;
;;; Define selection-related methods.
;;;

(defmethod l-data-selected-cw ((clasp-window clasp-window))
  ;"Returns a list of clasp-window's 'dataset-view's l-selected-items-dv ."
  ;;
  #-TEST
  (l-selected-items-dv (view-named 'dataset-view clasp-window)))


(defmethod l-history-selected-cw ((clasp-window clasp-window))
  ;"Returns a list of clasp-window's 'history-view's l-selected-items-hv ."
  ;;
  #-TEST
  (l-selected-items-hv (view-named 'history-view clasp-window)))


(defmethod f-one-data-item-selected-cw ((clasp-window clasp-window))
  ;"Returns the dataset when clasp-window's l-data-selected-cw
;contains exactly one data item, either a variable or a dataset. FIX:
;What other kinds of items? Otherwise returns nil."
  ;;
  (let* ((l-data-selected (l-data-selected-cw clasp-window))
         (data-item-first (first l-data-selected)))
    (when (= 1 (length l-data-selected))
      data-item-first)))


(defmethod f-one-dataset-selected-cw ((clasp-window clasp-window))
;  "Returns the dataset when clasp-window's l-data-selected-cw
;contains exactly one dataset. Otherwise returns nil."
  ;;
  (let* ((l-data-selected (l-data-selected-cw clasp-window))
         (data-item-first (first l-data-selected)))
    (when (and (= 1 (length l-data-selected))
               (typep data-item-first 'clasp::dataset))
      data-item-first)))

(defmethod f-one-variable-selected-cw ((clasp-window clasp-window))
;  "Returns the variable when clasp-window's l-data-selected-cw
;contains exactly one variable. Otherwise returns nil."
  ;;
  (let* ((l-data-selected (l-data-selected-cw clasp-window))
         (data-item-first (first l-data-selected)))
    (when (and (= 1 (length l-data-selected))
               (typep data-item-first 'clasp::variable))
      data-item-first)))


(defmethod f-two-variables-selected-cw ((clasp-window clasp-window))
;  "Returns the variables as multiple values when clasp-window's
;l-data-selected-cw contains exactly two variables. Otherwise returns
;nil."
  ;;
  (let* ((l-data-selected (l-data-selected-cw clasp-window))
         (data-item-first (first l-data-selected))
         (data-item-second (second l-data-selected)))
    (when (and (= 2 (length l-data-selected))
               (typep data-item-first 'clasp::variable)
               (typep data-item-second 'clasp::variable))
      (values data-item-first data-item-second))))

(defmethod f-some-variables-selected-cw ((clasp-window clasp-window))
;  "Returns the variables as a list of variables when clasp-window's
;l-data-selected-cw contains one or more variables. Otherwise returns nil."
  ;;
  (let* ((l-data-selected (l-data-selected-cw clasp-window))
         (result '()))
    (loop while (not (null l-data-selected)) do
          (if (typep (car l-data-selected) 'clasp::variable)
            (setq result (cons (car l-data-selected) result)))
          (setq l-data-selected (cdr l-data-selected)))
    result))

(defmethod f-some-numbers-selected-cw ((clasp-window clasp-window))
 ; "Returns all the numbers selected in the current clasp window or
;NIL if none are selected."
  (let* ((l-history-selected (l-history-selected-cw clasp-window))
         (result '()))
    (loop while (not (null l-history-selected)) do
          (if (typep (car l-history-selected) 'number)
            (setq result (cons (car l-history-selected) result)))
          (setq l-history-selected (cdr l-history-selected)))
    (reverse result)))

(defmethod f-selection-openable-cw ((clasp-window clasp-window))
;  "Returns non-nil if clasp-window's selected history or data items is
;openable via f-can-open-clasp-object . Returns nil if not or if there
;is no selection. If openable then the list of the selected items is
;returned."
  ;;
  ;; FIX: Problem: Some selections of a list of items is considered
  ;;  openable as a list (e.g. lists of variables). Others are
  ;;  openable as separate items (e.g. graphs and datasets).
  ;;
  (let* ((l-selection (when clasp-window
                        (append (l-history-selected-cw (clasp-window-front))
                                (l-data-selected-cw (clasp-window-front)))))
         (f-selection-openable (and l-selection
                                    (every #'f-can-open-clasp-object l-selection))))
    (if f-selection-openable
      l-selection
      nil)))


;;;
;;; Define clipboard-related functions.
;;;

(defmethod copy ((clasp-window clasp-window))
;  ""
  ;;
  (copy-history-as-pict-cw (clasp-window-front)))


;;;
;;; Define command file/command history io.
;;;

(defmethod save-history-cw ((clasp-window clasp-window)
                             (pn-save pathname))
;  "Saves clasp-window's history in readable form to pn-save via calls
;to l-recreate-form-object . The format of the saved file is a series
;of calls to run-clasp-command for each clasp-command in clasp-window,
;with "" passed to :str-message and nil passted to
;:f-update-dataset-view . The final form saved is a call to
;update-dataset-view to show the results. Use restore-history-cw on
;pn-save to restore the environment."
  ;;
  (let* ((l-clasp-commands
          (remove-if-not #'(lambda (item)
                             (typep item 'clasp-command))
                         (table-sequence (view-named 'history-view
                                                     clasp-window)))))
    (with-open-file (stream pn-save :direction :output
                                    :if-exists :supersede)
      (loop for clasp-command in l-clasp-commands
            for l-recreate-form = (l-recreate-form-object clasp-command)
            do
            (write `(run-clasp-command ,l-recreate-form
                                       :str-message ""
                                       :f-update-dataset-view nil)
                   :stream stream)
            (terpri stream)(terpri stream)(terpri stream)
            finally
            (write `(let ((clasp-window (clasp-window-front)))
                      (when clasp-window
                        (update-dataset-view clasp-window)))
                   :stream stream)))
    ;; Return the pathname instead of the possibly long form.
    pn-save))


(defmethod save-history-dialog-cw ((clasp-window clasp-window))
;  "Prompts the user for a pathname then calls save-history-cw on it."
  ;;
  (let ((pn-save (choose-new-file-dialog :prompt "Save history as")))
    (eval-enqueue
     `(with-clasp-status (format nil "Saving ~A's history" ,clasp-window)
        (save-history-cw ,clasp-window ,pn-save)))))


(defmethod restore-history-cw ((clasp-window clasp-window)
                               (pn-restore pathname))
;  "Restores clasp-window's history from pn-restore, which must be a
;file created via save-history-cw ."
  ;;
  (load pn-restore))


(defmethod restore-history-dialog-cw ((clasp-window clasp-window))
;  "Prompts the user for a pathname then calls restore-history-cw on it."
  ;;
  (let ((pn-restore (choose-file-dialog :mac-file-type :TEXT))
        (*load-verbose* t))
    (eval-enqueue
     `(with-clasp-status ,(format nil "Restoring history to ~A" clasp-window)
        (restore-history-cw ,clasp-window ,pn-restore)))))

(defmethod window-close :before ((clasp-window clasp-window))
;  "Saves the contents of the history-buffer before we close the window."
  #-TEST
  (setf *history-buffer-last-close* 
        (table-sequence (view-named 'history-view clasp-window))))
