;;;
;;; data-view.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines view subclasses for examining (and eventually editing) dataset
and variable contents.


================================================================
Status =========================================================
================================================================
In-progress.

BUG: Changing size causes data-view to not draw in clipped region.
FIX: Print variable headers?
ISSUE: Use a fixed cell width (perhaps calculated) and horiz
 scrolling, so as not to squeeze data together in set-view-size !? DEC: Yes!
ISSUE: Is a list of variables openable? Or just a dataset? Ideally a
 list of vars is openable; vars in the same dataset would be shown in
 the same window using data-view's l-var-viewed-data-view. DEC: For
 now datasets and variables are openable, and vars are opened in their
 own windows. <- FIX.
ISSUE: Print spreadsheet-like labels (variable names across top and
 index numbers across left).


================================================================
Change history =================================================
================================================================
17-Jun-94 mc	Created.

|#


(in-package "CLASP-INTERFACE")

;(require "ARRAY-DIALOG")

(defvar *default-edit-window-size* (min #@(100 200)
                                             (make-point *screen-width* *screen-height*)))

;;;
;;; Specify that datsets and variables are openable.
;;;

(defmethod f-can-open-clasp-object ((object clasp::dataset))
  (if (or (not (funcall *open-click-p*)) *vars-openable*)
    t
    NIL))

(defmethod open-clasp-object ((object clasp::dataset))
  (eval-enqueue `(edit-data ,object)))


(defmethod f-can-open-clasp-object ((object clasp::variable))
  (if (or (not (funcall *open-click-p*)) *vars-openable*)
    t
    NIL))

(defmethod open-clasp-object ((object clasp::variable))
  (eval-enqueue `(edit-data ,object)))

;;;
;;; Define utilities.
;;;

(defmethod array-from-l-var ((l-variables list))
  "Returns an array filled with the variables in l-variables . The
array looks like a table of variables where its first dimension is the
column number (or variable number) and its second dimension is the row
number (or value number). Errors if any of l-variables are not
variables or if they are of different lengths."
  ;;
  ;; Check the argument(s).
  ;;
  (unless (every #'(lambda (item)
                     (typep item 'clasp::variable))
                 l-variables)
    (error "Not every one of ~A was a variable." l-variables))
  ;;
  (let ((l-unique-lengths
          (remove-duplicates
           (map 'list #'(lambda (variable)
                          (length (clasp::simplify variable)))
                l-variables))))
    (unless (= 1 (length l-unique-lengths))
      (error "Not every one of ~A was of the same length. (~A ~
different lengths found.)"
             l-variables
             (length l-unique-lengths)))
    ;;
    ;; Compute and return the array.
    ;;
    (make-array (list (length l-variables)      ;first dimension is column/var
                      (first l-unique-lengths))         ;2nd dim is row/value
                :initial-contents (map 'list #'clasp::simplify l-variables))))


;;;
;;; Define data-view .
;;;

(defclass data-view (array-dialog-item)
  ((dataset
    :initform nil
    :initarg :dataset
    :type CLASP::DATASET
    :accessor dataset-data-view
    :documentation "The dataset being viewed.")
   (l-var-viewed
    :initform ()
    ;:initarg :l-var-viewed	;init'd indirectly by initialize-instance :before
    :type LIST
    :accessor l-var-viewed-data-view
    :documentation "Lists those variables contained in dataset that
are being viewed. If null then data-view sets this slot to all the
variables in dataset."))
  (:documentation "An array-dialog-item subclass for showing a
dataset's variable's contents. The :dataset argument is required, and
the array is computed from it.")
  (:default-initargs
    :view-font *data-view-font*
    :cell-size #@(50 12)                ;FIX: compute?
    :table-hscrollp t
    :table-vscrollp t))


(defmethod initialize-instance :before ((data-view data-view)
                                      &key dataset l-var-viewed)
  "Errors if dataset is not a clasp::dataset. Initizalizes
l-var-viewed-data-view to all dataset's variables if l-var-viewed is
null."
  ;;
  (unless (typep dataset 'clasp::dataset)
    (error "~S must be a 'clasp::dataset." dataset))
  ;;
  (setf (l-var-viewed-data-view data-view)
        ;; FIX: Check that l-var-viewed are in (clasp::dataset-variables dataset)
        (or l-var-viewed (clasp::dataset-variables dataset))))


(defmethod initialize-instance :after ((data-view data-view) &key)
  "Computes data-view's table-array based on l-var-viewed-data-view ."
  ;;
  (set-cell-size data-view 
                 (string-width (make-string 14 :initial-element #\w) *data-view-font*)
                 (font-height *data-view-font*))
  (set-table-array data-view
                   (array-from-l-var (l-var-viewed-data-view
                                      data-view))))

(defmethod data-from-data-view-cell ((data-view data-view)
                                     (pt-cell fixnum))
  "Returns information about the clasp data in pt-cell. Multiple
values returned for pt-cell are: contents, dataset, variable, and
int-variable-index . Int-variable-index is numbered like
ELT. Returns nil if pt-cell is out of data-view's range."
  ;;
  (let* ((contents (cell-contents data-view pt-cell))
         (dataset (dataset-data-view data-view))
         (variable (and pt-cell
                        (elt (l-var-viewed-data-view data-view)
                             (point-h pt-cell))))
         (int-variable-index (and pt-cell
                                  (point-v pt-cell))))
    (when contents
      (values contents dataset variable int-variable-index))))

#+OBSOLETE
(defmethod view-click-event-handler :after ((data-view data-view)
                                            (pt-where fixnum))
  "Prints the contents of the clicked cell's variable."
  (when (double-click-p)
    (let* ((pt-cell-clicked (point-to-cell data-view pt-where)))
      (when pt-cell-clicked
        (eval-enqueue `(edit-selected-cell-data-view ,data-view))))))

;;;
;;; Define data-view-window .
;;;

(defclass data-view-window (window)
  ()                                    ;no new slots
  (:documentation "A window subclass that contains one data-view that
fills the whole window. You must pass :dataset and optionally pass
:l-var-viewed .")
  (:DEFAULT-INITARGS
    :view-subviews (list (make-instance 'hack-text-item :view-nick-name 'entry-item :allow-tabs t
                                        :dialog-item-text "<enter data item here>" :view-font '("Geneva" 8))
                         #+NO (make-instance 'button-dialog-item :view-nick-name 'insert-item
                                        :dialog-item-text "insert" :view-font '("Geneva" 9)
                                        :dialog-item-action #'(lambda (item)
                                                                (insert-dvi (view-container item))))
                         (make-instance 'button-dialog-item :view-nick-name 'replace-item
                                        :dialog-item-text "R" :view-font '("Geneva" 9) :border-p NIL
                                        :dialog-item-action #'(lambda (item)
                                                                (replace-dvi (view-container item))))
                         #+NO (make-instance 'button-dialog-item :view-nick-name 'add-item
                                        :dialog-item-text "add" :view-font '("Geneva" 9)
                                        :dialog-item-action #'(lambda (item)
                                                                (add-dvi (view-container item)))))))


(defmethod initialize-instance :after ((data-view-window data-view-window)
                                     &key dataset l-var-viewed)
  "Adds to data-view-window a 'data-view initialized with dataset and
l-var-viewed . Also sizes the window."
  ;;
  (let* ((array-dialog-item (make-instance 'data-view :view-nick-name 'array-dialog-item
                               :dataset dataset
                               :l-var-viewed l-var-viewed))
        (variables (l-var-viewed-data-view array-dialog-item))
        (cell-size (cell-size array-dialog-item)))
    
    (add-subviews data-view-window
                  ;; add the column lable item
                  (make-instance 'sequence-dialog-item :view-nick-name 'column-labels
                                 :table-sequence (mapcar #'clasp::name variables) 
                                 :sequence-order :horizontal 
                                 :table-dimensions (make-point (length variables) 1)
                                 :view-font `(:BOLD ,@(rest *data-view-font*))
                                 :cell-size cell-size
                                 :table-hscrollp nil
                                 :table-vscrollp nil)
                  array-dialog-item))
  ;; PROB: Window's initialize-instance calls window-size-parts before
  ;;  the above add-subviews is called. But we can't add-subviews in a
  ;;  :before method because slot VIEW-SUBVIEWS is unbound. SOL: Call
  ;;  window-size-parts again.
  (window-size-parts data-view-window))


(defmethod window-size-parts :after ((data-view-window data-view-window))
  "Sets data-view-window's 'array-dialog-item's size and position."
  ;;
  (let ((array-dialog-item (view-named 'array-dialog-item data-view-window))
        (enter (view-named 'entry-item data-view-window))
        #+NO (insert (view-named 'insert-item data-view-window))
        #+NO (add (view-named 'add-item data-view-window))
        (replace (view-named 'replace-item data-view-window))
        (column-labels (view-named 'column-labels data-view-window))
        (ws (view-size data-view-window)))
    #+OLD
    (when array-dialog-item
      (set-view-position array-dialog-item 0 0)
      (set-view-size array-dialog-item (point-h ws) (- (point-v ws) 15)))
    (when array-dialog-item
      (set-view-position array-dialog-item 0 16)
      (set-view-size array-dialog-item (point-h ws) (- (point-v ws) 15 16)))
    (when column-labels
      (set-view-position column-labels 0 0)
      (set-view-size column-labels (point-h ws) 15))
    (when enter
      (set-view-position enter 0 (- (point-v ws) 12))
      (set-view-size enter (- (point-h ws) 18) 11))
    #+NO (when insert
      (set-view-position insert 0 (- (point-v ws) 13)))
    (when replace
      (set-view-position replace (- (point-h ws) 14) (- (point-v ws) 28)))
    #+NO (when add
      (set-view-position add (- (round (point-h ws) 2) 22) (- (point-v ws) 13)))))

(defmethod view-click-event-handler :after ((self data-view-window)
                                            (pt-where fixnum))
  (synchronize-column-labels self))

(defmethod synchronize-column-labels ((self data-view-window))
  ;; Synchronize the column-labels with the array
  (let ((array-position (scroll-position (view-named 'array-dialog-item self))))
    ;(spy (point-h array-position))
    (scroll-to-cell (view-named 'column-labels self) (point-h array-position) 0)))

(defmethod view-key-event-handler :around ((dvw data-view-window) char)
  (call-next-method)
  (when (eql char #\Return)
    (replace-dvi dvw)
    (handle-arrow dvw #\Tab)
    ;; 1-20-95 removed this - Westy
    ;(select-all dvw)
    )
  (synchronize-column-labels dvw))

(defmethod cell-visible-p ((self data-view) (pt-cell integer))
  (let* ((upper-left-cell (scroll-position self))
         (visible-cells (visible-dimensions self))
         (upx (point-h upper-left-cell))
         (upy (point-v upper-left-cell))
         (ptx (point-h pt-cell))
         (pty (point-v pt-cell))
         (lrx (1- (+ upx (point-h visible-cells))))
         (lry (1- (+ upy (point-v visible-cells)))))

    (and (<= upx ptx lrx) (<= upy pty lry))))
    
(defmethod handle-arrow ((dialog data-view-window) char)
  "do the nasty hack on the sequence-view according to the char."
  (let* ((item (view-named 'array-dialog-item dialog))
         (selected-cells (selected-cells item))
         (delta-v (if (equal char #\UpArrow) -1 (if (eql char #\DownArrow) 1 0)))
         (delta-h (if (eql char #\Tab) 1 0))
         (cols (array-dimension (table-array item) 0))
         )
    (mapcar #'(lambda (in) (cell-deselect item (point-h in) (point-v in)))
            selected-cells)
    (mapcar #'(lambda (in) 
                (let ((nh (mod (+ (point-h in) delta-h) cols)))
                       (cell-select item nh (+ (if (and (= delta-v 0) (= 0 nh)) 1 delta-v) (point-v in)))))
            selected-cells)
    (when (selected-cells item)
      (let ((selected-cell (first (selected-cells item))))
        (unless (cell-visible-p item selected-cell)
          (scroll-to-cell item selected-cell))))))

;;;
;;; Define edit-data .
;;;

(defmethod edit-data ((data clasp::data))
  "Displays data it an a new data-view-window instance."
  ;;
  ;;
  ;; FIX: Make a clasp-command.
  ;; FIX: Data should be either a list of variables
  ;;  or a dataset. If a list then their common dataset should be
  ;;  passed below. Also, must define f-can-open-clasp-object and
  ;;  open-clasp-object for lists of variables. For now, if data is a
  ;;  variable, show it alone in its own dataset window.
  ;;
  (with-cursor *watch-cursor*
    (let* ((dataset (if (typep data 'clasp::dataset)
                      data
                      (clasp::variable-dataset data)))
           (f-limit-view (typep data 'clasp::variable))
           (l-var-viewed (if f-limit-view (list data) ()))
           (size (make-point (min (* (string-width (make-string 20 :initial-element #\w) *data-view-font*)
                                     (if (typep data 'clasp::dataset) (length (dataset-variables data)) 1))
                                  *screen-width*)
                             (min (* (font-height *data-view-font*) 13)
                                  *screen-height*))))
      (if dataset
        (let ((view-window (find-if #'(lambda (w) (ignore-errors 
                                                   (and (eql dataset (dataset-data-view (view-named 'array-dialog-item w)))
                                                        (equal (if (atom data) (list data) data)
                                                               (l-var-viewed-data-view (view-named 'array-dialog-item w))))))
                                    (windows))))
          (if view-window
            (progn
              (window-select view-window)
              view-window) ;;return the view window for history view purposes
            (make-instance 'data-view-window
              :window-title (clasp::name data)
              :view-size size
              :view-position (best-spot (point-h size)
                                        (point-v size))
              :dataset dataset
              :l-var-viewed l-var-viewed)))
        (message-dialog (format nil "Couldn't find a dataset for ~A."
                                (clasp::name data)))))))

;;; do the actual editing

(defmethod replace-dvi ((dvw data-view-window))
  (let* ((data-view (view-named 'array-dialog-item dvw))
         (pt-cell-selected (first (selected-cells data-view)))
         (x (when pt-cell-selected (point-h pt-cell-selected)))
         (y (when pt-cell-selected (point-v pt-cell-selected)))
         (string (dialog-item-text (view-named 'entry-item dvw))))
    (when pt-cell-selected
      (multiple-value-bind (contents dataset variable index)
                           (data-from-data-view-cell data-view pt-cell-selected)
        (let* ((value-new (typecase contents
                            (number (read-from-string string))
                            (symbol (intern string))
                            (t (read-from-string string)))))
          (setf (aref (table-array data-view) x y) value-new)
          (setf (elt (clasp::variable-value variable) index) value-new)
          (invalidate-view data-view)
          )))))

;;; the andro-genesis mothership (code for entering datasets)
#+FINISH-THIS-MATTTTTTTT
(defclass create-dset-window (window)
  ((dataset :initform NIL :initarg dataset)
   (columns :initform 1 :initarg columns))
  (:default-initargs 
    :view-size #@(200 100)
    :window-type :document-with-grow
    :view-subviews (list (make-instance 'hack-text-item :view-nick-name 'entry-item)
                         (make-instance 'button-dialog-item :dialog-item-action
                                        #'(lambda (item) 
                                            )))))

