;;;
;;; function-dialog-utils.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines classes, functions, etc. for statistics drag-and-drop dialog
boxes.


================================================================
Status =========================================================
================================================================
In-progress.

BUG: drag-gray-region-lisp is changing the received view's pattern.


================================================================
Change history =================================================
================================================================
6-Jun-94 mc	Created.

|#


(in-package "CLASP-INTERFACE")

#+OLD
(require "POP-UP-MENU")
#+OLD
(require "OBJECT-DROPPER")

;;;
;;; Define variable-menu-item .
;;;

(defclass variable-menu-item (menu-item)
  ((variable
    :initform nil
    :initarg :variable
    :type CLASP::VARIABLE
    :accessor variable-vmi
    :documentation "The clasp variable represented by this menu-item."))
  (:documentation "A menu-item subclass that represents a clasp
variable. The title is set automatically based on the variable. Used
by variable-pop-up-menu . Abbreviated 'vmi' ."))


(defmethod initialize-instance :after ((variable-menu-item variable-menu-item)
                                     &key variable)
  "Errors if variable is not a variable. Sets variable-menu-item's title."
  ;;
  (check-type variable clasp::variable)
  ;; FIX: Is there ever not a name?:
  (set-menu-item-title variable-menu-item
                       (clasp::name (variable-vmi variable-menu-item))))


#|
;;;
;;; Define boolean-pop-up-menu .
;;;

(defclass boolean-pop-up-menu (pop-up-menu)
  ()
  (:documentation "A pop-up-menu subclass with two items, 'True' and
'False'. Abbreviated 'boolean-pum'."))


(defmethod initialize-instance :after ((boolean-pop-up-menu boolean-pop-up-menu)
                                     &key)
  "Adds the two menu items."
  ;;
  (add-menu-items boolean-pop-up-menu
                  (make-instance 'menu-item :menu-item-title "True")
                  (make-instance 'menu-item :menu-item-title "False")))


(defmethod f-menu-value-boolean-pum ((boolean-pop-up-menu boolean-pop-up-menu))
  "Returns non-nil if boolean-pop-up-menu's 'True' menu item is
selected, and nil if its 'False' item is selected."
  ;;
 (let* ((menu-item-selected (selected-item boolean-pop-up-menu))
        (str-pop-up-selected-item (menu-item-title menu-item-selected)))
   (cond ((string-equal str-pop-up-selected-item "True")
          t)
         ((string-equal str-pop-up-selected-item "False")
          nil)
         (t
          (error "~S was neither 'True' or 'False'." str-pop-up-selected-item)))))
|#

;;;
;;; Define variable-pop-up-menu .
;;;

(defclass variable-pop-up-menu (drag-aware-mixin pop-up-menu)
  ((allow-categorical-p :initform t :initarg :allow-categorical-p
                        :accessor allow-categorical-p))
  (:documentation "A pop-up-menu subclass whose items must be
variable-menu-item, and that supports drag-and-drop of lists of
variables. Abbreviated 'var-pum'."))


(defmethod delete-selected-var-pum ((variable-pop-up-menu variable-pop-up-menu))
  "Removes the selected menu-item from variable-pop-up-menu ."
  ;;
  (let ((menu-item-selected (selected-item variable-pop-up-menu)))
    (cond (menu-item-selected
           (remove-menu-items variable-pop-up-menu menu-item-selected)
           (invalidate-view variable-pop-up-menu t)     ;show the removed item
           )
          (t
           ;; Beep to show there's no selection. FIX: Should we show a dialog?
           (ed-beep)))))


(defmethod l-variables-var-pum ((variable-pop-up-menu variable-pop-up-menu))
  "Returns a list of variables corresponding to variable-menu-item in
variable-pop-up-menu, ordered from top to bottom."
  ;;
  (map 'list #'(lambda (variable-menu-item)
                 (variable-vmi variable-menu-item))
       (menu-items variable-pop-up-menu)))


;;;

(defmethod f-can-receive-drop ((view variable-pop-up-menu)
                               (object t)
                               (pt-global fixnum))
  "Return non-nil if object is a clasp variable or a sequence of clasp
variables."
  ;;
  (or (typep object 'clasp::variable)
      (and (typep object 'sequence)
           (every #'(lambda (item)
                      (typep item 'clasp::variable))
                  object))))


(defmethod invert-drop-highlight ((view variable-pop-up-menu)
                                 (object t)
                                 (pt-global integer))
  ;; Invert a gray frame just inside view. (Headache if it's thick!)
  (invert-inside-gray-frame view))


(defmethod receive-drop ((view variable-pop-up-menu)
                         (object clasp::variable)
                         (pt-global integer))
  "Object is a clasp::variable. Adds to view a 'variable-menu-item for
object."
  ;;
  ;; FIX: Make the last item (the one just dropped) selected/current.
  ;;
  (add-menu-items view (make-instance 'variable-menu-item
                         :variable object))
  (invalidate-view view t)              ;show the new item
  )


(defmethod receive-drop ((view variable-pop-up-menu)
                         (object list)
                         (pt-global integer))
  "Object is a list of clasp variables. Recurses on each element in object."
  ;;
  (let (cats)
    (when (not (allow-categorical-p view))
      (when (setf cats (remove-if-not
                        #'(lambda (v) (eql (slot-value v 'clasp::type) :categorical))
                        object))
        (gui-warn (format nil "Categorical variable~A not allowed as ~A in ~A."
                          (if (> (length cats) 1) "s" (format nil " ~A" (clasp::name (car cats))))
                          #-CCL-3 (dialog-item-text view)
                          #+CCL-3 (menu-title view)
                          (window-title (view-container view))))))
    (loop for variable in (reverse object)
          do (unless (and (eql (slot-value variable 'clasp::type) :categorical)
                          (not (allow-categorical-p view)))
               (receive-drop view variable pt-global)))))


;;;
;;; Define variable-delete-button-di .
;;;

(defclass variable-delete-button-di (button-dialog-item)
  ((var-pum
    :initform nil
    :initarg :var-pum
    :type VARIABLE-POP-UP-MENU
    :accessor var-pum-vdbdi
    :documentation "The variable-pop-up-menu instance that this button
deletes selected items from."))
  (:documentation "A button-dialog-item subclass that's associated
with a variable-pop-up-menu and allows deleting that pop-up's
currently selected variable. The default dialog-item-text is 'Delete'.
Abbreviated 'vdbdi'.")
  (:default-initargs
    :dialog-item-text "Delete"
    :dialog-item-action
    #'(lambda (item)
        (delete-selected-var-pum (var-pum-vdbdi item)))))


(defmethod initialize-instance :after ((variable-delete-button-di variable-delete-button-di)
                                     &key var-pum)
  "Errors if :var-pum is not a variable-pop-up-menu ."
  ;;
  (check-type var-pum variable-pop-up-menu))


;;;
;;; Define compute-button-di .
;;;

(defclass compute-button-di (button-dialog-item)
  ()
  (:documentation "A button-dialog-item subclass that is used by
statistic-dialog-mixin subclasses via its default action, which calls
compute-statistics-dialog on its container.")
  (:default-initargs
    :dialog-item-text "Compute"
    :default-button t
    :dialog-item-action
    #'(lambda (item)
        (compute-statistics-dialog (view-container item)))))

;;;
;;; Define statistic-dialog-mixin .
;;;

(defclass statistic-dialog-mixin ()
  ((help :initform "No help for this operation." :initarg :help
         :accessor help))
  (:documentation "A mixin that's used by the 'compute' and 'clone'
menu-items .")
  (:default-initargs
    :WINDOW-DO-FIRST-CLICK t))

(defmethod view-key-event-handler :around ((thing statistic-dialog-mixin) char)
  (if (and (char= char #\/)
           (command-key-p))
    (display-help thing))
  (call-next-method))

(defmethod initialize-instance :after ((statistic-dialog-mixin statistic-dialog-mixin)
                                     &key)
  "Sets up the statistical dialog by adding delete buttons to the pums
and grabbing valid data from the CLASP dock."
  ;;
  ;; CLASP dock init 24-Jul-94 ms
  (let* ((dock (find-clasp-dock))
         (xs   (if dock (reverse (items-in-dock 'x-var-dock :select-type :select-or-all))))
         (ys   (if dock (reverse (items-in-dock 'y-var-dock :select-type :select-or-all))))
         (zs   (if dock (reverse (items-in-dock 'gp-dock    :select-type :select-or-all))))
         (x-item (view-named 'x-var-item statistic-dialog-mixin))
         (y-item (view-named 'y-var-item statistic-dialog-mixin))
         (z-item (view-named 'z-var-item statistic-dialog-mixin)))
    (if (and xs x-item) (if (f-can-receive-drop x-item xs 0)
                          (receive-drop x-item xs 0)))
    (if (and ys y-item) (if (f-can-receive-drop y-item ys 0)
                          (receive-drop y-item ys 0)))
    (if (and zs z-item) (if (f-can-receive-drop z-item zs 0)
                          (receive-drop z-item zs 0))))
  (apply #'add-subviews statistic-dialog-mixin
         (loop for var-pum across (remove-if-not
                                   #'(lambda (item)
                                       (typep item 'variable-pop-up-menu))
                                   (view-subviews statistic-dialog-mixin))
               for pt-position = (view-position var-pum)
               collect
               (make-instance 'variable-delete-button-di
                 :view-position (add-points pt-position
                                            (make-point (+ 20 (point-h (view-size var-pum)))
                                                        2))
                 :view-size #@(62 16)
                 :var-pum var-pum)))
  ;; ms added 19-Jul-94
  (apply #'add-subviews statistic-dialog-mixin
         (loop for dataset-pum across (remove-if-not
                                       #'(lambda (item)
                                           (typep item 'dataset-pop-up-menu))
                                       (view-subviews statistic-dialog-mixin))
               for pt-position = (view-position dataset-pum)
               collect
               (make-instance 'dataset-delete-button-di
                 :view-position (add-points pt-position
                                            (make-point (+ 20 (point-h (view-size dataset-pum)))
                                                        2))
                 :view-size #@(62 16)
                 :dataset-pum dataset-pum))))


(defgeneric f-can-compute-statistics-dialog (statistic-dialog-mixin)
  (:documentation "Returns two values that indicate whether
statistic-dialog-mixin has the proper values to compute its function.
The first value (f-can-compute) is a boolean that is non-nil when
statistic-dialog-mixin CAN compute. The second value (str-why-cant-compute)
is a string, meaningful when the first value is nil, that describes
why the function can't be currently be computed."))

(defmethod f-can-compute-statistics-dialog ((statistic-dialog-mixin statistic-dialog-mixin))
  (error "No f-can-compute-statistics-dialog method is defined for ~A"
         statistic-dialog-mixin))


(defgeneric compute-statistics-dialog (statistic-dialog-mixin)
  (:documentation "Computes the statistical function associated with
statistic-dialog-mixin. Only called when
f-can-compute-statistics-dialog returns non-nil."))

(defmethod compute-statistics-dialog ((statistic-dialog-mixin statistic-dialog-mixin))
  (error "No compute-statistics-dialog method is defined for ~A"
         statistic-dialog-mixin))

(defmethod compute-statistics-dialog :around ((statistic-dialog-mixin statistic-dialog-mixin))
  (multiple-value-bind (f-can-compute str-why-cant-compute)
                       (f-can-compute-statistics-dialog statistic-dialog-mixin)
    (if f-can-compute
      (call-next-method)
      (message-dialog (format nil "Can't compute with current ~
information: ~A." str-why-cant-compute) :size #@(350 150)))))

#+IGNORE
(defgeneric clone-statistics-dialog (statistic-dialog-mixin)
  (:documentation "Clones the data in statistic-dialog-mixin and
applies it to another statistical function."))
#+IGNORE
(defmethod clone-statistics-dialog ((statistic-dialog-mixin statistic-dialog-mixin))
  (error "No clone-statistics-dialog method is defined for ~A"
         statistic-dialog-mixin))


;;;
;;; Define dataset-stdi .
;;;

(defclass dataset-stdi (drag-aware-mixin static-text-dialog-item)
  ((dataset-dset-stdi
    :initform nil
    :initarg :dataset-dset-stdi
    :type CLASP::DATASET
    :accessor dataset-dset-stdi
    :documentation "The most recently stored dataset in this item."))
  (:documentation "A static-text-dialog-item subclass that supports
drag-and-drop of lists of a clasp dataset. Abbreviated 'dset-stdi'.")
  (:default-initargs
    :dialog-item-text "<NONE>"))


;;; FIX: Define (setf dataset-dset-stdi) :before to error if new-value
;;;  isn't typep clasp::dataset .


(defmethod (setf dataset-dset-stdi) :after ((dataset-new clasp::dataset)
                                          (dataset-stdi dataset-stdi))
  "Updates dataset-stdi's dialog-item-text to be dataset-new's name or
'<NONE>' ."
  ;;
  (set-dialog-item-text dataset-stdi (if dataset-new
                                       (clasp::name dataset-new)
                                       "<NONE>")))


(defmethod view-draw-contents :after ((dataset-stdi dataset-stdi))
  "Draws a frame around dataset-stdi ."
  ;;
  ;; FIX: Frame dataset-stdi in its container.
  ;;
  )


;;;

(defmethod f-can-receive-drop ((view dataset-stdi)
                               (object t)
                               (pt-global fixnum))
  "Return non-nil if object is a clasp dataset or a sequence of clasp
datasets."
  ;;
  (or (typep object 'clasp::dataset)
      (and (typep object 'sequence)
           (every #'(lambda (item)
                      (typep item 'clasp::dataset))
                  object))))


(defmethod invert-drop-highlight ((view dataset-stdi)
                                 (object t)
                                 (pt-global integer))
  ;; Invert a gray frame just inside view. (Headache if it's thick!)
  (invert-inside-gray-frame view))


(defmethod receive-drop ((view dataset-stdi)
                         (object clasp::dataset)
                         (pt-global integer))
  "Object is a clasp::dataset . Sets view's dataset-dset-stdi to
object ."
  ;;
  (setf (dataset-dset-stdi view) object))


(defmethod receive-drop ((view dataset-stdi)
                         (object list)
                         (pt-global integer))
  "Object is a list of clasp datasets. Recurses on each element in object."
  ;;
  (loop for dataset in (reverse object)
        do (if (f-can-receive-drop view dataset pt-global)
             (receive-drop view dataset pt-global))))
