;;;
;;; graph-edit.lisp
;;;

(in-package "CLASP-INTERFACE")

#|

Code to generate dialog boxes for graphs and modify the graphs according to
their contents.

11-Aug-94  ms   Created

|#

;;; a support function for changing the graphs
(defun install-plot-options (plot options &key (newname nil))
  "takes a SINGLE plot sequence and installs the new options, leaving plot type intact."
  (declare (ignore newname))
  (let* ((header (car plot))
         (old-options (car (last header)))
         (points (cdr plot)))
    (mapcar #'(lambda (option) (if (keywordp option) (remf old-options option)))
            options)
    (cons (list (car header) (cadr header) (append old-options options))
          points)))

(defun plot-color (plot)
  (let* ((header (car plot))
         (options (caddr header))
         (color-at (search (list :color) options))
         (color (if color-at (car (subseq options (1+ color-at))))))
    (if color
      color
      0)))
   

;;; a menu-item class for common colors

(defvar *color-table* #+mcl(list (list "Yellow" *yellow-color*)
                                    (list "Gray" *gray-color*)
                                    (list "Tan" *tan-color*)
                                    (list "Orange" *orange-color*)
                                    (list "Black" *black-color*)
                                    (list "Purple" *purple-color*)
                                    (list "Pink" *pink-color*)
                                    (list "Brown" *brown-color*)
                                    (list "Blue" *blue-color*)
                                    (list "Red" *red-color*)
                                    (list "Green" *green-color*)
                                    ))
                                    

(defclass colors-pum (pop-up-menu)
  ()
  (:DOCUMENTATION "A pop up menu with several common choices for colors."))

(defmethod initialize-instance :after ((color-pum colors-pum) &key)
  "puts the colors into the pum"
  (add-menu-items color-pum (make-instance 'menu-item :menu-item-title "Custom"))
  (add-menu-items color-pum (make-instance 'menu-item :menu-item-title "-"
                                           :disabled t))
  (mapcar #'(lambda (item) (add-menu-items color-pum
                                           (make-instance 'menu-item 
                                             :menu-item-title (car item))))
             *color-table*))

(defmethod view-click-event-handler :after ((menu colors-pum) (pt-where fixnum))
  (set-rgb-items (view-container menu)))

;;; the graph editor

(defclass graph-editor (window)
  ((graph
    :initform nil
    :initarg :graph
    :type graph
    :accessor graph-editor-graph
    :documentation "The graph that is being operated on."))
  (:DOCUMENTATION "The window class of dialog boxes to edit graphs.")
  (:DEFAULT-INITARGS
    :window-type :document
    :window-show nil
    :VIEW-POSITION (best-spot 320 200)))

(defmethod initialize-instance :after ((graph-editor graph-editor) &key)
  ;; Get the subviews and the size which depend on the graph we are editing.
  (with-slots (graph) graph-editor
    ;; Compute the size first so we can use it to figure out where some of the subviews go.
    (set-view-size graph-editor (graph-edit-compute-size graph))
    (apply #'add-subviews graph-editor (graph-edit-compute-subviews graph))
    )
  ;; Might actually want to make some of this code be the default method.
  (let ((dp (view-named 'data-points graph-editor))
        (pt (view-named 'plot-type graph-editor))
        (ri (view-named 'red-value graph-editor))
        (gi (view-named 'green-value graph-editor))
        (bi (view-named 'blue-value graph-editor))
        (xbc (view-named 'use-x-bounds graph-editor))
        (xmindi (view-named 'x-min graph-editor))
        (xmaxdi (view-named 'x-max graph-editor))
        (ybc (view-named 'use-y-bounds graph-editor))
        (ymindi (view-named 'y-min graph-editor))
        (ymaxdi (view-named 'y-max graph-editor))
        (graph (graph-editor-graph graph-editor)))
    ;; set up string tags
    (set-window-title graph-editor (concatenate 'string
                                                (clasp::name graph)
                                                " Settings"))
    (set-dialog-item-text dp (format nil "~D data points" (length (cdar (variable-plot-data
                                                                         graph)))))
    (set-dialog-item-text pt (format nil "~:(~A~)" (type-of graph)))
    ;; set up the rgb dialogs
    (set-rgb-items graph-editor)
    (multiple-value-bind (red green blue)
                         (color-values (plot-color
                                        (car (variable-plot-data ;; FIX no good for multiple plots
                                              graph))))
      (progn
        (set-dialog-item-text bi (format nil "~D" blue))
        (set-dialog-item-text gi (format nil "~D" green))
        (set-dialog-item-text ri (format nil "~D" red))))
    (select-all ri)
    ;; set up the custom bounds gadgets
    (multiple-value-bind (xmin xmax xcp ymin ymax ycp)
                         (xy-bounds graph)
      (if xcp (progn (dialog-item-enable xmindi) (dialog-item-enable xmaxdi) 
                     (check-box-check xbc)))
      (if ycp (progn (dialog-item-enable ymindi) (dialog-item-enable ymaxdi) 
                     (check-box-check ybc)))
      (set-dialog-item-text xmindi (if xmin
                                     (format nil "~S" xmin)
                                     (dialog-item-disable xbc)))
      (set-dialog-item-text xmaxdi (if xmax
                                     (format nil "~S" xmax)
                                     (dialog-item-disable xbc)))
      (set-dialog-item-text ymindi (if ymin
                                     (format nil "~S" ymin)
                                     (dialog-item-disable ybc)))
      (set-dialog-item-text ymaxdi (if ymax
                                     (format nil "~S" ymax)
                                     (dialog-item-disable ybc))))
    )
  (select-all graph-editor)
  (window-show graph-editor))

(defmethod graph-edit-editor-class ((graph graph))
  'graph-editor)

(defmethod graph-edit-compute-size ((graph graph))
   (values #@(320 145)))

(defmethod graph-edit-compute-subviews ((graph graph))
  (let* ((size (graph-edit-compute-size graph))
         (height (point-v size)))
    (list (make-instance 'static-text-dialog-item
            :view-nick-name 'data-points
            :view-position #@(165 5)
            :view-size #@(140 15))
          (make-instance 'default-button-dialog-item
            :view-nick-name 'change-button 
            :dialog-item-text "Change"
            :view-position (make-point 230 (- height 25))
            :dialog-item-action #'(lambda (item)
                                    (change-graph (view-container item))))
          (make-instance 'static-text-dialog-item
            :view-nick-name 'plot-type
            :view-position #@(10 5)
            :view-size #@(140 15))
          (make-instance 'colors-pum
            :view-nick-name 'color-menu
            :view-position #@(90 28)
            :dialog-item-text "Color")
          (make-instance 'static-text-dialog-item
                         :dialog-item-text "Red"
                         :view-position #@(5 55))
          (make-instance 'number-entry-item :view-nick-name 'red-value
                         :allow-negatives nil :allow-floats nil
                         :view-position #@(40 55) :view-size #@(50 15))
          (make-instance 'static-text-dialog-item
                         :dialog-item-text "Green"
                         :view-position #@(108 55))
          (make-instance 'number-entry-item :view-nick-name 'green-value
                         :allow-negatives nil :allow-floats nil
                         :view-position #@(156 55) :view-size #@(50 15))
          (make-instance 'static-text-dialog-item
                         :dialog-item-text "Blue"
                         :view-position #@(222 55))
          (make-instance 'number-entry-item :view-nick-name 'blue-value
                         :allow-negatives nil :allow-floats nil
                         :view-position #@(260 55) :view-size #@(50 15))

          (make-instance 'disabler-checkbox-item :view-nick-name 'use-x-bounds
                         :dialog-item-text "Custom X bounds"
                         :view-position #@(15 85)
                         :check-box-checked-p nil
                         :watches-over '(x-min x-max))
          (make-instance 'number-entry-item :view-nick-name 'x-min
                         :dialog-item-enabled-p nil
                         :view-position #@(160 85) :view-size #@(59 15))
          (make-instance 'number-entry-item :view-nick-name 'x-max
                         :dialog-item-enabled-p nil
                         :view-position #@(240 85) :view-size #@(59 15))
          (make-instance 'disabler-checkbox-item :view-nick-name 'use-y-bounds
                         :dialog-item-text "Custom Y bounds"
                         :view-position #@(15 110)
                         :check-box-checked-p nil
                         :watches-over '(y-min y-max))
          (make-instance 'number-entry-item :view-nick-name 'y-min
                         :dialog-item-enabled-p nil
                         :view-position #@(160 110) :view-size #@(59 15))
          (make-instance 'number-entry-item :view-nick-name 'y-max
                         :dialog-item-enabled-p nil
                         :view-position #@(240 110) :view-size #@(59 15))
          )))


(defvar *graph-editor-affects-keywords* '(:X-FIRST :X-LAST :Y-FIRST :Y-LAST)
  "The keywords that the graph editor works on.")

(defmethod replace-keywords ((old-keywords LIST) (new-keywords LIST) (replace-keywords LIST))
  "replaces the keywords with the new ones."
  (let ((result '())
        marker)
    (loop for item in old-keywords do
          (if (typep item 'keyword)
            (if (not (or (find item replace-keywords)
                         (find item new-keywords)))
              (progn
                (push (elt old-keywords (setf marker (search (list item) old-keywords))) result)
                (push (elt old-keywords (1+ marker)) result)))))
    (append (reverse result) new-keywords)))

(defmethod change-graph ((graph-editor graph-editor))
  (let* ((graph (graph-editor-graph graph-editor))
         (uxb (view-named 'use-x-bounds graph-editor))
         (uyb (view-named 'use-y-bounds graph-editor))
         (xlb (number-item-value (view-named 'x-min graph-editor)))
         (xub (number-item-value (view-named 'x-max graph-editor)))
         (ylb (number-item-value (view-named 'y-min graph-editor)))
         (yub (number-item-value (view-named 'y-max graph-editor)))
         new-keyword-list
         new-plot-options)    
    (setf new-plot-options `(:color ,(make-color
                                      (number-item-value (view-named 'red-value graph-editor))
                                      (number-item-value (view-named 'green-value graph-editor))
                                      (number-item-value (view-named 'blue-value graph-editor)))))
    (setf (variable-plot-data (graph-editor-graph graph-editor))
          (mapcar #'(lambda (plot) 
                      (install-plot-options plot new-plot-options))
                  (variable-plot-data (graph-editor-graph graph-editor))))
    (if (check-box-checked-p uxb)
      (progn (push xlb new-keyword-list) (push :x-first new-keyword-list)
             (push xub new-keyword-list) (push :x-last new-keyword-list)))
    (if (check-box-checked-p uyb)
      (progn (push ylb new-keyword-list) (push :y-first new-keyword-list)
             (push yub new-keyword-list) (push :y-last new-keyword-list)))
    (setf (graph-keywords graph)
          (replace-keywords (graph-keywords graph)
                            new-keyword-list
                            *graph-editor-affects-keywords*))))

(defmethod change-graph :around ((graph-editor graph-editor))
  (let* ((right-graph (find (clasp::name (graph-editor-graph graph-editor))
                            (windows) :test #'(lambda (item1 item2)
                                                     (string-equal item1 (window-title item2))))))
    (call-next-method)
    (if right-graph (ignore-errors (window-close right-graph) (open-clasp-object 
                                                               (graph-editor-graph graph-editor))))))

(defun color-value-by-name (name)
  (let ((result 0))
    (loop for x in *color-table* do
          (if (equal (car x) name)
            (setq result (cadr x))))
    result))

(defmethod set-rgb-items ((graph-editor graph-editor))
  (let* ((c-menu (view-named 'color-menu graph-editor))
         (c-picked (menu-item-title (selected-item c-menu)))
         (ri (view-named 'red-value graph-editor))
         (gi (view-named 'green-value graph-editor))
         (bi (view-named 'blue-value graph-editor))
         (rgb-state-set (if (equal c-picked "Custom") 'dialog-item-enable 'dialog-item-disable)))
    (funcall rgb-state-set ri)
    (funcall rgb-state-set gi)
    (funcall rgb-state-set bi)
    (if (equal rgb-state-set 'dialog-item-disable)
      (multiple-value-bind (red green blue)
                           (color-values (color-value-by-name c-picked))
        (progn
          (set-dialog-item-text bi (format nil "~D" blue))
          (set-dialog-item-text gi (format nil "~D" green))
          (set-dialog-item-text ri (format nil "~D" red)))))))

(defmethod make-graph-editor ((graph graph))
  (let ((edit-window (find-window (concatenate 'string (clasp::name graph) " Settings"))))
    (if edit-window 
      (window-select edit-window)
      (make-instance (graph-edit-editor-class graph) :graph graph))))

;;; -------------------------------------------------------------------------------------
;;; Histogram editor

(defclass histogram-editor (graph-editor)
  ()
  )

(defmethod graph-edit-editor-class ((graph histogram))
  'histogram-editor)

(defmethod graph-edit-compute-subviews ((graph histogram))
  ;; In lieu of a real append method combination we do this...
  (let* ((size (graph-edit-compute-size graph))
         (height (point-v size)))
    (append (call-next-method)
            (list (make-instance 'static-text-dialog-item
                    :dialog-item-text "Number of Bins"
                    :view-position (make-point 15 (- height 20)))
                  (make-instance 'number-entry-item 
                    :view-nick-name 'number-of-bins
                    :dialog-item-text (princ-to-string (slot-value graph 'number-of-bins))
                    :allow-negatives nil :allow-floats nil
                    :view-position (make-point 140 (- height 25))
                    :view-size #@(50 15))
                  ))))

(defmethod graph-edit-compute-size ((graph histogram))
  ;; In lieu of a real add-sizes method combination we do this...
  (let* ((size (call-next-method))
         (width (point-h size))
         (height (point-v size)))
    (make-point width (+ height 30))))

(defmethod change-graph :after ((graph-editor histogram-editor))
  (let ((graph (graph-editor-graph graph-editor)))
    (setf (slot-value graph 'number-of-bins)
          (number-item-value (view-named 'number-of-bins graph-editor)))
    (setf (variable-plot-data graph) (generate-plots graph))))

;;; -------------------------------------------------------------------------------------
;; Scatter, row-line, line, regression editor

(defclass line-or-symbol-plot-editor (graph-editor)
  ()
  )

(defmethod graph-edit-editor-class ((graph line-or-symbol-plot-mixin))
  'line-or-symbol-plot-editor)

;; The above should be enough, but strangely does not work. This is a hack.
#+CRAP
(defmethod graph-edit-editor-class ((graph row-line-plot))
  'line-or-symbol-plot-editor)
#+CRAP
(defmethod graph-edit-editor-class ((graph scatterplot))
  'line-or-symbol-plot-editor)
#+CRAP
(defmethod graph-edit-editor-class ((graph line-plot))
  'line-or-symbol-plot-editor)
#+CRAP
(defmethod graph-edit-editor-class ((graph regression-plot))
  'line-or-symbol-plot-editor)

(defmethod initialize-instance :after ((graph-editor line-or-symbol-plot-editor) &key)
  (let ((ucc (view-named 'use-custom-char graph-editor))
        (ccdi (view-named 'custom-char graph-editor))
        (xsh (view-named 'x-shift graph-editor))
        (graph (graph-editor-graph graph-editor))
        marker)
    ;; set up custom character thingy
    (if (typep graph 'scatterplot)
      (if (setf marker (search (list :char) (car (cddaar (variable-plot-data graph)))))
        (progn 
          (set-dialog-item-text ccdi (string (elt (car (cddaar (variable-plot-data graph))) (1+ marker))))
          (dialog-item-enable ccdi)
          (check-box-check ucc)))
      (dialog-item-disable ucc))
    ;; set up x shift box
    (if (ignore-errors (shift-x graph))
      (progn (dialog-item-enable xsh)))))

(defmethod graph-edit-compute-size ((graph line-or-symbol-plot-mixin))
  ;; In lieu of a real add-sizes method combination we do this...
  (let* ((size (call-next-method))
         (width (point-h size))
         (height (point-v size)))
    (make-point width (+ height 65))))

(defmethod graph-edit-compute-subviews ((graph line-or-symbol-plot-mixin))
  ;; In lieu of a real append method combination we do this...
  (let* ((size (graph-edit-compute-size graph))
         (height (point-v size)))
    (append (call-next-method)
            (list
             (make-instance 'disabler-checkbox-item :view-nick-name 'use-custom-char
                            :dialog-item-text "Use custom symbol"
                            :check-box-checked-p nil
                            :watches-over '(custom-char)
                            :view-position (make-point 32 (- height 60)))
             (make-instance 'editable-text-dialog-item :view-nick-name 'custom-char
                            :dialog-item-enabled-p nil :dialog-item-text "o"
                            :view-position (make-point 215 (- height 60))
                            :view-size #@(59 15))
             ;; This might be taken out of here
             (make-instance 'static-text-dialog-item :dialog-item-text "X Shift"
                            :view-position (make-point 15 (- height 25)))
             (make-instance 'number-entry-item :view-nick-name 'x-shift
                            :dialog-item-enabled-p nil :dialog-item-text "0"
                            :view-position (make-point 75 (- height 25))
                            :view-size #@(59 15))))))

(defmethod change-graph :after ((graph-editor line-or-symbol-plot-editor))
  (let* ((graph (graph-editor-graph graph-editor))
         (ucc (view-named 'use-custom-char graph-editor))
         (ccdi (view-named 'custom-char graph-editor))
         (xsh (number-item-value (view-named 'x-shift graph-editor)))
         new-plot-options)
    (if (and (dialog-item-enabled-p (view-named 'x-shift graph-editor))
             (ignore-errors (/= xsh (shift-x graph))))
      (progn (setf (shift-x graph) xsh) (setf (variable-plot-data graph) (generate-plots graph))))
    (if (and (check-box-checked-p ucc) (> (length (dialog-item-text ccdi)) 0))
      (let ((text (dialog-item-text ccdi)))
        (cond 
         ;; This is a hack to get it to work now. We really should update the dialog box to have radio 
         ;; buttons to select the one you want.
         ((plotter::symbol-displayer (string-upcase text) nil nil nil)
          (setf new-plot-options (append `(:SYMBOL ,(clasp::form-keyword (string-upcase text))) new-plot-options)))
         (t
          ;; allow old style slow stuff
          (setf new-plot-options (append `(:SYMBOL ,(aref (dialog-item-text ccdi) 0)) new-plot-options))))))
      (setf (variable-plot-data (graph-editor-graph graph-editor))
          (mapcar #'(lambda (plot) (install-plot-options plot new-plot-options))
                  (variable-plot-data (graph-editor-graph graph-editor)))
          )))

  

;;;
;;; overlaid graph editing. it's easy because we just spawn editors for each component graph.
;;;

(defclass overlaid-graph-editor (window)
  ((graph
    :initform nil
    :initarg :graph
    :type overlaid-graph
    :accessor overlaid-graph-editor-graph
    :documentation "The graph that is being operated on."))
  (:documentation "a way of launching overlaid graph component's editors.")
  (:DEFAULT-INITARGS
    :WINDOW-TYPE :DOCUMENT
    :VIEW-SIZE #@(300 150)
    :VIEW-POSITION (best-spot 300 150)
    :VIEW-SUBVIEWS (list (make-instance 'button-dialog-item 
                           :view-position #@(230 20)
                           :dialog-item-text "edit"
                           :dialog-item-action #'(lambda (item) (launch-graph-editor (view-container item))))
                         (make-instance 'pop-up-menu
                           :view-position #@(10 16)
                           :dialog-item-text "Graph"
                           :view-size #@(200 18)
                           :view-nick-name 'component-pum)
                         (make-instance 'disabler-checkbox-item :view-nick-name 'use-x-bounds
                                        :dialog-item-text "Custom X bounds"
                                        :view-position #@(5 50)
                                        :check-box-checked-p nil
                                        :watches-over '(x-min x-max))
                         (make-instance 'number-entry-item :view-nick-name 'x-min
                                        :dialog-item-enabled-p nil
                                        :view-position #@(150 50) :view-size #@(59 15))
                         (make-instance 'number-entry-item :view-nick-name 'x-max
                                        :dialog-item-enabled-p nil
                                        :view-position #@(230 50) :view-size #@(59 15))
                         (make-instance 'disabler-checkbox-item :view-nick-name 'use-y-bounds
                                        :dialog-item-text "Custom Y bounds"
                                        :view-position #@(5 75)
                                        :check-box-checked-p nil
                                        :watches-over '(y-min y-max))
                         (make-instance 'number-entry-item :view-nick-name 'y-min
                                        :dialog-item-enabled-p nil
                                        :view-position #@(150 75) :view-size #@(59 15))
                         (make-instance 'number-entry-item :view-nick-name 'y-max
                                        :dialog-item-enabled-p nil
                                        :view-position #@(230 75) :view-size #@(59 15))
                         (make-instance 'default-button-dialog-item
                           :view-nick-name 'change-button 
                           :dialog-item-text "Change"
                           :view-position #@(130 110)
                           :dialog-item-action #'(lambda (item)
                                                   (change-graph (view-container item))))
                         )))

(defmethod initialize-instance :after ((oge overlaid-graph-editor) &key)
  (set-window-title oge (clasp::name (overlaid-graph-editor-graph oge)))
  (mapcar #'(lambda (item) (add-menu-items (view-named 'component-pum oge)
                                           (make-instance 'menu-item 
                                             :menu-item-title (clasp::name item))))
          (graphs-overlaid-graphs (overlaid-graph-editor-graph oge)))
  (let ((xbc (view-named 'use-x-bounds oge))
        (xmindi (view-named 'x-min oge))
        (xmaxdi (view-named 'x-max oge))
        (ybc (view-named 'use-y-bounds oge))
        (ymindi (view-named 'y-min oge))
        (ymaxdi (view-named 'y-max oge))
        (graph (overlaid-graph-editor-graph oge)))
    (multiple-value-bind (xmin xmax xcp ymin ymax ycp)
                         (xy-bounds graph)
      (if xcp (progn (dialog-item-enable xmindi) (dialog-item-enable xmaxdi) 
                     (check-box-check xbc)))
      (if ycp (progn (dialog-item-enable ymindi) (dialog-item-enable ymaxdi) 
                     (check-box-check ybc)))
      (set-dialog-item-text xmindi (if xmin
                                     (format nil "~S" xmin)
                                     (dialog-item-disable xbc)))
      (set-dialog-item-text xmaxdi (if xmax
                                     (format nil "~S" xmax)
                                     (dialog-item-disable xbc)))
      (set-dialog-item-text ymindi (if ymin
                                     (format nil "~S" ymin)
                                     (dialog-item-disable ybc)))
      (set-dialog-item-text ymaxdi (if ymax
                                     (format nil "~S" ymax)
                                     (dialog-item-disable ybc))))))
  
(defmethod launch-graph-editor ((oge overlaid-graph-editor))
  (let* ((graph-number (search (list (selected-item (view-named 'component-pum oge))) (menu-items (view-named 'component-pum oge))))
         (graph (elt (graphs-overlaid-graphs (overlaid-graph-editor-graph oge)) graph-number)))
    (make-instance 'graph-editor :graph graph)))

(defmethod change-graph ((oge overlaid-graph-editor))
  (let ((uxb (view-named 'use-x-bounds oge))
         (uyb (view-named 'use-y-bounds oge))
         (xlb (number-item-value (view-named 'x-min oge)))
         (xub (number-item-value (view-named 'x-max oge)))
         (ylb (number-item-value (view-named 'y-min oge)))
         (yub (number-item-value (view-named 'y-max oge)))
         new-keyword-list)
    (if (check-box-checked-p uxb)
      (progn (push xlb new-keyword-list) (push :x-first new-keyword-list)
             (push xub new-keyword-list) (push :x-last new-keyword-list)))
    (if (check-box-checked-p uyb)
      (progn (push ylb new-keyword-list) (push :y-first new-keyword-list)
             (push yub new-keyword-list) (push :y-last new-keyword-list)))
    (setf (graph-keywords (overlaid-graph-editor-graph oge))
          (replace-keywords (graph-keywords (overlaid-graph-editor-graph oge))
                            new-keyword-list
                            *graph-editor-affects-keywords*))))
