;;;
;;; history-view.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines the history-view class and methods.


================================================================
Status =========================================================
================================================================
In-progress.


================================================================
Change history =================================================
================================================================
22-May-94 mc	Created.
19-Jul-94 ms    Added definitions & support for variable history fonts.
16-Aug-94 ms    Added the comment support functions.

|#


(in-package "CLASP-INTERFACE")

(defvar *history-buffer-last-close* '()
  "The contents of the history buffer after the CLASP window was last closed")

;;;
;;; Define history-view .
;;;

(defclass history-view (sequence-dialog-item)
  ()
  (:documentation "Shows pairs of user commands and command results,
with a heavy line separating pairs. Selection is :disjoint . Selected
commands and results can be 'logged' to the logging program for
editing and reporting. Abbreviated 'hv'.")
  (:DEFAULT-INITARGS
    :view-nick-name 'history-view
    ;:view-position #@(3 2)              ; set by parent window
    ;:view-size #@(294 146)              ; set by parent window
    :CELL-SIZE #@(350 11)
    :SELECTION-TYPE :DISJOINT
    :TABLE-HSCROLLP NIL
    :TABLE-VSCROLLP T
    :TABLE-SEQUENCE *history-buffer-last-close*   ; modified by add-command&result-hv
    :view-font *font-spec-variables*
    :table-print-function
    #'(lambda (command-or-result stream)
        "Formats command-or-result to stream and draws a heavy
line if command-or-result is typep 'clasp-command, to visually
separate command-result pairs."
        (declare (ignore stream))
        ;;
        (let* ((*package* (find-package "CLASP-INTERFACE"))
              (font-spec   ;; FIX: Use different fonts or draw a line.
               ;; choose the correct font spec based on item type
                           (cond 
                             ((typep command-or-result 'clasp-command)
                              *font-spec-history-command*)
                             ((typep command-or-result 'simple-string)
                              *font-spec-history-label*)
                             ((typep command-or-result 'comment)
                              *font-spec-history-comment*)
                             ((typep command-or-result 'clasp-result)
                              *font-spec-history-result*)
                             (t             *font-spec-history-result*)))
              ;; Print control var (~A, ~S, or *float-format-string) based on type
              (string (format nil (if (or (stringp command-or-result) (typep command-or-result 'clasp-result)
                                          (typep command-or-result 'comment))  ;; ascii output
                                    "~A"
                                    (if (or (typep command-or-result 'ratio)  ;; do float formatting ?
                                            (typep command-or-result 'float))
                                      *float-format-string*
                                      "~S"))
                              ;; this hairy sequence of if's returns the actual print item depending on type.
                              (if (typep command-or-result 'clasp-command)    ;; pretty-print commands
                                (let ((describe (sexp-describe-clasp-command command-or-result))
                                      (command (get-descript-tag command-or-result 'COMMAND)))
                                  (if command  ;; prepare the command for printing
                                    (mapcar #'(lambda (item) (if (equal (ignore-errors (car item))
                                                                        'COMMAND)
                                                               (cadr item)
                                                               item))
                                            describe)
                                    (sexp-clasp-command command-or-result))) ;; for untagged commands
                                (if (typep command-or-result 'ratio)         ;; ratios are ugly
                                  (coerce command-or-result 'float)
                                  (if (typep command-or-result 'comment)     ;; comments must be extracted
                                    (comment-contents command-or-result)
                                    (if (typep command-or-result 'clasp-result)  ;; print the text of the result
                                      (clasp-result-text command-or-result)
                                      command-or-result)))))))                   ;; what is this ?? (floats)
          (setf string (trim-to-maxlen string 255)) ; pstrs cannot exceed 255
          (multiple-value-bind (ff ms) (font-codes font-spec)
            (with-font-codes ff ms
              (with-pstrs ((p-str string))
                (#_DrawString p-str))))
          ))))

(defmethod initialize-instance :after ((history-view history-view) &key)
  (set-cell-size history-view (point-h (cell-size history-view)) (font-height *font-spec-variables*)))


(defmethod l-selected-items-hv ((history-view history-view))
  "Returns a list of items selected in history-view. FIX: Need
terminology for 'items', such as dataset, variable, result, graph, etc."
  ;;
  (let* ((l-items (table-sequence history-view))
         (l-cell-selected (selected-cells history-view))
         (l-selected-items (map 'list #'(lambda (cell)
                                          (elt l-items (point-v cell)))
                                l-cell-selected)))
    l-selected-items))

;; 16-Aug-94 added for comment capability
(defvar *record-p* 'option-key-p)

(defmethod receive-comment ((view history-view) (comment comment))
  (let* ((selected-cells (reverse (selected-cells view)))
         (first-selected-element (if selected-cells (point-v (car selected-cells))))
         (elements (table-sequence view)))
    (if first-selected-element
      (set-table-sequence view (append (subseq elements 0 first-selected-element)
                                         (list comment)
                                         (subseq elements first-selected-element)))
      (set-table-sequence view (append (table-sequence view) (list comment))))
    (scroll-to-cell view
                    (index-to-cell view
                                   (if first-selected-element
                                     first-selected-element
                                     (1- (length (table-sequence view))))))))

(defmethod view-click-event-handler :after ((view history-view)
                                            (pt-where fixnum))
  "Handles double-clicks in view by opening all selected objects that
f-can-open-clasp-object ."
  ;;
  ;; display the item in the ribbon
  (let ((command-or-result (ignore-errors (elt (table-sequence view) 
                                          (cell-to-index view 0 (point-v (point-to-cell view pt-where)))))))
    (if command-or-result (ribbon-display (format nil (if (or (stringp command-or-result) (typep command-or-result 'clasp-result)
                                                         (typep command-or-result 'comment))  ;; ascii output
                                                   "~A"
                                                   (if (or (typep command-or-result 'ratio)  ;; do float formatting ?
                                                           (typep command-or-result 'float))
                                                     *float-format-string*
                                                     "~S"))
                                             ;; this hairy sequence of if's returns the actual print item depending on type.
                                             (if (typep command-or-result 'clasp-command)    
                                               (sexp-clasp-command command-or-result)        ;; for untagged commands
                                               (if (typep command-or-result 'ratio)         ;; ratios are ugly
                                                 (coerce command-or-result 'float)
                                                 (if (typep command-or-result 'comment)     ;; comments must be extracted
                                                   (comment-contents command-or-result)
                                                   (if (typep command-or-result 'clasp-result)  ;; print the text of the result
                                                     (clasp-result-text command-or-result)
                                                     command-or-result))))))))
  (let* ((l-selected-items (l-selected-items-hv view))
         (l-openable-items (remove-if-not
                            #'(lambda (item)
                                (f-can-open-clasp-object item))
                            l-selected-items)))
    (if (funcall *record-p*)
      (progn
        (setq *clasp-recorded-items* (append *clasp-recorded-items* (reverse l-selected-items)))
        (let ((notebook (find-window "CLASP Notebook")))
          (if notebook (fill-caddy notebook)))))
    (when (and (funcall *open-click-p*)
               (< (point-h pt-where) (- (point-h (view-size (view-container view))) 16)))
      (cond (l-openable-items
             (eval-enqueue `(map nil #'open-clasp-object ',l-openable-items)))
            (t nil )))))


;;;
;;; Define command-related functions.
;;;

(defmethod add-command&result-hv ((clasp-command clasp-command)
                                    (history-view history-view))
  "The top-level function for adding commands to the history view. Adds
items to the end of history-view's table-sequence corresponding to
clasp-command. The first item is clasp-command, which is printed as
its sexp-clasp-command . Remaining items are added for each of
clasp-command's l-result-clasp-command ."
  ;;
  (set-table-sequence history-view
                      (append (table-sequence history-view)
                              (cons clasp-command
                                    (l-result-clasp-command clasp-command))))
  (scroll-to-cell history-view
                  (index-to-cell history-view
                                 (1- (length (table-sequence history-view))))))


(defmethod clear-history-view ((history-view history-view))
  "The top-level command for clearing the history view of all elements."
  ;;
  (set-table-sequence history-view ()))

;;;
;;;  some extraneous openable items in the history view
;;;


(defmethod f-can-open-clasp-object ((object window))
  t)

(defmethod open-clasp-object ((object window))
  "open a clasp window that's still in the history buffer."
  (if (wptr object)         ;; a hack way to see if the window still exists
    (window-select object)
    (gui-warn "That window has been closed.")))
