;;;
;;; menu-items.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines menu-item subclasses that are enabled/disabled in certain
conditions.


================================================================
Status =========================================================
================================================================
In-progress.


================================================================
Change history =================================================
================================================================
22-May-94 mc	Created.
6-Jul-94  ms    Added 'some-variables-menu-item for X(s) functions
|#


(in-package "CLASP-INTERFACE")

(export '())

;;;
;;; The most common menu item...clasp-window-menu-item only good when there is one
;;; 21-Jul-94 ms
;;;

(defclass clasp-window-menu-item (menu-item)
  ()
  (:documentation "A menu-item subclass that is enabled when there exists
a valid clasp window."))

(defmethod menu-item-update :after ((clasp-window-menu-item clasp-window-menu-item))
  "Dis/enables the item according to the definition."
  (if (clasp-window-front)
    (menu-item-enable clasp-window-menu-item)
    (menu-item-disable clasp-window-menu-item)))

;;;
;;; Define dataset-related items.
;;;


;;;

;;;

(defclass one-dataset-menu-item (menu-item)
  ()
  (:documentation "A menu-item subclass that is enabled when the front
window is a clasp-window and exactly one dataset is selected in the
front clasp-window."))


(defmethod menu-item-update :after ((one-dataset-menu-item one-dataset-menu-item))
  "Disables/enables the item according to its definition."
  ;;
  (if (and (eq (clasp-window-front)
               (first (windows)))
           (f-one-dataset-selected-cw (clasp-window-front)))
    (menu-item-enable one-dataset-menu-item)
    (menu-item-disable one-dataset-menu-item)))


;;;

;;;

;;;

;;;
;;; Define history-related items.
;;;

;;;
(defclass history-selection-menu-item (menu-item)
  ()
  (:documentation "A menu-item subclass that is enabled when the front
window is a clasp-window and there is a history selection."))


(defmethod menu-item-update :after ((history-selection-menu-item history-selection-menu-item))
  "Disables/enables the item according to its definition."
  ;;
  (if (and (eq (clasp-window-front)
               (first (windows)))
           (l-history-selected-cw (clasp-window-front)))
    (menu-item-enable history-selection-menu-item)
    (menu-item-disable history-selection-menu-item)))
;;;

(defclass openable-selection-menu-item (menu-item)
  ()
  (:documentation "A menu-item subclass that is enabled when the front
window is a clasp-window and there is a history and or data selection
that is openable via f-can-open-clasp-object ."))


(defmethod menu-item-update :after ((openable-selection-menu-item openable-selection-menu-item))
  "Disables/enables the item according to its definition."
  ;;
  (let* ((f-clasp-window-front (eq (clasp-window-front)
                                   (first (windows))))
         (f-selection-openable (and f-clasp-window-front
                                    (f-selection-openable-cw (clasp-window-front)))))
    (if f-selection-openable
      (menu-item-enable openable-selection-menu-item)
      (menu-item-disable openable-selection-menu-item))))
;;;

(defclass ribbon-open-menu-item (menu-item)
  ()
  (:documentation "A menu-item that is enabled when the ribbon window is."))

(defmethod menu-item-update :after ((romi ribbon-open-menu-item))
  (if (find-window "gibbon-monkey" 'ribbon-window)
    (menu-item-enable romi)
    (menu-item-disable romi)))

;;;
;;; variable moving menu items
;;;

;;; the special one that sends things to X and Y of undisclosed locations

(defclass dock-menu-item (menu-item)
  ((dock-name :initform "Dock" :initarg :dock-name :accessor dock-name)
   (affects :initform nil :initarg :affects :accessor affects))
  (:documentation "A menu item that changes depending on any and
every little thing you could imagine."))

(defmethod menu-item-update :after ((dock-menu-item dock-menu-item))
  (let* ((fw (progn (menu-item-enable dock-menu-item) (front-window)))
         (sourcewin (if (typep fw 'clasp-window) fw
                        (or (find-window "CLASP Dock") NIL)))
         (destwin (if (typep fw 'statistic-dialog-mixin) (prog1 fw (setf sourcewin (or (find-window "CLASP Dock") (clasp-window-front))))
                      (or (and (not (typep sourcewin 'clasp-dock)) (find-window "CLASP Dock"))
                          (find-if #'(lambda (item) (typep item 'clasp-interface::statistic-dialog-mixin)) (windows)))))
         )
    (if (and destwin sourcewin)
      (set-menu-item-title dock-menu-item
                           (format nil "~A -> ~A"
                                   (typecase sourcewin
                                     (clasp-window "Selected CLASP variables")
                                     (clasp-dock (dock-name dock-menu-item))
                                     (t (progn (menu-item-disable dock-menu-item) "Nothing ")))
                                   (typecase destwin
                                     (clasp-dock (dock-name dock-menu-item))
                                     (statistic-dialog-mixin 
                                      (with-slots (affects) dock-menu-item
                                        (or (ignore-errors 
                                             (format nil "~A ~A" (window-title destwin)
                                                     (dialog-item-text (view-named affects destwin))))
                                            (progn (menu-item-disable dock-menu-item) "Nothing"))))
                                     (t (progn (menu-item-disable dock-menu-item) "Nothing ")))))
      (menu-item-disable dock-menu-item))))

;;; operates on the most recent result (most recent variable result)

(defclass mrr-menu-item (menu-item)
  ((affects :initform NIL :initarg :affects))
  (:documentation "A menu item that changes depending on the most-recent-result."))

(defmethod menu-item-update :after ((mrr-menu-item mrr-menu-item))
  (menu-item-enable mrr-menu-item)
  (if (and (typep (front-window) 'statistic-dialog-mixin) *most-recent-vresult*)
    (with-slots (affects) mrr-menu-item
      (set-menu-item-title mrr-menu-item
                           (format nil "~:(~A~) -> ~A ~A"
                                   (clasp::name *most-recent-vresult*)
                                   (or (ignore-errors 
                                        (dialog-item-text (view-named affects (front-window))))
                                       (progn (menu-item-disable mrr-menu-item) ""))
                                   (window-title (front-window)))))
    (menu-item-disable mrr-menu-item)))

;;;

(defclass graph-window-menu-item (menu-item)
  ()
  (:documentation "A menu item that is active only when the front window is a graph."))

(defmethod menu-item-update :after ((gw-menu-item graph-window-menu-item))
  (if (typep (front-window) 'plotter::plot-window)
    (menu-item-enable gw-menu-item)
    (menu-item-disable gw-menu-item)))

;;;

(defclass table-window-menu-item (menu-item)
  ((slot-to-check :initarg :slot-to-check :accessor slot-to-check))
  (:documentation "A menu item that becomes active when the front window is a table
view and has a matching slot."))

(defmethod menu-item-update :after ((tw-menu-item table-window-menu-item))
  (let ((string (ignore-errors
                 (slot-value (table-view-table
                              (view-named 'table-array-item (front-window)))
                             (slot-to-check tw-menu-item)))))
    (if string
      (set-menu-item-title tw-menu-item string))
    (if (or string (and (typep (front-window) 'table-view-window) (null (slot-to-check tw-menu-item))))
      (menu-item-enable tw-menu-item)
      (menu-item-disable tw-menu-item))))