;;;
;;; New Menus
;;;

#| This is a new scheme of defining the CLASP menus that is amenable to patch-work |#

;;;
;;; 
;;;

(in-package :clasp-interface)

;; ================================================================
;; macros for defining menu items

(defparameter *menu-defs* NIL)

(defclass tag-mixin ()
  ((tag :initform NIL :initarg :tag :accessor tag)))

(defclass menudef (tag-mixin)
  ((title :initform NIL :initarg :title :accessor menu-title)
   (help-spec :initform NIL :initarg :help-spec :accessor menu-help-spec)
   (class :initform NIL :initarg :class :accessor menu-class)
   (menu-items :initform NIL :initarg :menu-items :accessor menu-menu-items))
  (:documentation "A definition for a menu."))

(defclass menu-itemdef (tag-mixin)
  ((class :initform NIL :initarg :class :accessor menu-item-class)
   (keyargs :initform NIL :initarg :keyargs :accessor menu-item-keyargs)
   (action :initform NIL :initarg :action :accessor menu-item-action))
  (:documentation "A definition of a menu item."))

(defmethod tag ((item t))
  NIL)

(defun find-menudef (tag list)
  (if (null list)
    NIL
    (if (typep (car list) 'menudef) 
      (if (eql (tag (car list)) tag)
        (car list)
        (or (find-menudef tag (menu-menu-items (car list))) (find-menudef tag (cdr list))))
      (find-menudef tag (cdr list)))))

(defmethod weed-menudef ((menudef menudef) tag)
  (setf (menu-menu-items menudef) (delete tag (menu-menu-items menudef) :key #'tag))
  (mapc #'(lambda (md) (if (typep md 'menudef) (weed-menudef md tag))) (menu-menu-items menudef)))

(defmacro destroy-menudef (tag)
  `(progn (setf *menu-defs* (delete ,tag *menu-defs* :key #'tag))
          (mapc #'(lambda (md) (if (typep md 'menudef) (weed-menudef md ,tag))) *menu-defs*)))

(defmethod print-menu (list &optional (l 0))
  (if (null list)
    NIL
    (progn
      (format t "~A~S~%" (make-string (* l 2) :initial-element #\space) (tag (car list)))
      (when (typep (car list) 'menudef)
        (print-menu (menu-menu-items (car list)) (1+ l)))
      (print-menu (cdr list) l))))

(defmacro defmenu (menu-parent-name menu-name menu-class title help-spec)
  (let ((menudef (gensym "MENUDEF"))
        (parent  (gensym "PARENT")))
  `(let ((,menudef (make-instance 'menudef :tag ,menu-name :class ,menu-class
                                 :title ,title :help-spec ,help-spec))
         (,parent (find-menudef ,menu-parent-name *menu-defs*)))
     (if ,parent
       (push ,menudef (menu-menu-items ,parent))
       (push ,menudef *menu-defs*)))))

(defmacro defmenu-item (menu-item-name (menu-name menu-item-class &rest keys) &body action)
  (let ((menu (gensym "MENU"))
        (mitem (gensym "MITEM")))
    `(let ((,menu (find-menudef ,menu-name *menu-defs*))
           (,mitem (make-instance 'menu-itemdef :tag ,menu-item-name
                                  :class ,menu-item-class
                                  :keyargs ',keys
                                  :action #'(lambda () ,@action))))
       (destroy-menudef ,menu-item-name)
       (if ,menu
          (push ,mitem (menu-menu-items ,menu))
          (error "Illegal forward reference to menu ~S" ,menu-name)))))

(defmacro defmenu-raw (menu)
  `(push ,menu *menu-defs*))

(defmethod menumake ((mid menu-itemdef))
  (apply #'make-instance (menu-item-class mid) :menu-item-action (menu-item-action mid)
         (menu-item-keyargs mid)))

(defmethod menumake ((md menudef))
  (make-instance (menu-class md) :menu-title (menu-title md) :help-spec (menu-help-spec md)
                 :menu-items (mapcar #'menumake (reverse (menu-menu-items md)))))

(defmethod menumake ((m menu))
  m)

(defmethod menumake ((fun function))
  (funcall fun))

(proclaim '(inline compute-menus))
(defun compute-menus (list)
  (mapcar #'menumake (reverse list)))

(defun install-as-menu (menu-name defs &optional (leaders NIL))
  (let ((oldmenu (find-menu menu-name)))
    (when oldmenu (menu-deinstall oldmenu))
    (menu-install (make-instance 'menu :menu-title menu-name
                                 :menu-items (append leaders (compute-menus defs))))))

(defun install-menubar (defs)
  (set-menubar (compute-menus defs)))

;;; ================================================================
;;; CLASP Basic Menus

;;; ================================================================
;;; file menu
(defmenu NIL  :file 'menu "File" "File menu for I/O and general options.")
(defmenu-item :load-dataset (:file 'MENU-ITEM :menu-item-title "Load Dataset"
                                   :help-spec "Prompts you to select a dataset file then loads it."
                                   #-MCG-TEST :command-key #-MCG-TEST #\O
                                   #+MCG-TEST :command-key #+MCG-TEST #\0)
  (eval-enqueue '(choose-and-load-dataset)))
(defmenu-item :import-dataset (:file 'MENU-ITEM :MENU-ITEM-TITLE "Import Dataset"
                                     :help-spec "Prompts you to select a data file then imports it. Use the Preferences commands to specify parameters such as file separator."
                                     :command-key #\I)
  (eval-enqueue '(choose-and-import-dataset)))
(defmenu-item :save-dataset (:file 'one-dataset-menu-item
                                   :MENU-ITEM-TITLE "Save Dataset"
                                   :help-spec "Prompts you to name a data file in which to save the selected dataset. Enabled when exactly one dataset is selected in the frontmost clasp window."
                                   :command-key #\S)
  (eval-enqueue '(save-selected-dataset)))
(defmenu-item :export-dataset (:file 'one-dataset-menu-item
                                     :MENU-ITEM-TITLE "Export Dataset"
                                     :help-spec "Prompts you to name a data file then exports the selected dataset to it. Use the Preferences commands to specify parameters such as file separator.")
  (eval-enqueue '(export-selected-dataset)))
(defmenu-item :close (:file 'menu-item :MENU-ITEM-TITLE "Close"
                            :command-key #\W)
  (eval-enqueue '(window-close (front-window))))
(defmenu-item :close-all (:file 'menu-item :menu-item-title "Close All" 
                                ;:command-key #\-
                                )
  (mapc #'(lambda (w) (unless (or (typep w 'listener) (typep w 'clasp-dock)
                                  (typep w 'fred-window) (typep w 'clasp-window))
                        (window-close w))) (windows)))
;;;
(defmenu-item :bar11 (:file 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu-item :clear-history (:file 'clasp-window-menu-item :MENU-ITEM-TITLE "Clear History"
                                    :command-key #\Backspace
                                    :help-spec "Clears the frontmost CLASP window's history view. Enabled when the front window is a CLASP window.")
  (eval-enqueue `(clear-history-view (clasp-window-front))))
(defmenu-item :update (:file 'clasp-window-menu-item :MENU-ITEM-TITLE "Update Dataset View"
                             :help-spec "Updates the dataset view in the frontmost CLASP window. Enabled when the front window is a CLASP window.")
  (eval-enqueue `(update-dataset-view (clasp-window-front))))
(defmenu-item :log-sel (:file 'history-selection-menu-item :MENU-ITEM-TITLE "Log Selection"
                              :command-key #\T
                              :help-spec "Logs the frontmost CLASP window's selected history items. Enabled when the front window is a CLASP window.")
  (eval-enqueue `(log-selected-items)))
(defmenu-item :copy-clip (:file 'menu-item :menu-item-title "Copy to Clipboard" 
                                :command-key #\C
                                :help-spec "Cuts the current window's selection to the clipboard.")
  #+DEMO (sorry-demo-version)
  #-DEMO (eval-enqueue `(cut-selection (front-window))))
(defmenu-item :copy-rib (:file 'ribbon-open-menu-item :menu-item-title "Copy Ribbon to Clipboard"
                               :command-key #\B
                               :help-spec "Cuts the ribbon window's contents to the clipboard.")
  (eval-enqueue `(cut-selection (find-window "gibbon-monkey" 'ribbon-window))))
(defmenu-item :paste (:file 'menu-item :menu-item-title "Paste from Clipboard"
                            :command-key #\V
                            :help-spec "Pastes the current clipboard to the current window.")
  (eval-enqueue `(paste-into-clasp (front-window))))
;;;
(defmenu-item :bar12 (:file 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu-item :prefs (:file 'MENU-ITEM :MENU-ITEM-TITLE "Preferences"
                            :command-key #\`)
  (eval-enqueue '(edit-preferences)))
;;;
(defmenu-item :bar13 (:file 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu-item :quit (:file 'MENU-ITEM :MENU-ITEM-TITLE "Quit"
                           :COMMAND-KEY #\Q
                           :help-spec "Closes all windows and exits from CLASP. If the contents of any window have not been saved, you will loose that data.")
  (eval-enqueue '(quit-clasp)))

;;; ================================================================
;;; graph menu

(defmenu NIL :graph 'MENU "Graph" "A menu full of graphing commands.")
(defmenu-item :histogram (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Histogram"
                                 :command-key #\H)
  (eval-enqueue '(graph-ov-dialog-make 'histogram "Histogram" :help '*histogram-help*)))
(defmenu-item :scatterplot (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Scatter Plot"
                                   :command-key #\A)
  (eval-enqueue '(graph-tv-dialog-make 'scatterplot "Scatterplot" t 
                  :help '*scatterplot-help* :categorical-x t :categorical-y t)))
(defmenu-item :clusterplot (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Cluster Plot")
  (eval-enqueue '(graph-tv-dialog-make 'cluster-plot "Cluster Plot" t
                  :help '*cluster-plot-help* :categorical-x nil :categorical-y nil)))
(defmenu-item :lineplot (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Line Plot"
                                :command-key #\L)
  (eval-enqueue '(graph-tv-dialog-make 'line-plot "Line Plot" NIL 
                  :help '*line-plot-help* :categorical-x nil :categorical-y nil)))
(defmenu-item :rowplot (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Row Line Plot"
                               :command-key #\N )
  (eval-enqueue '(graph-ov-dialog-make 'row-line-plot "Row Line Plot" 
                  :help '*row-line-help* :categorical-ok nil)))
(defmenu-item :regplot (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Regression Plot"
                               :command-key #\G )
  (eval-enqueue '(graph-tv-dialog-make 'regression-plot "Regression Plot" t 
                  :help '*regression-plot-help* :categorical-x nil :categorical-y nil)))
(defmenu-item :tableplot (:graph 'table-window-menu-item :MENU-ITEM-TITLE "Table plot"
                                 :command-key #\P :slot-to-check nil)
  (eval-enqueue `(make-table-plot (front-window))))
(defmenu-item :overlay (:graph 'clasp-window-menu-item :MENU-ITEM-TITLE "Overlay Graphs"
                               :command-key #\=)
  (eval-enqueue '(make-instance 'overlaid-graph-dialog)))
;;;
(defmenu-item :bar21 (:graph 'menu-item :MENU-ITEM-TITLE "-" :DISABLED t) NIL)
;;;
(defmenu-item :editgraph (:graph 'graph-window-menu-item :MENU-ITEM-TITLE "Edit"
                                 :command-key #\E)
  (eval-enqueue `(make-graph-editor (window-graph (front-window)))))
(defmenu-item :key (:graph 'graph-window-menu-item :MENU-ITEM-TITLE "Graph Key"
                           :command-key #\K)
  (eval-enqueue `(plotter::pop-key (front-window))))

;;; ================================================================
;;; describe menu

(defmenu NIL :describe 'MENU "Describe" "Descriptive stats")

(defmenu-item :length (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Data Length")
  (eval-enqueue `(one-variable-dialog-make 'clasp::data-length "Data Length"
                                           (make-line-mapper "~D is data length of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*data-length-help*)))
(defmenu-item :mean (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Mean")
  (eval-enqueue `(one-variable-dialog-make 'clasp::mean "Mean"
                                           (make-line-mapper "~,4F is mean of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*mean-help*
                                           :categorical-ok nil)))
(defmenu-item :variance (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Variance")
  (eval-enqueue `(one-variable-dialog-make 'clasp::variance "Variance"
                                           (make-line-mapper "~,4F is variance of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*variance-help* :categorical-ok nil)))
(defmenu-item :stddev (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Standard Deviation")
  (eval-enqueue `(one-variable-dialog-make 'clasp::standard-deviation "Standard Deviation"
                                           (make-line-mapper "~,4F is standard deviation of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*std-dev-help* :categorical-ok nil)))
(defmenu-item :minimum (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Minimum")
  (eval-enqueue `(one-variable-dialog-make 'clasp::minimum "Minimum"
                                           (make-line-mapper "~,4F is minimum of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*minimum-help* :categorical-ok nil)))
(defmenu-item :maximum (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Maximum")
  (eval-enqueue `(one-variable-dialog-make 'clasp::maximum "Maximum"
                                           (make-line-mapper "~,4F is maximum of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*maximum-help* :categorical-ok nil)))
(defmenu-item :range (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Range")
  (eval-enqueue `(one-variable-dialog-make 'clasp::range "Range"
                                           (make-line-mapper "~,4F is range of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*range-help* :categorical-ok nil)))
(defmenu-item :quantile (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Quantile")
  (eval-enqueue '(quantile-dialog-make)))
(defmenu-item :median (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Median")
  (eval-enqueue `(one-variable-dialog-make 'clasp::median "Median"
                                           (make-line-mapper "~,4F is median of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*median-help* :categorical-ok nil)))
(defmenu-item :trimmean (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Trimmed Mean")
  (eval-enqueue '(trimmed-mean-dialog-make)))
(defmenu-item :mode (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Mode")
  (eval-enqueue `(mode-dialog-make)))
(defmenu-item :iqrange (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Interquartile Range")
  (eval-enqueue `(one-variable-dialog-make 'clasp::interquartile-range "Interquartile Range"
                                           (make-line-mapper "~,4F is interquartile Range of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*iqr-help* :categorical-ok nil)))
(defmenu-item :skew (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Skew")
  (eval-enqueue `(one-variable-dialog-make 'clasp::skewness "Skew"
                                           (make-line-mapper "~,4F is skew of ~:(~A~)"
                                                             '(0 x-args))
                                           :help '*skew-help* :categorical-ok nil)))
(defmenu-item :summary (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Statistical Summary" :command-key #\M)
  (eval-enqueue `(one-variable-dialog-make 'clasp::statistical-summary "Statistical Summary" 
                                           (make-table-maker *statistical-summary-labels*
                                                             :table-type :column-table
                                                             :name-format "Statistical Summary of ~A"
                                                             :name-args '(X-ARGS))
                                           :help '*statistical-summary-help* :categorical-ok nil)))
(defmenu-item :covariance (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Covariance")
  (eval-enqueue `(two-variable-dialog-make 'clasp::covariance "Covariance"
                                           (make-line-mapper "~,4F is covariance of ~:(~A~) and ~:(~A~)"
                                                             '(0 X-ARGS Y-ARGS))
                                           :help '*covariance-help*
                                           :categorical-x NIL :categorical-y NIL)))
(defmenu-item :correlation (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Correlation")
  (eval-enqueue `(two-variable-dialog-make 'clasp::correlation "Correlation"
                  (make-line-mapper "~,4F is correlation of ~:(~A~) and ~:(~A~)"
                                    '(0 X-ARGS Y-ARGS))
                  :help '*correlation-help*
                  :categorical-x NIL :categorical-y NIL)))
(defmenu-item :parcorr (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Partial Correlation")
  (eval-enqueue `(partial-corr-dialog-make)))
(defmenu-item :crosscorr (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Cross Correlation")
  (eval-enqueue '(cross-correlation-dialog-make)))
(defmenu-item :autocorr (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Autocorrelation")
  (eval-enqueue '(autocorrelation-dialog-make)))
(defmenu-item :corrtable (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Correlation Table")
  (eval-enqueue `(corr-table-dialog-make)))
(defmenu-item :conttable (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "Contingency Table")
  (eval-enqueue `(two-variable-dialog-make 'clasp::make-contingency-table "Contingency Table"
                                           'ctable-format :help '*contingency-help*)))
#+FBD
(defmenu-item :fbd-model (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "FBD Causal Model")
  (eval-enqueue `(fbd-dialog-make)))
(defmenu-item :cdf (:describe 'clasp-window-menu-item :MENU-ITEM-TITLE "CDF Functions")
  (eval-enqueue `(cdf-dialog-make)))

;;; ================================================================
;;; manipulate menu

(defmenu NIL :manipulate 'MENU "Manipulate" "Dataset manipulation commands.")

(defmenu-item :rename (:manipulate 'clasp-window-menu-item :MENU-ITEM-TITLE "Rename"
                                   :help-spec "Renames the selected dataset. Enabled when the front window is a CLASP window with one dataset selected."
                                   :command-key #\R)
  (eval-enqueue '(rename-selected-things)))
(defmenu-item :delete (:manipulate 'clasp-window-menu-item :MENU-ITEM-TITLE "Delete Result" :command-key #\D)
  (eval-enqueue '(delete-selected-items)))
(defmenu-item :partition (:manipulate 'clasp-window-menu-item :MENU-ITEM-TITLE "Partition Dataset")
  (eval-enqueue '(partition-dataset-dialog-make)))
(defmenu-item :partition-on (:manipulate 'clasp-window-menu-item :MENU-ITEM-TITLE "Partition On")
  (eval-enqueue '(partition-on-dialog-make)))
(defmenu-item :merge (:manipulate 'clasp-window-menu-item :MENU-ITEM-TITLE "Merge Datasets")
  (eval-enqueue '(merge-datasets-dialog-make)))
;;;
(defmenu-item :bar41 (:manipulate 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu :manipulate :tables 'MENU "Tables" "Table manipulation stuff.")
(defmenu-item :cclick-table (:tables 'table-window-menu-item :MENU-ITEM-TITLE "do nothing"
                                     :help-spec "Control-click menu item for table."
                                     :slot-to-check 'control)
  (eval-enqueue '(progn (control-click (table-view-table
                                        (view-named 'table-array-item (front-window))))
                  (revalidate-view (view-named 'table-array-item (front-window))))))
(defmenu-item :oclick-table (:tables 'table-window-menu-item :MENU-ITEM-TITLE "do nothing"
                                     :help-spec "Option-click menu item for table."
                                     :slot-to-check 'option)
  (eval-enqueue '(progn (option-click (table-view-table (view-named 'table-array-item (front-window))))
                  (revalidate-view (view-named 'table-array-item (front-window))))))
(defmenu-item :dclick-table (:tables 'table-window-menu-item :MENU-ITEM-TITLE "do nothing"
                                     :help-spec "Double-click menu item for table."
                                     :slot-to-check 'double)
  (eval-enqueue '(progn (dbl-click (table-view-table (view-named 'table-array-item (front-window))))
                  (revalidate-view (view-named 'table-array-item (front-window))))))
;;;
(defmenu :manipulate :variables 'MENU "Variables" "Variable/data manipulation.")
(defmenu-item :xdock (:variables 'dock-menu-item :MENU-ITEM-TITLE "X Dock" :dock-name "X Dock"
                                 :command-key #-MCG-TEST #\X #+MCG-TEST #\- :affects x-var-item
                                 :help-spec "Sends the selected variables to the x dock.")
  (eval-enqueue `(operate-on-dock (front-window) 'x-var-dock)))
(defmenu-item :ydock (:variables 'dock-menu-item :MENU-ITEM-TITLE "Y Dock" :dock-name "Y Dock"
                                 :command-key #-MCG-TEST #\Y #+MCG-TEST #\= :affects y-var-item
                                 :help-spec "Sends the selected variables to the y dock.")
  (eval-enqueue `(operate-on-dock (front-window) 'y-var-dock)))
(defmenu-item :zdock (:variables 'dock-menu-item :MENU-ITEM-TITLE "Cache" :Dock-name "Cache"
                                 :command-key #\Z :affects z-var-item)
  (eval-enqueue `(operate-on-dock (front-window) 'gp-dock)))
;;;
(defmenu-item :bar42 (:variables 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu-item :lastx (:variables 'mrr-menu-item :MENU-ITEM-TITLE "X"
                                 :command-key #\[ :affects x-var-item
                                 :help-spec "Replaces the x arguments with the most recent result.")
  (eval-enqueue `(operate-on-mrr (front-window) 'x-var-item)))
(defmenu-item :lasty (:variables 'mrr-menu-item :MENU-ITEM-TITLE "Y"
                                 :command-key #\] :affects y-var-item
                                 :help-spec "Replaces the y arguments with the most recent result.")
  (eval-enqueue `(operate-on-mrr (front-window) 'y-var-item)))
(defmenu-item :lastz (:variables 'mrr-menu-item :MENU-ITEM-TITLE "Z"
                                 :command-key #\} :affects z-var-item
                                 :help-spec "Replaces the y arguments with the most recent result.")
  (eval-enqueue `(operate-on-mrr (front-window) 'z-var-item)))
;;;
(defmenu-item :bar43 (:variables 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
;;;
(defmenu-item :changetype (:variables 'openable-selection-menu-item :MENU-ITEM-TITLE "Change Type"
                                      :command-key #\J :help-spec "Changes the type to categorical or whatever.")
  (eval-enqueue '(change-selected-variable-types)))
(defmenu-item :edit (:variables 'openable-selection-menu-item :MENU-ITEM-TITLE "Edit"
                                :command-key #-MCG-TEST #\E #+MCG-TEST #\Tab
                                :help-spec "Opens the selected variables and datasets from the data view for editing/viewing.")
  (eval-enqueue '(open-openable-selection)))

;;; ================================================================
;;; transform menu

(defmenu NIL :transform 'MENU "Transform" "Data transformations.")

(defmenu-item :userdefined (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "User Defined")
  (eval-enqueue '(make-user-defined-dialog)))
(defmenu-item :sort (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Sort")
  (eval-enqueue '(sort-dialog-make)))
(defmenu-item :sortindex (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Sort by Index")
  (eval-enqueue '(sort-index-dialog-make)))
(defmenu-item :addconst (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Add a Constant")
  (eval-enqueue '(add-constant-dialog-make)))
(defmenu-item :mulconst (:transform  'clasp-window-menu-item :MENU-ITEM-TITLE "Multiply by Constant")
  (eval-enqueue '(mul-constant-dialog-make)))
(defmenu-item :natlog (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Natural Log")
  (eval-enqueue '(one-var-trans-dialog-make 'clasp::natural-log "Natural Log"
                  :help '*natural-log-help* :categorical-ok NIL)))
(defmenu-item :log10 (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Log10")
  (eval-enqueue '(one-var-trans-dialog-make 'clasp::log10 "Log Base 10"
                  :help '*log10-help* :categorical-ok NIL)))
(defmenu-item :recode (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Recode Categorical")
  (eval-enqueue '(make-instance 'recode-variable-dialog)))
(defmenu-item :reassign (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Reassign Bins")
  (eval-enqueue '(rebin-dialog-make)))
(defmenu-item :smooth (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Smooth")
  (eval-enqueue '(smooth-dialog-make)))
(defmenu-item :smooth4253h (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Smooth 4253h")
  (eval-enqueue '(one-var-trans-dialog-make 'clasp::smooth-variable-4253h "Smooth 4253h"
                  :help '*smooth-4253h-help* :categorical-ok NIL)))
(defmenu-item :derivative (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Discrete Derivative")
  (eval-enqueue '(one-var-trans-dialog-make 'clasp::Difference-a-column "Discrete Derivarive"
                  :help '*derivative-help* :categorical-ok NIL)))
(defmenu-item :standardize (:transform 'clasp-window-menu-item :MENU-ITEM-TITLE "Standardize")
  (eval-enqueue '(one-var-trans-dialog-make 'clasp::standardize-variable "Standardize"
                  :help '*standardize-help* :categorical-ok NIL)))

;;; ================================================================
;;; test menu

(defmenu nil :test 'MENU "Test" "Hypothesis testing.")
(defmenu-item :zconfi (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Confidence Interval Using z Statistic")
  (eval-enqueue '(confidence-interval-data-dialog-make 'clasp::confidence-interval-z 
                  (make-table-maker *confidence-labels* :table-type :row-table
                                    :name-format (format nil "~A Z Confidence Interval on ~A" "~5F" "~:(~A~)")
                                    :name-args '(CONFIDENCE X-ARGS)) "Confidence Interval Using Z Statistic"
                  :help '*confi-z-help*)))
(defmenu-item :tconfi (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Confidence Interval Using t Statistic")
  (eval-enqueue '(confidence-interval-data-dialog-make 'clasp::confidence-interval-t 
                  'confidence-t-formatter "Confidence Interval Using T Statistic"
                  :help '*confi-t-help*)))
(defmenu-item :pconfi (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Confidence Interval of a Proportion")
  (eval-enqueue '(confidence-prop-dialog-make)))
(defmenu-item :ttest1 (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "t-test One Sample")
  (eval-enqueue '(t-test-one-sample-dialog-make)))
(defmenu-item :ttest2 (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "t-test Two Samples")
  (eval-enqueue '(t-test-dialog-make)))
(defmenu-item :ttestm (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "t-test Matched Pairs")
  (eval-enqueue '(t-test-matched-dialog-make)))
(defmenu-item :dtest (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "D test")
  (eval-enqueue '(d-test-dialog-make)))
(defmenu-item :contrasts (:test 'clasp-window-menu-item :menu-item-title "Analysis of Contrasts")
  (eval-enqueue `(make-instance 'contrast-toolbox-dialog)))
(defmenu-item :anova-new (:test 'clasp-window-menu-item :menu-item-title "Analysis of Variance")
  (eval-enqueue `(make-instance 'anova-toolbox-dialog)))
#+REPLACED-AS-OF-1.4.5
(defmenu-item :anova-ow (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Anova - One Way")
  (eval-enqueue '(anova-one-way-dialog-make)))
#+REPLACED-AS-OF-1.4.5
(defmenu-item :anova-tw (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Anova - Two Way")
  (eval-enqueue '(anova-two-way-dialog-make)))
(defmenu-item :chicount (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Chi-Square Counts")
  (eval-enqueue '(chi-square-2x2-counts-dialog-make)))
(defmenu-item :chi2x2 (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Chi-Square 2x2")
  (eval-enqueue `(two-variable-dialog-make 'clasp::chi-square-2x2 "Chi-Square 2x2"
                                           'chi-2x2-format :help '*chi-2x2-help*)))
(defmenu-item :chiRxC (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Chi-Square RxC")
  (eval-enqueue `(two-variable-dialog-make 'clasp::chi-square-rxc "Chi-Square RxC"
                                           'chi-rxc-format :help '*chi-rxc-help*)))
(defmenu-item :3danalyze (:test 'clasp-window-menu-item :menu-item-title "3d Contingency Table Analysis")
  (eval-enqueue `(make-instance 'analyze-3d-dialog)))
(defmenu-item :regrbrief (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Linear Regression - brief")
  (eval-enqueue '(two-variable-dialog-make 'clasp::linear-regression-brief "Linear Regression - brief"
                  (make-table-maker *brief-lr-labels* :table-type :column-table
                                    :name-format "Linear Regression of ~:(~A~) against ~:(~A~)"
                                    :name-args '(X-ARGS Y-ARGS))
                  :help '*brief-lr-help* :categorical-x NIL :categorical-y NIL)))
(defmenu-item :regrverbose (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Linear Regression - verbose")
  (eval-enqueue '(two-variable-dialog-make 'clasp::linear-regression-verbose
                  "Linear Regression - verbose" 'verbose-lr-formatter
                  :help '*verbose-lr-help* :categorical-x NIL :categorical-y NIL)))
(defmenu-item :multiregr (:test 'clasp-window-menu-item :MENU-ITEM-TITLE "Multiple Linear Regression - verbose")
  (eval-enqueue '(multi-linear-regr-dialog-make)))

;;; ================================================================
;;; sample menu

(defmenu nil :sample 'MENU "Sample" "Commands that make datasets.")
(defmenu-item :uniform (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Uniform Distribution")
  (eval-enqueue '(uniform-dist-dialog-make)))
(defmenu-item :normal (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Normal Distribution")
  (eval-enqueue '(normal-dist-dialog-make)))
(defmenu-item :binomial (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Binomial Distribution")
  (eval-enqueue '(binomial-dist-dialog-make)))
(defmenu-item :poisson (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Poisson Distribution")
  (eval-enqueue '(poisson-dist-dialog-make)))
(defmenu-item :gamma (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Gamma Distribution")
  (eval-enqueue '(gamma-dist-dialog-make)))
(defmenu-item :monte-carlo (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Monte Carlo Sampling")
  (eval-enqueue '(monte-carlo-dialog-make)))
(defmenu-item :bootstrap (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Bootstrap/Randomize")
  (eval-enqueue '(data-sampling-dialog-make)))
(defmenu-item :mixed-resample (:sample 'clasp-window-menu-item :MENU-ITEM-TITLE "Mixed Resampling")
  (eval-enqueue `(bootstrap-dialog-make)))

;;; ================================================================
;;; windows menu
;;; THIS MENU USES THE FUNCTION FORMAT. THAT MAKES IT HARD TO PATCH, BUT HEY...

(defclass clasp-windows-menu (menu)
  ())

(defmethod menu-update ((item clasp-windows-menu))
  (let ((front (front-window))
        (windows (windows)))
    (apply #'remove-menu-items *clasp-windows-menu* (menu-items *clasp-windows-menu*))
    (apply #'add-menu-items *clasp-windows-menu* 
           (append *clasp-windows-menu-default-items*
                   (loop for w in windows
                         collect (make-instance 'ccl::windows-menu-menu-item
                                   :disabled (eq w front)
                                   :command-key (when (eq w *top-listener*) #\L) 
                                   :menu-item-title (ccl::window-title w)
                                   :window w))))))

(defvar *clasp-windows-menu*)
(defvar *clasp-windows-menu-default-items*)
(defvar *rw-menu-item* nil) ; stores the ribbon window menu item so I can check it's box

(defun make-windows-menu ()
  (setq *clasp-windows-menu*
        (make-instance 'clasp-windows-menu :MENU-TITLE "Windows"
                       :help-spec "Window Menu
Use this select from the various windows available in the CLASP interface."
                       :MENU-ITEMS
                       (setf *clasp-windows-menu-default-items*
                             (list
                              ;; FIX: Put following somewhere else?
                              (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "CLASP Window"
                                             :help-spec "Selects the current CLASP window or creates a new one if there isn't one."
                                             :command-key #\1
                                             :menu-item-action
                                             #'(lambda ()
                                                 (eval-enqueue '(select-current-or-new-clasp-window))))
                              (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "CLASP Variable Palette"
                                             :help-spec "Selects the CLASP variable dock or creates one if there isn't one."
                                             :command-key #\2
                                             :menu-item-action
                                             #'(lambda ()
                                                 (eval-enqueue '(find-clasp-dock :force-access t))))
                              (make-instance 'menu-item :menu-item-title "CLASP Notebook"
                                             :help-spec "Selects the CLASP Notebook or opens a new one if there isn't one."
                                             :command-key #\3
                                             :menu-item-action
                                             #+DEMO 'sorry-demo-version
                                             #-DEMO 'open-clasp-notebook)
                              (make-instance 'menu-item :menu-item-title "Graph Palette"
                                             :help-spec "a list of generated graphs"
                                             :command-key #\4
                                             :menu-item-action 'open-graph-caddy)
                              (make-instance 'menu-item :menu-item-title "Table Palette"
                                             :help-spec "a list of generated tables"
                                             :command-key #\5
                                             :menu-item-action 'open-table-caddy)
                              (setf *rw-menu-item* 
                                    (make-instance 'menu-item :menu-item-title "Ribbon Window"
                                                   :help-spec "is the ribbon window open ?"
                                                   :command-key #\6
                                                   :menu-item-checked *use-ribbon-window*
                                                   :menu-item-action
                                                   #'(lambda ()
                                                       (set-menu-item-check-mark *rw-menu-item*
                                                                                 (setf *use-ribbon-window* (not *use-ribbon-window*)))
                                                       (if *use-ribbon-window*
                                                         (open-ribbon)
                                                         (ignore-errors (window-close (find-window "gibbon-monkey")))))))
                              (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T))))))
  
(defmenu-raw #'make-windows-menu)

;;; ================================================================
;;; debug menu

#+CLASP-MCL-INTERFACE-DEBUG
(defmenu NIL :debug 'MENU "Debug" "Debug Menu
This menu should be only be enabled in the development version.")
#+CLASP-MCL-INTERFACE-DEBUG
(defmenu-item :clasp-menu (:debug 'MENU-ITEM :MENU-ITEM-TITLE "Install CLASP Menu")
  (install-clasp-menu))
#+CLASP-MCL-INTERFACE-DEBUG
(defmenu-item :clasp-menubar (:debug 'MENU-ITEM :MENU-ITEM-TITLE "Install CLASP Menubar")
  (install-clasp-menubar))
#+CLASP-MCL-INTERFACE-DEBUG
(defmenu-item :barD1 (:debug 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T) NIL)
#+CLASP-MCL-INTERFACE-DEBUG
(defmenu-item :defaultmenu (:debug 'MENU-ITEM :MENU-ITEM-TITLE "Install Default Menubar")
  (install-default-menubar))

;;; =================================================================
;;; makes some menu items for the apple menu

(defun l-menu-item-clasp-apple-menu ()
  "Returns a list of CLASP menu items that are installed on the apple
menu, including the separator. See note in clasp-menubar-make
concerning binding this to a global variable."
  ;;
  (LIST (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "About CLASP"
                       :menu-item-action
                       'display-about-box)
        (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "Help"
                       :COMMAND-KEY #\?
                       :menu-item-action
                       #'(lambda ()
                           (display-help (front-window))))
        ;;
        (MAKE-INSTANCE 'MENU-ITEM :MENU-ITEM-TITLE "-" :DISABLED T)
        ;;
        ))

;;; =================================================================
;;; replacements for the old menu functions

(defun install-clasp-menu ()
  (install-default-menubar t)
  (install-as-menu "CLASP" *menu-defs* (l-menu-item-clasp-apple-menu)))

(defun install-clasp-menubar ()
  (LET ((APPLE-MENU *APPLE-MENU*))
    (APPLY #'REMOVE-MENU-ITEMS APPLE-MENU (MENU-ITEMS APPLE-MENU))
    (APPLY #'ADD-MENU-ITEMS APPLE-MENU (l-menu-item-clasp-apple-menu)))
  (install-menubar *menu-defs*))

(defun install-default-menubar (&optional no-query)
  (if no-query
    (set-menubar *default-menubar*)
    (eval-enqueue '(progn
                     (when (yes-or-no-p "Do you really want to install a menubar without an CLASP menus?")
                       (format *query-io* "To reinstall the CLASP menubar use (clasp-interface::install-clasp-menubar).")
                       (set-menubar *default-menubar*))))))

;;; =================================================================

(install-clasp-menubar)