;;; -*- Mode: LISP; Syntax: Common-Lisp; Base: 10; Package:CLASP-INTERFACE -*-

;;; Copyright (c) 1990 - 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.

;;; For more information write to clasp-support@cs.umass.edu

(in-package :CLASP-INTERFACE)

;;;----------------------------------------------------------------------------

(defclass textual-results-view (textual-dialog-view) ())

;; Should be a constant, but that did not work...
(defparameter +textual-results-view+ (make-instance 'textual-results-view))

;;;----------------------------------------------------------------------------

(defclass window-mixin ()
  ((frame :initform nil :initarg :frame :accessor display-frame)
   (stream :initform nil :initarg :stream :accessor display-stream)
   (window-open-p :initform nil :initarg :window-open-p :accessor window-open-p)))

(defmethod window-open-p ((self t))
  nil)

(defmethod (setf display-frame) :after (frame (self window-mixin))
  (setf (display-stream self) (frame-standard-output frame)))

(defmethod (setf display-stream) :after (stream (self window-mixin))
  (setf (window-open-p self) (not (null stream))))

(defconstant *open-result-text-style* (make-text-style nil :italic nil))
(defconstant *closed-result-text-style* (make-text-style nil nil nil))

(defmethod presentation-text-style ((object window-mixin))
  (if (window-open-p object)
      *open-result-text-style*
    *closed-result-text-style*))

(defmethod presentation-text-style ((object t))
  *closed-result-text-style*)

;;;----------------------------------------------------------------------------

(defclass modifiable-display-option-mixin (window-mixin)
  ((display-option :initarg :display-option
		   :type (member :notebook :window :icon :none)
		   :accessor display-option
		   :allocation :class
		   :initform :notebook)))

(defmethod modifiable-display-option-p ((class t))
  nil)

(defmethod modifiable-display-option-p
    ((class modifiable-display-option-mixin))
  t)

(defmethod display-option ((self t))
  :notebook)

;;;----------------------------------------------------------------------------

(defclass save-before-quit-mixin ()
	  ((modified-p :initarg :modified-p
		       :type (member t nil)
		       :accessor modified-p
		       :initform t)
	   (save-pathname :initarg :save-pathname
			  :type string
			  :accessor save-pathname
			  :initform "")))

(defmethod save-before-quit-p ((self t)) nil)
(defmethod save-before-quit-p ((self save-before-quit-mixin)) t)

(defmethod modified-p ((self t))
  nil)

(defmethod save :after ((self save-before-quit-mixin) pathname)
  (setf (modified-p self) nil)
  (setf (save-pathname self) pathname))

;;;----------------------------------------------------------------------------

(defclass result (window-mixin #+OLD icon clasp::data
		  #+ignore clasp::named-object
		  #+ignore clasp::remember-instances)
  ())

(defmethod display-option ((self result))
  :notebook)

(defmethod show-in-results-display ((self result))
  t)

(defmethod initialize-instance :after ((self result) &key)
  (unless (and (slot-boundp self 'name)
	       (> (length (name self)) 0))
    (setf (name self) (format nil "~a" (name self)))))

(defclass dataset (result clasp::dataset save-before-quit-mixin)
   ())

(defmethod display-option ((self dataset))
  :icon)

(defmethod show-in-results-display ((self dataset))
  nil)

(defclass numbers (result)
  ())

(defclass column (numbers clasp::variable)
  ())

#|

A problem arises because CLIM needs to reparse the input. When you
click on items to enter arguments CLIM puts a special marker in the
input-editor buffer so that when it reparses the input it gets hold of
the actual object. Unfortunately this is all done at the stream layer
in the input editor code. This is OK as long as you're clicking on
presentations that represent complete arguments but if you have an
argument that consists of multiple presentations eg a sequence of
items then this breaks badly. This is because by the time it gets down
to the accept-method for a sequence of items it doesn't know about
these special markers in the input stream.

So we have this workaround where you define textual based accept and
present methods for all your datatypes. This way when you click on an
item as an argument it inserts into the input buffer a string of
characters which when accept'ed will get you back the original
object.

We only define these for `column' for now because they are the only things
accepted in this problematic way (eg., the `histogram' command).

|#

(define-presentation-method accept ((type clasp::variable) stream (view textual-view) &key)
  (let ((dataset (with-delimiter-gestures (#\!)
		   (read-token stream)))
	(separator (read-char stream))
	(variable (read-token stream)))
    (declare (ignore separator))
    ;(spy dataset variable)
    (when (setf dataset (get-dataset dataset t nil))
      (get-variable variable dataset t nil))))

(define-presentation-method present (x (type clasp::variable) stream (view textual-view) &key)
  (when x ; bullet proof
    (format stream "~a!~a"  (name (clasp::variable-dataset x)) (name x))))

(define-presentation-method present (x (type result) stream (view textual-results-view) &key)
  (with-output-as-presentation
      (stream x (type-of x) :single-box t)
    (with-text-style (stream (merge-text-styles (presentation-text-style x) (medium-merged-text-style stream)))
      (write-string (name x) stream))))

(defclass row-number-column (column row-number-variable)
  ())

(defmethod display-option ((self column))
  :icon)

(defmethod show-in-results-display ((self column))
  nil)

(defclass number-sequence (numbers)
  ((value :initarg :value :accessor value)))

(defclass graph-data-icon (modifiable-display-option-mixin result)
	  ((display-option :allocation :class :initform :notebook)
	   (graph-data :initarg :graph-data :reader graph-data)
           (x-label :initform nil :initarg :x-label :reader x-label)
           (y-label :initform nil :initarg :y-label :reader y-label)
	   (graph :initarg :graph :initform nil :reader graph)))

(defclass graph (result graph:annotated-graph) ())

(defmethod show-in-results-display ((self graph))
  nil)

(defmethod display-option ((self graph))
  nil)

(defclass window-point (standard-point)
	  ((window :initarg :window :accessor window)))

