;;; -*- Mode: LISP; Syntax: Common-Lisp; Base: 10; Package:CLASP-INTERFACE -*-

(in-package :CLASP-INTERFACE)

;; ----------------------------------------------------------------------------
;;; The FILE menu
;; (more to come!)

(define-command-table clasp-file)

(add-menu-item-to-command-table 'clasp "File" :menu 'clasp-file :errorp nil
				:documentation "I/O and customization commands"
				:after :start)

#+ignore
(define-command-table clasp-map)

#+ignore
(define-command (com-help-me :command-table clasp-file :name t
			     :menu  "Help Menu") ()) 

(define-command (com-load-dataset
		 :command-table clasp-file :name t
		 :menu ("Load Dataset" 
			:documentation
			"Load a dataset from a CLASP format file"))
    ((pathname 'pathname :prompt "Pathname"))
  (let (dataset-result)
    (fresh-line)
    (formatting-item-list
     (*standard-output*
      :max-height 60)
     (formatting-cell
      (*standard-output*)
      (present
       (setf dataset-result (load-dataset pathname nil))
       'dataset
       :stream *standard-output*
       :view +dialog-view+)))
    (terpri)))

(define-command (com-import-dataset
		 :command-table clasp-file :name t
		 :menu ("Import Dataset"
			:documentation
			"Load a dataset from an ascii file"))
    ((pathname 'pathname :prompt "Pathname")
     (separator 'character :prompt "Separator Character")
     (include-labels-p 'boolean
		       :prompt "Include Variable Names [Yes or No]"))
  (let (dataset-result)
    (fresh-line)
    (formatting-item-list
     (*standard-output*
      :max-height 60)
     (formatting-cell
      (*standard-output*)
      (present
       (setf dataset-result
	 (import-dataset pathname
			 :separator separator
			 :include-labels-p include-labels-p))
     'dataset
     :stream *standard-output*
     :view +dialog-view+)))
    (terpri)
    (clim:redisplay-frame-pane clim:*application-frame* 'datasets :force-p t)
    (draw-datasets-window)
    (auto-open dataset-result)))
	       
(define-command (com-save-dataset :command-table clasp-file :name t
				  :menu
				  ("Save Dataset"
				   :documentation
				   "Save a dataset to a CLASP format file"))
    ((dataset 'dataset :prompt "Dataset")
     (pathname 'pathname :prompt "Pathname"))
  (save-dataset dataset pathname))

(define-command (com-export-dataset :command-table clasp-file :name t
				    :menu ("Export Dataset"
					   :documentation
					   "Save a dataset to an ascii file"))
    ((dataset 'dataset :prompt "Dataset")
     (pathname 'pathname :prompt "Pathname")
     (separator 'character :prompt "Separator Character")
     (include-labels-p 'boolean
		       :prompt "Include Variable Names [Yes or No]"))
  (export-dataset dataset pathname
		  :separator separator
		  :include-labels-p include-labels-p))

#+ignore
(define-clasp-command (com-set-default-directory :command-table clasp-file
						 :name t
						 :menu "Set Default Directory")
    ((directory 'string :prompt "Directory")
     (type 'string :prompt "Extension"))
  (setf clasp::*data-file-defaults* (make-pathname :host "CLASP-DATA"
						   :directory directory
						   :type type)))

(define-command (com-kill-everything :command-table clasp-file
				     :name t
				     :menu ("Kill Everything"
					    :documentation
					    "Clear All Data"))
    ()
  (clasp::forget-instances 'dataset)
  (clasp::forget-instances 'result t)
  (dolist (result (clasp::find-instances 'result))
	  (when (and (slot-boundp result 'display-window)
		     (not (null (display-window result))))
		(setf (window-visibility (display-window result)) nil)))
  ;; Handle the clasp variables as well.
  (setf clasp::*clasp-active-datasets* nil)
  (setf clasp::*clasp-datasets* nil)
  (clim:redisplay-frame-panes clim:*application-frame* :force-p t)
  (draw-datasets-window)
  (draw-results-window)
  (window-clear *clasp-frame-main-window*))

(define-command (com-clear-desktop :command-table clasp-file :name t
				   :menu ("Clear Desktop"
					  :documentation
					  "Clear the interactor pane"))
    ()
  (window-clear *clasp-frame-main-window*))

(define-clasp-command (com-print :command-table clasp-file :name t
				 :menu
				 ("Print"
				  :documentation
				  "Print a CLASP object to a postscript file"))
    ((result 'result :prompt "Print what?")
     (pathname 'pathname :prompt "Filename for Postscript File"))
  (hardcopy result pathname))

(defmethod hardcopy ((the-result result) filename)
  (with-open-file (file-stream filename :direction :output
		   :if-exists :overwrite
		   :if-does-not-exist :create)
    (clim:with-output-to-postscript-stream
	(postscript-stream file-stream :multi-page t)
      (display the-result postscript-stream))))

(define-command (com-preferences :command-table clasp-file
				 :name t
				 :menu ("Preferences"
					:documentation
					"Customize display"))
    ()
  (let ((auto-open-types (auto-open-types *application-frame*))
	(datasets-and-results-display (datasets-and-results-display
				       *application-frame*)))
    (with-inferior-stream
	(stream *application-frame* :label "Preferences")
      (window-expose stream)
      (restart-case
	  (progn
	    (accepting-values
	     (stream)
	     (setf auto-open-types
	       (accept
		(list 'subset-sequence
		      (remove-if #'(lambda (x)
				     (or (eq x 'result)
					 (eq x 'graph)))
				 *result-types*))
		:stream stream
		:prompt "What types of results should open automatically?"
		:default auto-open-types))
	       (terpri stream)
	       (setf datasets-and-results-display
		 (accept '(member #-MCL :window :pane :none) :stream stream
			 :prompt "Display datasets and results"
			 :default datasets-and-results-display))
	       (terpri stream))
	    (setf (auto-open-types *application-frame*)
	      auto-open-types)
	    (setf (datasets-and-results-display *application-frame*)
	      datasets-and-results-display)
	    (setf (window-visibility stream) nil))
	(abort () (setf (window-visibility stream) nil))))
    (cond
     ((eq (datasets-and-results-display *application-frame*) :none)
      (setf (window-visibility (datasets-window *application-frame*)) nil)
      (setf (window-visibility (results-window *application-frame*)) nil)
      (set-frame-layout *application-frame* 'datasets-and-results-not-in-pane))
     ((eq (datasets-and-results-display *application-frame*) :window)
      (setf (window-visibility (datasets-window *application-frame*)) t)
      (setf (window-visibility (results-window *application-frame*)) t)
      (draw-datasets-window)
      (draw-results-window)
      (set-frame-layout *application-frame* 'datasets-and-results-not-in-pane))
     (t
      (setf (window-visibility (datasets-window *application-frame*)) nil)
      (setf (window-visibility (results-window *application-frame*)) nil)
      (set-frame-layout *application-frame* 'datasets-and-results-in-pane)))))

(define-command (com-quit-clasp :command-table clasp-file :name t
				:menu ("Quit"
				       :documentation
				       "Leave CLASP"))
   ()
  (setf (window-visibility (datasets-window *application-frame*)) nil)
  (setf (window-visibility (results-window *application-frame*)) nil)
  (dolist (window (inferiors *application-frame*))
    (setf (window-visibility window) nil))
  (dolist (result (clasp::find-instances 'result))
    (when (and (slot-boundp result 'display-window)
	       (not (null (display-window result))))
      (setf (window-visibility (display-window result)) nil)))
  (frame-exit *application-frame*))

#+documentation
(define-clasp-command (com-print-menus :command-table clasp :name t)
    ()
  (do-command-table-inheritance
      (command-table (find-command-table 'clasp))
    (let ((name (format nil "~A" (command-table-name command-table))))
      (when (search "CLASP-" name)
	(with-open-file (file-stream (concatenate 'string name "-menu.ps")
			 :direction :output :if-exists :overwrite)
	  (with-output-to-postscript-stream (stream file-stream)
	    (display-command-table-menu command-table stream)))))))

#+documentation
(define-clasp-command (com-print-interactor :command-table clasp :name t)
    ()
  (with-open-file (file-stream "interactor-pane.ps"
		   :direction :output :if-exists :overwrite
		   :if-does-not-exist :create)
    (with-output-to-postscript-stream (stream file-stream)
      (let ((*standard-output* stream))
	(redisplay-frame-pane *application-frame* 'interactor :force-p t)))))

;;; ----------------------------------------------------------------------------

#| 
(define-command-table clasp-show)

(add-menu-item-to-command-table 'clasp "Show" :menu 'clasp-show :errorp nil :after "Plotting")

(define-command (COM-PRINT-ALL-DATA :command-table clasp-show :name t :menu  "Show All Data") ()
  (print-all-data))

(define-command (com-print-data-multiple-columns :command-table clasp-show :name t :menu  "Show Data Multiple Columns") ()
  "Displays the values for the selected variable."
  (let ((data-set (current-data-set *application-frame*)))
		(dolist (var (get-variable :data-set data-set :get-multiple t :label "Select variables to view."))
		  (print-column-data (column-number var) :data-set data-set))))

(define-command (com-print-data-row :command-table clasp-show :name t :menu  "Show All Data by Row") ()
  "Display all values for a selected Dataset one row at a time."
  (if (#+MCL ccl::y-or-n-dialog #-MCL y-or-n-p "Show just the current dataset?")
    (let ((d-set (current-data-set *application-frame*)))
      (dotimes (i (array-dimension (data-set-data-array d-set) 0))
        (unless (print-row-data i :data-set d-set) (return))))
    (catch 'up 
      (dolist (set *data-set-list*)
        (let ((d-set (eval set)))
          (dotimes (i (array-dimension (data-set-data-array d-set) 0))
            (unless (print-row-data i :data-set d-set) 
              (throw 'up nil))))))))

(define-command (com-print-variable-column-and-data-set :command-table clasp-show :name t :menu  "Show Variable Stuff") ()
  "Display the column number and data-set title of a variable."
  (multiple-value-bind (num set)
    (column-number (caar (get-multi-set-data)))
		(format *clasp-report-stream* "~%Column number: ~d, dataset: ~a~%"
			num set)))

(define-command (print-variable-symbols-and-names :command-table clasp-show :name t :menu  "Show Variable Symbols and Names") ()
  "Display all internal symbols and variable name strings for a selected dataset, or all datasets."
  (if (#+MCL ccl::y-or-n-dialog #-MCL y-or-n-p "Show just the current dataset?")
		  (print-variable-symbol-name 
		    :data-set (current-data-set *application-frame*))
		  (print-variable-symbol-name)))

|#
;;; ----------------------------------------------------------------------------
