;;; -*- Syntax: Common-lisp; Package: GRAPH -*-
;;;
;;; Copyright (c) 1990 - 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.

;;; For more information write to clasp-support@cs.umass.edu

(in-package :graph)

;; From AXIS.LISP
;; 11-21-93 Lispworks does not support making strings using `make-array' with
;; a :ELEMENT-TYPE of 'character - must use 'base-character.  I have
;; sent them a bug report about this.  (Westy)

(defun make-adjustable-string (&optional (size 7))
  ;; (internal to float-to-string)
  ;; Cons a new string every time; if you reuse an old string, you get graphics
  ;; turds.
  (make-array size
	      :element-type
	      ; Westy
	      #+lispworks 'base-character
	      #+lucid 'string-char
	      #-(or lucid lispworks) 'character
	      :adjustable t
	      :fill-pointer 0))


;; FROM BASIC-GRAPH.LISP
;; These next two functions previously were using the `float' function 
;; to produce single-floats (see top p. 350 CLTL2) and then declaring 
;; them to be short-floats (and also using `the' forms which is where
;; Lispworks complained). I hacked all the declarations to be
;; single-float, but this is prpbably not what the author intended. (Westy)

(defmethod xy-to-uv-transform ((self basic-graph-coordinates-mixin))
  ;; Carefully optimized to death.  Calling the closure should not
  ;; cons any floats, unless the inputs X and Y are not floats.
  (let* ((xll       (slot-value self 'xll))
	 (yll       (slot-value self 'yll))
	 (ull       (slot-value self 'ull))
	 (vll       (slot-value self 'vll))
	 (x-u-scale (float (slot-value self 'x-u-scale)))
	 (y-v-scale (float (slot-value self 'y-v-scale)))
	 (temp1     (float (- ull (* xll x-u-scale))))
	 (temp2     (float (- vll (* yll y-v-scale)))))
    ;; 1-19-94 Changed all the `short-float's to `single-float' - Westy
    (declare (single-float x-u-scale temp1 y-v-scale temp2))
    #'(lambda (x y)
        (values (truncate
		 (the (single-float -1000000.0 1000000.0)
		   (+ temp1 (* (the single-float (float x))
			       x-u-scale))))
		(truncate
		 (the (single-float -1000000.0 1000000.0)
		   (+ temp2 (* (the single-float (float y))
			       y-v-scale))))))))

(defmethod xy-to-stream-transform ((self basic-graph-coordinates-mixin)
				   stream)
  (let* ((xll       (slot-value self 'xll))
	 (yll       (slot-value self 'yll))
	 (ull       (slot-value self 'ull))
	 (vll       (slot-value self 'vll))
	 (x-u-scale (float (slot-value self 'x-u-scale)))
	 (y-v-scale (float (slot-value self 'y-v-scale)))
	 (height (the fixnum (sheet-inside-height stream)))
	 (temp1     (float (- ull (* xll x-u-scale))))
	 (temp2     (float (- vll (* yll y-v-scale)))))
    ;; 1-19-94 Changed all the `short-float's to `single-float' - Westy
    (declare (single-float x-u-scale temp1 y-v-scale temp2))
    (setq y-v-scale (- y-v-scale))
    (setq temp2 (- height temp2))
    #'(lambda (x y)
        (values (truncate
		 (the (single-float -1000000.0 1000000.0)
		      (+ temp1 (* (the single-float (float x))
				  x-u-scale))))
		(truncate
		 (the (single-float -1000000.0 1000000.0)
		      (+ temp2 (* (the single-float (float y))
				  y-v-scale))))))))

;; From POPUP-ACCEPT.LISP
;; 11-4-93 Added a Lispworks conditionalization similar to the previously
;; added one for MCL for MOPish functions that are not exported from
;; the CL package. (Westy)

(defmethod popup-accept-variables ((self standard-object))
    "Get list of instance variables."
    (let ((slots 
	   ;; Blech.  What's wrong with my package?
           #FEATURE-CASE
           (((not (or :mcl :lispworks))
             (mapcar #'slot-definition-name (class-slots (class-of self))))
            ;; 11-4-93 Westy
            (:lispworks
	     (mapcar #'clos:slot-definition-name
		     (clos:class-slots (class-of self))))
            (:mcl
             (mapcar #'ccl:slot-definition-name
		     (ccl:class-class-slots (class-of self)))))
           ))
      (sort slots #'string-lessp)))

;;; Changes to the default color of graphs. Also allow CLIM colors.

;; From GRAPH-DATA.LISP

(defclass GRAPH-DATA-COLOR-MIXIN ()
    ((color :initform :royal-blue :initarg :color :accessor color)))

;; From COLORS.LISP
;; 3-17-93

(defvar *color-name-to-color-table* (make-hash-table))
  
(defun ALU-FOR-COLOR (color-name)
  "Translate a color name to an ink/alu"
  (or *colors* (initialize-color-system))
  (or (gethash color-name *color-name-to-color-table* nil)
      ;; 3-17-94 Allow CLIM color objects (Westy)
      (when (colorp color-name)
	    color-name)
      ;; 3-17-94 Allow CLIM color names, ie '+green+ (Westy)
      (and (symbolp color-name)
           (when (char= (char (symbol-name color-name) 0) #\+)
	     (symbol-value (intern (symbol-name color-name) (find-package 'clim)))))
      (gethash :White *color-name-to-color-table* nil)))

(defun MAKE-COLOR-CHOICES (color-specifications &optional reinitialize)
  "Makes an Alist of colors to choose from."
  (if reinitialize (setq *color-name-to-color-table* (make-hash-table)))
  (cond ((color-exists-p)
	 (loop for (name red green blue) in color-specifications
	       do
	       (setf (gethash name *color-name-to-color-table*) (make-color-rgb red green blue))
	       (pushnew (list (string-capitalize name) :value name)
			*colors*
			:test #'equal)))
	#+genera
	(t
	 (pushnew (list "White" tv:alu-ior) *colors*)
	 (pushnew (list "Black" tv:alu-andca) *colors*)
	 (setf (gethash :white *color-name-to-color-table*) tv:alu-ior)
	 (setf (gethash :black *color-name-to-color-table*) tv:alu-andca))))

;;; ***************************************************************************
;;; EOF
