;;;; -*- Mode:Common-Lisp; Package:EKSL-UTILITIES; Fonts:(MEDFNT); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/clos-mixins.lisp *-*
;;;; *-* Last-Edit: Wednesday, November 1, 1995  15:53:11; Edited-By: Westy *-*
;;;; *-* Machine: Earhart (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                   Underlying Basic Class Mixins                        *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Copyright (c) 1995 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  01-19-93 File Created.  (Westy)
;;;  05-12-95 Added to eksl-utilities system.  (Westy)
;;;  05-16-95 Changed Explorer methods `class-prototype-p' and `print-method'
;;;           to specialize on `named-object-mixin' instead of `basic-object.'
;;;           (anderson)
;;;  05-23-95 Changed `find-instance-by-name' to use a more efficient. Also
;;;           removed `forget-instance' and `forget-instances' from the
;;;           exported protocol.
;;;  06-28-95 Added error checking to `find-instance-by-name.'  (Anderson)
;;;  06-30-95 Copied `remember-original-values-mixin' from Phoenix.  (Anderson)
;;;  07-03-95 Incompatible change to `find-instance-by-name'.  The previous 
;;;           implementation allowed the user to specify which class to look 
;;;           it up under, which would fail unless they guessed right.  
;;;           Perhaps an improved representation will allow this behavior.   
;;;           (Anderson)
;;;  07-20-95 Added an "improved representation" for `find-instance-by-name'.
;;;           (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(name name-string named-object-mixin remember-instances
               object-with-properties persistent-instances find-direct-instances
               find-instances find-instance-by-name kill reset
               kill-all-instances kill-all-remembered-instances get-value
               remember-original-values-mixin restore-original-values))

;;; --*--
;;; ***************************************************************************

(defgeneric name (object))

(defmethod name ((the-object t))
  (format nil "~a" the-object))

;;;----------------------------------------------------------------------------
;;; NAMED-OBJECTS have a NAME and a DESCRIPTION slot. They store themselves on the
;;; :OBJECT property of their name so that they can be accessed by the NAME.
;;; The preceding comment is not true.  SDA 7/3/95

;;; Note that the names of named objects define a name-space. 
;;; They also store themselves under the class-name property of their name symbol.

(define-debugging-class named-object-mixin ())

(defclass named-object-mixin ()
     ((name :type symbol :initarg :name :reader name))
     (:documentation "Allows each instance to have an name. One is generated
for it if not provided. The name is always a symbol.
`named-objects' store themselves, i.e., their instances, under
the (class-name (class-of the-object) property of their name symbol."))

(defmethod print-object ((object named-object-mixin) stream)
  (let ((name (and (slot-boundp object 'name) (slot-value object 'name))))
    (if *print-escape*
	(format stream "#.(~a ~:[<unnamed>~;'~s~])" (class-name (class-of object)) name name)
        (format stream "<~a ~:[(unnamed)~;~a~]>" (class-name (class-of object)) name name))))
  
(defmethod make-name ((object named-object-mixin) &optional new-name)
  "Make a name for yourself if necessary. This version insures name is a symbol."
  (let ((class-name (class-name (class-of object))))
    (macrolet ((form-name-symbol (&rest strings)
                                 `(form-symbol-in-package *package* ,@strings)))
      (cond ((not new-name) (form-name-symbol
			     (string-upcase class-name)
			     "-"
			     (princ-to-string
			      (setf (get class-name 'name-index)
				    (1+ (or (get class-name 'name-index) 0))))))
	    ((symbolp new-name) new-name)
	    ((stringp new-name) (form-name-symbol new-name))
	    (t (form-name-symbol (princ-to-string new-name)))))))

(defmethod initialize-instance :around ((object named-object-mixin) &rest initargs &key name)
  (apply #'call-next-method object :name (make-name object name) initargs)
  #+OLD
  (cond (name
	 (when-debugging-format named-object-mixin
	   "Using user-supplied name ~s for object of type ~s."
	   name (class-name (class-of object)))
	 (call-next-method))
	(t
	 (setf name (make-name object name))
	 (when-debugging-format named-object-mixin
	   "Using created name ~s for object of type ~s."
	   name (class-name (class-of object)))
	 (apply #'call-next-method object :name name initargs))))

(defmethod initialize-instance :after ((object named-object-mixin) &key)
  ;; Put a `name' property on the object for this class.
  (when-debugging-format named-object-mixin
    "Initializing named-object: ~s of ~s to ~s"
    (name object) (class-name (class-of object)) object)
  (setf (get (name object) (class-name (class-of object))) object))

(defmethod (setf name) (new-name (object named-object-mixin))
  (setf (slot-value object 'name)
        (make-name object new-name)))

(defmethod name-string ((object t))
  "Returns name as a string."
  (string-capitalize (name object)))

(defmethod kill :after ((object named-object-mixin))
  (awhen (name object)
    (remprop it (class-name (class-of object)))))

;; Set the index back to zero if we are forgetting all the instances.
(defmethod forget-instances :after ((class standard-class)
                                    &optional do-subclasses-p)
  (declare (ignore do-subclasses-p))
  (setf (get (class-name class) 'name-index) 0))

(defmethod find-instance-by-name ((object-spec symbol) (class standard-class) &optional no-error-p)
  (when-debugging-format named-object-mixin
    "Looking up object named ~s of class ~s" object-spec (class-name class))
  (or (get object-spec (class-name class))
      (if no-error-p
	  nil
	  (error "Instance of ~s not found:  ~s" (class-name class) object-spec))))
  
(defmethod find-instance-by-name (object-spec (class symbol) &optional no-error-p)
  (find-instance-by-name object-spec (find-class class) no-error-p))

(defmethod find-instance-by-name ((object-spec string) class &optional no-error-p)
  (find-instance-by-name (intern object-spec) class no-error-p))

(defmethod find-instance-by-name ((object-spec named-object-mixin) class &optional no-error-p)
  (find-instance-by-name (name object-spec) class no-error-p))

;;;----------------------------------------------------------------------------
;;; REMEMBER INSTANCES CLASS......
;;; Previously, we remembered the instances in a class slot and accessed the set by
;;; looking at the class prototype.  However, there are too many problems with
;;; accessing the prototype, especially for classes that haven't had any instances
;;; created.  So, a much simpler approach is to use a hash table.

;;; Changed this so it does do things hierarchically.  Use find-direct-instances if
;;; you need the old behavior.

;;; Note that to get rid of an instance one must KILL IT, so that the connection,
;;; if any, between the instance and its name is broken.  If you do not
;;; kill instances properly they will come back to haunt you.

(defvar *remember-instances-table* (make-hash-table)
  "table of class names (the keys) and List of Instances of that Class.")

(defclass remember-instances ()
  ())

(defmethod initialize-instance :after ((x remember-instances) &key)
  (push x (gethash (class-name (class-of x))
		   *remember-instances-table*)))

(defmethod find-direct-instances ((class-name symbol))
  (values (gethash class-name *remember-instances-table*)))

;;; This returns me and all other instances of my class.
(defmethod find-direct-instances ((object remember-instances))
  (find-direct-instances (class-name (class-of object))))

;;; This does what we normally want...return all instances of this specific
;;; class and any classes that inherit from this.

(defmethod find-instances ((class standard-class))
  (let ((subclasses-examined-so-far nil))
    (labels ((grab-instances (class)
	       (let ((subclasses (class-direct-subclasses
                                  class)))
		 (unless (member class subclasses-examined-so-far)
		   (push class subclasses-examined-so-far)
		   (cond ((null subclasses)
			  (find-direct-instances (class-name class)))
			 (t
			  (append
			    (find-direct-instances (class-name class))
			    (loop for subclass in subclasses
				  append (grab-instances subclass))))
			 )))))
      (grab-instances class))))

(defmethod find-instances ((class-name symbol))
  (find-instances (find-class class-name)))

(defmethod find-instances ((object remember-instances))
  (find-instances (class-of object)))

;;;----------------------------------------------------------------------------
;;; `forget-instance' and `forget-instances' are not public; use `kill' and
;;; `kill-all-instances'.

(defmethod forget-instance ((object remember-instances))
  (setf (gethash (class-name (class-of object)) *remember-instances-table*)
	(remove object (gethash (class-name (class-of object)) *remember-instances-table*)
		:test #'eq)))

#|
                  ;; There was a bug (that you will not believe until you see it)
                  ;; where (eq <object-1> <object-1>) was returning nil. I didn't
                  ;; track it down, but this gets around it.
                  :test #'(lambda (a b) 
                            (eq (name a) (name b)))))))
|#

;;; Note that we remove them one-by-one unless we are killing an entire class or everything.

(defmethod forget-instances ((class-name symbol) &optional do-subclasses-p)
  (forget-instances (find-class class-name) do-subclasses-p))

(defmethod forget-instances ((class standard-class) &optional do-subclasses-p)
  (remhash (class-name class) *remember-instances-table*)
  (when do-subclasses-p
    (dolist (subclass (class-direct-subclasses class))
      (forget-instances subclass t))))

;;;----------------------------------------------------------------------------
;;; We need to know how to kill efficiently.

(defvar *killing-all-instances* nil
   "This is bound to a non-nil value when we are killing all instances of all
the classes we know about so forgetting instances can be done more efficiently.")

(defvar *killing-class* nil
  "This is bound to a class that we are killing all instances of so that
forgetting instances can be done more efficiently.")

;; Make sure we do not waste time killing things more than once.
(defvar *killed-objects-table* (make-hash-table :test #'eq :size 1000))

(defgeneric kill (object))

(defmethod kill ((object t))
  (values nil))

(defmethod kill-all-instances ((class-name symbol) &optional do-subclasses-p)
  (kill-all-instances (find-class class-name) do-subclasses-p))

(defmethod kill-all-instances ((class standard-class) &optional do-subclasses-p)
  (let ((*killing-class* class))
    (map nil #'kill (find-instances class))
    (forget-instances class do-subclasses-p)))

(defun kill-all-remembered-instances (&key even-the-persistent-ones (verbose nil) (report-stream *standard-output*))
  (when verbose
    (format report-stream "~&Killing~:[~;, with extreme predjudice,~] all instances...~%"
	    even-the-persistent-ones))
  (let ((*killing-all-instances* t))
    (maphash #'(lambda (class-name objects)
                 ;; This assumes that if we kill one of a particular class we kill them all.
                 (loop with really-killing? = nil
                       for object in objects
                       for first-time = t then nil do
                       (when (or even-the-persistent-ones
                                 (not (dont-kill-me-with-the-rest object)))
                         (setf really-killing? t)
			 (when (and first-time verbose)
			   (format report-stream "~&  of class ~a.~%"
                                   class-name))
			 (kill object))
		       finally (when really-killing?
                                 (forget-instances class-name))))
             *remember-instances-table*))
  (clrhash *killed-objects-table*)
  (values nil))

(defmethod kill :after ((object remember-instances))
  (forget-instance object))

#+UNUSED-EFFICIENCY-HACK
(defmethod kill :after ((object remember-instances))
  (unless (or *killing-all-instances*
              (class-built-on-class (class-of object) *killing-class*))
    (forget-instance object)))

(defmethod kill :around ((object t))
  (unless (gethash object *killed-objects-table*)
    (call-next-method)
    (setf (gethash object *killed-objects-table*) object)))

(defmethod reset ((object t))
  (values nil))

(defmethod reset :around ((object t))
  (unless (gethash object *killed-objects-table*)
    (call-next-method)))

(defclass persistent-instances ()
  ())

(defmethod dont-kill-me-with-the-rest ((object persistent-instances))
  t)

(defmethod dont-kill-me-with-the-rest ((object t))
  nil)

;;; ----------------------------------------------------------------------------
;;; OBJECT-WITH-PROPERTIES

(defclass object-with-properties ()
  ((property-list :initarg :property-list :initform nil)))

(defmethod get-value ((object object-with-properties) key)
  (cdr (assoc key (slot-value object 'property-list))))

(defmethod (setf get-value) (val (object object-with-properties) key)
  (with-slots (property-list) object
    (let ((cell (assoc key property-list)))
      (if cell
	  (rplacd cell val)
	  (push (cons key val) property-list))
      val)))

(defmethod initialize-instance :after ((object object-with-properties) &key property-list)
  ;; convert the supplied property list to an assoc list (for who knows what reason)
  (when property-list
    (setf (slot-value object 'property-list) nil)
    (doplist (key value property-list)
      (setf (get-value object key) value))))

;;; ----------------------------------------------------------------------------
;;; Make class prototypes print without an error. Most of my `print-object'
;;; methods assume that an object has been created using the standard
;;; initialization methods and will have the appropriate slots filled in.

#+Lispm
(defmethod class-prototype-p ((object named-object-mixin))
  (not (some #'(lambda (slot-object)
                 (and (eq (slot-definition-allocation slot-object) :INSTANCE)
                      (slot-boundp object (slot-definition-name slot-object))))
         (class-slots (class-of object)))))

#+Lispm
(defmethod print-object :around ((object named-object-mixin) stream)
;; This slot-bound check is so that class prototypes print out without error.
;; All their slots are unbound (at least on the Explorer).
  (if (class-prototype-p object)
    (format stream "~a Prototype" (class-name (class-of object)))
    (call-next-method)))


;;; ----------------------------------------------------------------------------

(defclass basic-named-object (named-object-mixin)
  ())

(defclass basic-remembered-object (remember-instances)
  ())

(defclass basic-named-remembered-object (named-object-mixin remember-instances)
  ())

(defclass basic-named-remembered-object-with-properties (named-object-mixin remember-instances)
  ())

#+test
(defclass test-object (basic-named-remembered-object-with-properties)
  ())

#+test
(defclass persistent-test-object (basic-named-remembered-object-with-properties persistent-instances)
  ())

#+test
(defun test-object-mixins ()
  (unwind-protect

      (progn
        (format t "~&Building 5 objects...~%")
	(loop repeat 5
	      collect (spy (make-instance 'test-object)))

	(format t "~&Finding all...~%")
	(spy (find-instances 'test-object))
	(format t "~&Finding one...~%")
	(spy (find-instance-by-name 'test-object-3 'test-object))
	(format t "~&Deleting one...~%")
	(spy (kill (find-instance-by-name 'test-object-3 'test-object)))
	(format t "~&Not finding that one...~%")
	(spy (find-instance-by-name 'test-object-3 'test-object t))
	(format t "~&Changing the name of one...~%")
        (spy (setf (name (find-instance-by-name 'test-object-1 'test-object t)) "Fred"))
	(format t "~&Finding all that are left...~%")
	(spy (find-instances 'test-object))
	(format t "~&Deleting all...~%")
	(spy (kill-all-instances 'test-object))
	(format t "~&Not finding another...~%")
	(spy (find-instance-by-name 'test-object-2 'test-object t))
        
	(format t "~&Building 5 more objects...~%")
	(loop repeat 5
	      collect (spy (make-instance 'test-object)))
	(format t "~&Building 5 more persistent Objects...~%")
	(loop repeat 5
	      collect (spy (make-instance 'persistent-test-object)))

        (kill-all-remembered-instances :verbose t)
	(format t "~&Finding all non-persistent ones that are left...~%")
	(spy (find-instances 'test-object))
	(format t "~&Finding all persistent ones that are left...~%")
	(spy (find-instances 'persistent-test-object))


	(format t "~&Building 1 more object...~%")
        (make-instance 'test-object :name "Sam")
	(format t "~&Finding all non-persistent ones that are left...~%")
	(spy (find-instances 'test-object))
        
	(kill-all-remembered-instances :verbose t :even-the-persistent-ones t)
	(format t "~&Finding all persistent ones that are left...~%")
	(spy (find-instances 'persistent-test-object)))
        
    (kill-all-remembered-instances :even-the-persistent-ones t)))

;;; ============================================================================

(defclass remember-original-values-mixin ()
  ((original-values :initform nil)))

(defmethod restore-original-values ((object remember-original-values-mixin))
  (doplist (key value (slot-value object 'original-values))
    (setf (slot-value object key) value)))

(defmethod initialize-instance :after ((object remember-original-values-mixin)
				       &key slots-to-remember &allow-other-keys)
  (dolist (slot slots-to-remember)
    (push (slot-value object slot) (slot-value object 'original-values))
    (push slot (slot-value object 'original-values))))

    
;;; ***************************************************************************
;;; EOF
