;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/clos-utils.lisp *-*
;;;; *-* Last-edit: Tuesday, July 4, 1995  19:12:58; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                   Utilities for Defining CLOS Stuff                    *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  03-23-95 File Created.  (anderson)
;;;  06-22-95 Extended parse-brief-slot and defclass-brief to have 
;;;           automatic accessors, initargs and conc-name  (anderson)
;;;  06-27-95 :conc-name now applies to slot-specific readers.  (anderson)
;;;  06-27-95 Implemented get-slots and set-slots.  (anderson)
;;;  07-04-95 Broke out `parse-method' from `undefmethod.'  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(defclass-brief parse-brief-slot define-after-init
	   undefclass parse-method undefmethod set-slots get-slots
           map-over-classes class-built-on-class class-slot-names))

;;; ---------------------------------------------------------------------------
;;; Macro to help in dynamically removing a method definition.  It's easy to
;;; redefine a method, but it's hard to make it go away entirely.  This uses 
;;; exactly the same syntax as `defmethod,' but removes it.
;;; ---------------------------------------------------------------------------

(eval-when #+CLtL2 (:compile-toplevel :load-toplevel)
	   #-CLtL2 (compile load)
(defun parse-method (defmethod)
  "Takes an s-expression defining a method--a list starting with `defmethod' and
returns four values: the function specifier, the qualifiers, the lambda-list and
the body.  The hard part is finding the qualifiers."
  (let ((name (cadr defmethod))
	(rest (cddr defmethod)))
    (let ((qualifiers nil))
      (do () ((listp (car rest)))
      (push (pop rest) qualifiers))
      (values name (nreverse qualifiers) (car rest) (cdr rest))))))

(defmacro undefmethod (&whole sexp spec lambda-list &body body)
  "Syntax is identical to `defmethod,' but removes the method.  Clearly, this
macro has no place in delivered code, but it can be a tremendous help during
development."
  (declare (ignore spec lambda-list body))
  (multiple-value-bind (name qualifiers lambda-list) (parse-method sexp)
    (let* ((specializers (subseq lambda-list
				 0
				 (position-if #'(lambda (elt)
						  (member elt lambda-list-keywords))
					      lambda-list)))
	   (classes      (mapcar #'(lambda (x)
			             (if (symbolp x) 't (find-class (second x))))
			         specializers)))
      `(let ((m (find-method #',name ',qualifiers ',classes)))
         ;; Querying turns out to be difficult in Lispworks, because y-or-n-p
         ;; goes to *terminal-io*, which is invisible.
         #+Explorer
         (if (y-or-n-p "delete ~s ? " m) (remove-method #',name m))
         #-Explorer
         (remove-method #',name m)))))

#+test
(undefmethod foo ((x mess::event)))

#+test
(undefmethod foo :before ((x mess::event)))

#+test
(undefmethod foo :before :sideways ((x mess::event)))

(defmacro undefclass (name supers slots &rest options)
  "Syntax is identical to `defclass,' but removes the class.  Clearly, this
macro has no place in delivered code, but it can be a tremendous help during
development."
  (declare (ignore supers slots options))
  `(setf (find-class ',name) nil))

;;; ============================================================================

(defvar *automatic-slot-accessors?* nil)
(defvar *automatic-slot-initargs?* nil)

(eval-when #+CLTL2 (:compile-toplevel :load-toplevel)
	   #-CLTL2 (compile load)
(defun parse-brief-slot
       (slot &optional
	(automatic-accessors? *automatic-slot-accessors?*)
	(automatic-initargs? *automatic-slot-initargs?*)
	conc-name)
  "Returns a verbose-style slot specification given a brief style, consisting of
a single symbol, the name of the slot, or a list of the slot name, optional
initform, optional symbol specifying whether there is an initarg, reader, or
accessor, and optional documentation string.  The specification of initarg,
reader and accessor is done by the letters I, R and A, respectively; to specify
none of those, give a symbol containing none of those letters, such as the
symbol *.  This function is used in the macro `defclass-brief,' but has been
broken out as a function in its own right for those writing variants on the
`defclass' macro.  If a verbose-style slot specification is given, it is
returned unchanged.

If `automatic-accessors?  is true, an accessor is defined, whether A is
specified or not.  If `automatic-initargs?  is true, an initarg is defined
whether I is specified or not.  If `conc-name' is specified, the accessor name
has that prepended, with a hyphen, and then the slot name."
  (etypecase slot
    (symbol (setf slot (list slot)))
    (list
     (when (intersection slot '(:initform :initarg :reader :writer :accessor :documentation))
       ;; Assume original CLOS syntax
       (return-from parse-brief-slot slot))))
  (let* ((name     (pop slot))
	 (new-slot (list name)))
    (when slot
      (let ((initform (pop slot)))
        (unless (eq initform :UNBOUND)
	  (push :initform new-slot)
	  (push initform new-slot))))
    (let ((spec (if (symbolp (car slot)) (pop slot) nil)))
      (when (find #\R (and spec (string spec)))
	(push :reader new-slot)
	(push (if conc-name
		  (form-symbol conc-name "-" name)
		  name)
	      new-slot))
      (when (or automatic-accessors? (and spec (find #\A (string spec))))
	(push :accessor new-slot)
	(push (if conc-name
		  (form-symbol conc-name "-" name)
		  name)
	      new-slot))
      (when (or automatic-initargs? (and spec (find #\I (string spec))))
	(push :initarg new-slot)
	(push (intern (string name) (find-package :keyword))
	      new-slot)))
    (when slot
      (push :documentation new-slot)
      (push (pop slot) new-slot))
    (nreverse new-slot))))

#+test
(defun test-parse-brief-slot ()
  (spy (parse-brief-slot 'foo))
  (spy (parse-brief-slot '(foo)))
  (spy (parse-brief-slot '(foo t)))
  (spy (parse-brief-slot '(foo t i)))
  (spy (parse-brief-slot '(foo t r)))
  (spy (parse-brief-slot '(foo t ia)))
  (spy (parse-brief-slot '(foo t ia "test slot")))
  (spy (parse-brief-slot '(foo t * "test slot")))
  (spy (parse-brief-slot '(foo t "slot")))
  (spy (parse-brief-slot '(bar :initform nil)))
  (spy (parse-brief-slot 'baz t t))
  (spy (parse-brief-slot 'baz t t 'class))
  (spy (parse-brief-slot '(baz nil) t t 'class))
  (spy (parse-brief-slot '(baz nil) nil t 'class))
  (spy (parse-brief-slot '(baz nil) t nil 'class))
  (spy (parse-brief-slot '(baz nil) nil nil 'class))
  (spy (parse-brief-slot '(baz nil "the baz slot") t t 'class))
  (spy (parse-brief-slot '(baz nil a) nil nil 'class))
  (spy (parse-brief-slot '(baz nil r) nil nil 'class))
)

(defmacro defclass-brief (name superclasses slots &rest class-options)
  "A macro with simpler, less powerful syntax than `defclass' and allows some
more brief ways of expressing things.  The syntax is more like `defstruct.' A
documentation string is optional.  Each slot is expressed as either a bare
symbol, or a list of the name of the slot, its initial value, a symbol with the
letters I, R, A, standing for :initarg, :reader, :writer and :accessor, and a
documentation string.  The symbol, whose package is unimportant, determines the
generation of reader, accessor and so forth; the names are the same as the slot
name.

In addition, three new class options are defined.
  :AUTOMATIC-ACCESSORS means that an accessor is defined for every slot
  :AUTOMATIC-INITARGS means that an initarg is defined for every slot
  (:CONC-NAME <symbol>) prepends `symbol' with a hyphen to each slot accessor.
The default symbol is the class name, in which case the wrapping parentheses are
optional."
  (let ((docstring (if (stringp slots)
		       (prog1 slots (setf slots (pop class-options)))
		       nil)))
    (flet ((delete-option (name)
	     ;; inefficient, but who cares, since it's at compile time and
	     ;; `options' will be short.
	     (let ((elt (or (find name class-options)
			    (find name class-options
				  :key #'(lambda (x) (and (consp x) (first x)))))))
	       (deletef elt class-options)
	       elt)))
      (let ((accessors (or (delete-option :automatic-accessors)
			   *automatic-slot-accessors?*))
	    (initargs  (or (delete-option :automatic-initargs)
			   *automatic-slot-initargs?*))
	    (conc-name (delete-option :conc-name)))
	(when conc-name
	  (setf conc-name
		(if (consp conc-name) (cadr conc-name) name)))
	`(defclass ,name ,superclasses
	   ,(mapcar #'(lambda (s) (parse-brief-slot s accessors initargs conc-name))
		    slots)
	   ,@(when docstring
	       `((:documentation ,docstring)))
	   . ,class-options)))))

#+test
(defclass-brief foo ()
  "the Foo class"
  (a
    (b)
    (c 1)
    (d 2 i)
    (e 3 ia "The E slot")))

#+test
(defclass-brief foo ()
  "the Foo class"
  (a b)
  :automatic-accessors
  :automatic-initargs
  :conc-name)

#+test
(defclass-brief foo ()
  "the Foo class"
  ((a 1 r)
   (c 3 a))
  :conc-name)

#+test
(defclass-brief foo ()
  "the Foo class"
  (a b)
  :automatic-accessors
  :conc-name)

#+test
(defclass-brief foo ()
  "the Foo class"
  (a b)
  :automatic-accessors
  :automatic-initargs
  (:conc-name ugly))

;;; ============================================================================

;;; The Explorer doesn't have this new CL declaration.  Neither does Lispworks
;;; or Lucid.
#+(or Explorer Lispworks Lucid)
(proclaim '(declaration ignoreable))

(defmacro define-after-init (class &body forms)
  "This macro helps ensure that you get the syntax of an :after method on
`shared-initialize' correct.  The arglist is (self &rest initargs)."
  ;; The intern of `self' makes it easy for the user to refer to the SELF
  ;; variable in the user's package rather than to U:SELF
  (let ((self     (form-symbol 'self))
	(initargs (form-symbol 'initargs)))
    `(defmethod initialize-instance :after
       ((,self ,class)  &rest ,initargs)
       #-(or Explorer Lispworks Lucid)
       (declare (ignoreable ,initargs))
       #+(or Explorer Lispworks Lucid)
       ,initargs
       ,@forms)))

#+test
(define-after-init foo
  (print (slot-value self 'b)))

;;; ============================================================================

(defun class-built-on-class (class potential-superclass)
  (member potential-superclass
          #-MCL
	  (if #+Explorer (ticlos::class-composed-p class)
              #-Explorer (class-finalized-p class)
	      (class-precedence-list class)
	      (class-direct-superclasses class))
          #+MCL
          (class-precedence-list class)
	  :test #'eq))

(defmacro map-over-classes (function class-list &key key predicate partial-order &aux (class (gensym)))
  `(dolist (,class ,@(if partial-order `((sort-using-list-order (copy-list ,class-list)
                                                                ,partial-order))
                         `(,class-list)))
     ,@(if key `((setf ,class (funcall ,key ,class))))
     (when (funcall ,predicate ,class)
       (funcall ,function ,class))))

(defmethod class-slot-names ((self standard-object))
  "Returns the slot names for an instance of a class."
  #+MCL
  (mapcar #'slot-definition-name (ccl::class-instance-slots self))
  #-MCL
  (mapcar #'slot-definition-name (class-slots self)))

;;; ============================================================================

(defmacro get-slots (slots object &body forms)
  "Like `with-slots,' except that the slot values are bound to the variables, giving read access only."
  `(let ,(mapcar #'(lambda (slot)
		     `(,slot (slot-value ,object ',slot)))
		 slots)
     . ,forms))

#+test
(get-slots (a b c) foo
  (+ a b c))

;;; ============================================================================

(defmacro set-slots (object &body slots-and-values)
  "Expands into setting `slot-value' with `object' as first arg, slot-name as second."
  `(progn . ,(mapcar #'(lambda (sv)
			 `(setf (slot-value ,object ',(car sv)) ,(cadr sv)))
		     (group slots-and-values 2))))

#+test
(set-slots foo
  a 1
  b 2
  c 3)

;;; ***************************************************************************
;;; EOF
