;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                             Parse-Timestring                           *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Matthew D. Schmill

;;; Department of Computer Science
;;; University of Massachusetts
;;; Amherst MA 01003-4610

;;; Copyright (c) 1995 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;; 06/05/95 Created (schmill)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(parse-timestring))

;;; ---------------------------------------------------------------------------
;;; these are lists of acceptable tokens

;;; make sure to use 'index-of' on these; it allows you to find an index in a deep list

(defparameter +month-list+ '((january february march april may june july august september october november december)
                             (jan feb mar apr may jun jul aug sep oct nov dec)))
(defparameter +day-list+   '(sunday monday tuesday wednesday thursday friday saturday))
(defparameter +fluff-list+ '(at))

;;; ---------------------------------------------------------------------------
;;; these are context-predicates

(proclaim '(inline month-ok-p day-ok-p yearp nth-ok-p))

;; okay to parse an nTH token
(defun nth-ok-p (context) (member context '(:month)))
;; okay to parse a month token
(defun month-ok-p (context) (not (member context '(:month))))
;; okay to parse a day of the week token
(defun day-ok-p (context) (not (member context '(:month :day-of-week))))
;; okay to parse a year
(defun yearp (n) (> n 1900))

;;; ---------------------------------------------------------------------------
;;; a support function

(defun index-of (item list)
  "finds the index of an item in a list, or, in a list embedded in a list.
this allows the user to make 'nicknames' for items in a list."
  (cond
   ((null list) NIL)
   ((consp (car list)) (or (index-of item (car list)) (index-of item (cdr list))))
   (t (let ((k (position item list)))
        (if k (1+ k) NIL)))))

;;; ---------------------------------------------------------------------------
;;; reader functions

(defun token-type (string)
  "Determines which reader to use for the next token."
  (let* ((ns (string-trim " " string))
         (pc (position #\: string))      ;; a colon requires the time reader
         (pb (position #\/ string))      ;; for convenience, we use a date reader for slashes
         (ps (position #\space string))) ;; this is the end of the next token
    (cond
     ((and pc (< pc (or ps (length ns)))) :time-reader)
     ((and pb (< pb (or ps (length ns)))) :date-reader)
     (t :lisp-reader))))

(defun read-time (string)
  "strips the time signature off of the front of string. can handle the following format:
[H]H(:|.)MM[(:|.)[SS]] [AM|PM]"
  (let* ((end (or (position #\space string) (length string)))
         (sbuff (subseq string 0 (min (1+ end) (length string))))
         la hour minute (second 0) (seq '(:h :m :s)))
    (loop until (string= sbuff "") do
          (setq la (min (or (position #\: sbuff) end)
                        (or (position #\. sbuff) end)
                        (length sbuff)))
          (case (pop seq)
	    (:h (setq hour   (read-from-string (subseq sbuff 0 la))))
	    (:m (setq minute (read-from-string (subseq sbuff 0 la))))
	    (:s (setq second (read-from-string (subseq sbuff 0 la))))
            (t  (error "unrecognized time format in ~S" string)))
          (setq sbuff (subseq sbuff (min (1+ la) (length sbuff)) (length sbuff))))
    (unless (and hour minute second)
      (error "incomplete time format in ~S" string))
    (if (search "PM" string)
        (incf hour 12))
    (values (list hour minute second)
            (+ (or (search "PM" string)
                   (search "AM" string)
                   (- end 2)) 2))))

(defun read-date (string &optional default-year)
  "strips the date signature off of the front of string. can handle slash-delimited format:
[M]M/[D]D[/YY]"
  (let* ((end (or (position #\space string) (length string)))
         (sbuff (subseq string 0 (min (1+ end) (length string))))
         la da mo yr (seq '(:m :d :y)))
    (loop until (string= sbuff "") do
          (setq la (or (position #\/ sbuff) (length sbuff)))
          (case (pop seq)
	    (:y (setq yr (read-from-string (subseq sbuff 0 la))))
	    (:d (setq da (read-from-string (subseq sbuff 0 la))))
	    (:m (setq mo (read-from-string (subseq sbuff 0 la))))
            (t  (error "unrecognized time format in ~S" string)))
          (setq sbuff (subseq sbuff (min (1+ la) (length sbuff)) (length sbuff))))
    (when (and yr (< yr 100)) (incf yr 1900))
    (unless (and da mo)
      (error "incomplete date format in ~S" string))
    (values (list mo da (or yr default-year))
            end)))

;;; ---------------------------------------------------------------------------
;;; this is the macro for handling tokens that the lisp reader gets

(defmacro handle-token ((mo da yr ho mi se context cont-q) token)
  (let ((n (gensym "N-")))
    `(progn
       (format t "HT:> ~S [~S/~S]~%" ,token ,context ,cont-q)
       (cond
        ;; a month by name
        ((and (index-of ,token +month-list+)
              (month-ok-p ,context))
         (when (eql ,context :number) (setf ,da ,cont-q) (setf ,cont-q NIL))
         (setf ,mo (index-of ,token +month-list+) ,context :month))
        ;; a day of the week by name
        ((and (index-of ,token +day-list+)
              (day-ok-p ,context))
         (setf ,context :day-of-week))
        ;; a fluff token
        ((index-of ,token +fluff-list+) NIL)
        ;; a symbol is likely to depend on context, but is probably an nth
        ((symbolp ,token)
         (cond
          ((nth-ok-p ,context) (let ((,n (read-from-string (string-trim "DRTHS" (symbol-name ,token)))))
                                 (setf ,da ,n ,context :day)))
          (t (error "unrecognized token ~S" ,token))))
        ;; a number. depends on context
        ((numberp ,token)
	 (cond
          ((yearp ,token) (setf ,yr ,token ,context :year))
          ((and (eql ,context :month) (< ,token 100))
           (setf ,yr (+ ,token 1900) ,context :year))
	  ((null ,cont-q) (setf ,cont-q ,token ,context :number))
          ((not (null ,cont-q)) (error "number token ~S in unexpected location." ,token))))
        (t (error "unparsable token ~S" ,token))))))

;;; ---------------------------------------------------------------------------
;;; this is the grand function for parsing time and date strings

(defun parse-timestring (time-string &key day month year)
  (let ((ts (string-trim ", " time-string))
        (hour 0)
	(minute 0)
	(second 0)
        (context :none)
        (context-queue NIL)
        month day year
        token next-stop)
    ;; set up default date
    (multiple-value-setq (token token token day month year token token token) (get-decoded-time))
    (loop until (string= ts "") do
          (case (token-type ts)
            (:time-reader
             (multiple-value-setq (token next-stop) (read-time ts))
	     (setq hour (first token) minute (second token) second (third token)))
            (:lisp-reader
             (multiple-value-setq (token next-stop) (read-from-string ts))
             (handle-token (month day year ho mi se context context-queue) token))
            (:date-reader
             (multiple-value-setq (token next-stop) (read-date ts year))
             (setq month (first token) day (second token) year (third token)))
            (otherwise (error "unsupported reader call")) ;; shouldn't ever happen
	    )
          (setq ts (string-trim ", " (subseq ts next-stop (length ts)))))
    (values day month year hour minute second)))

#+TEST
(defun test-parse-timestring ()
  (spy (parse-timestring "12/4/1961 4:03:15 PM"))
  (spy (parse-timestring "12/4/1961 4:03 AM"))
  (spy (parse-timestring "12/4/1961"))
  (spy (parse-timestring "8/1/88"))
  (spy (parse-timestring "8/1/88 4:03"))
  (spy (parse-timestring "8/1")))

;;; **************************************************************************
;;; EOF