;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                          MAIL Support Package                          *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Copyright (c) 1990 - 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  09-08-94 File Created.  (Schmill)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

(defpackage ekslmail
  (:nicknames em))

(in-package :ekslmail)
(provide :mail)

(export '(syscall
          initialize-mailbox
          mail-waiting-p
          archive-mailbox
          mailbox-append relocate-mailbox-append relocate-mailbox-overwrite
          mailbox mb-last-edited mb-messages
          mail-message msg-header msg-header-text msg-body msg-markings
          header hdr-from hdr-from-address hdr-reply-to hdr-date
          mark-message unmark-message
          read-mailbox
          sendmail
          dump-mailbox))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; parameter settings block for your architecture


(defparameter mail-path**          "/usr/spool/mail/schmill")
(defparameter archive-extension**  ".acv")
(defparameter severity-threshold** 0)

  ;; so far, the only support is for :standard, :unix, :generic
(defparameter mail-file-type**   :standard)  ;; file system/mailbox type
(defparameter command-style**    :unix)      ;; OS shell command style
(defparameter mail-header-type** :generic)   ;; mail header content

(defparameter system-shell**     (cond ((probe-file "/usr/local/bin/bash") "/usr/local/bin/bash")
                                       ((probe-file "/usr/local/bin/tcsh") "/usr/local/bin/tcsh")
                                       ((probe-file "/usr/local/bin/csh") "/usr/local/bin/csh")
                                       (t "/bin/sh")))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; exports 

(export '(initialize-mail-system
          mail-waiting-p archive-mailbox relocate-mailbox))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; macros

(defmacro uocase ((switch option) &body body)
  "Unsupported Option CASE macro."
  `(case ,switch
     ,@body
     (otherwise (error "Unsupported ~A option ~A." ,option ,switch))))

(defmacro interpreting-error-code (&body body)
  "Signals a warning if a system call returns an error (non zero) code."
  (let ((rc-tag (gensym "rcode-")))
    `(let ((,rc-tag ,@body))
       (when (> ,rc-tag severity-threshold**)
	 (warn "System call ~S returned error code ~S." ',body ,rc-tag)))))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; file operations

(defun syscall (system-command)
  (interpreting-error-code
   #+LISPWORKS 
   (system::call-system system-command system-shell**)
   #-LISPWORKS
   (error "system calls not supported in this environment")))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; file operations

(defun probe-file** (path)
  (uocase (command-style** :command-style)
          (:unix (probe-file path))))

(defun copy-file** (source dest)
  (uocase (command-style** :command-style)
          (:unix (syscall (format nil "cp ~A ~A" source dest)))))

(defun append-file** (source dest)
  (uocase (command-style** :command-style)
          (:unix (syscall (format nil "cat ~A >> ~A" source dest)))))

(defparameter unix-pmasks** '((:user-read . 400) (:user-write . 200) (:user-execute . 100)
                              (:group-read . 40) (:group-write . 20) (:group-execute . 10)
                              (:world-read . 4) (:world-write . 2) (:world-execute . 1)))

(defun protect** (path &optional (level '(:user-read :user-write)))
  (uocase (command-style** :command-style)
          (:unix
	   (let ((mask (loop for i in level sum (or (cdr (assoc i unix-pmasks**)) 0))))
	     (syscall (format nil "chmod ~D ~A" mask path))))))

;;; ---------------------------------------------------------------------------
(defun delete-file** (path)
  (uocase (command-style** :command-style)
    (:unix (delete-file path))))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; mailbox manipulation commands

(defun initialize-mail-system (path file-type command-style &key (shell NIL))
  (setq mail-file-type** file-type)
  (setq mail-path** path)
  (setq command-style** command-style)
  (when shell (setq system-shell** shell)))

(defun mail-waiting-p (&optional (path mail-path**))
  (uocase (mail-file-type** :file-type)
    (:standard (probe-file** path))))

(defun archive-mailbox (&optional (path mail-path**))
  (uocase (mail-file-type** :file-type)
    (:standard (copy-file** path (merge-pathnames path archive-extension**)))))

(defun mailbox-append (&optional (path mail-path**) (append-path NIL))
  (uocase (mail-file-type** :file-type)
    (:standard (append-file** path (or append-path (merge-pathnames path archive-extension**))))))

(defun relocate-mailbox-append (&optional (path mail-path**) (append-path NIL))
  (uocase (mail-file-type** :file-type)
    (:standard
     (append-file** path (or append-path (merge-pathnames path archive-extension**)))
     (delete-file** path))))

(defun relocate-mailbox-overwrite (&optional (path mail-path**) (append-path NIL))
  (uocase (mail-file-type** :file-type)
    (:standard
     (delete-file** (or append-path (merge-pathnames path archive-extension**)))
     (copy-file** path (or append-path (merge-pathnames path archive-extension**)))
     (delete-file** path))))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; message manipulation commands

(defclass mailbox ()
  ((last-edited     :initform NIL :initarg :last-edited :accessor mb-last-edited)
   (messages        :initform NIL :accessor mb-messages))
  (:documentation "An instance of a mailbox image."))

(defclass mail-message ()
  ((header      :initform NIL :accessor msg-header)
   (header-text :initform NIL :accessor msg-header-text)
   (body        :initform NIL :accessor msg-body)
   (markings    :initform NIL :accessor msg-markings))
  (:documentation "A mail message."))

(defclass header ()
  ((from :initform "???" :initarg :from :accessor hdr-from)
   (fadd :initform NIL   :initarg :fadd :accessor hdr-from-address)
   (repto :initform NIL  :initarg :reply-to :accessor hdr-reply-to)
   (date :initform "unknown date" :accessor hdr-date)
   (subject :initform NIL :accessor hdr-subject)))

(defmethod print-object ((obj mail-message) stream)
  (format stream "~A ~{~A~} from ~A: ~A"
          (hdr-date (msg-header obj))
          (mapcar #'(lambda (s) (elt (symbol-name s) 0)) (msg-markings obj))
	  (hdr-from (msg-header obj))
	  (or (hdr-subject (msg-header obj)) "Mail")
          ))

(defmethod mark-message ((msg mail-message) (marking symbol))
  (pushnew marking (msg-markings msg)))

(defmethod unmark-message ((msg mail-message) (marking symbol))
  (setf (msg-markings msg) (delete marking (msg-markings msg))))

(defvar *mrmb* NIL)

(defun parse-header (header-text)
  (let ((new-hdr (make-instance 'header)))
    (loop for cline in header-text do
          (when (> (length cline) 4)
            (cond
             ((string-equal "From:" (subseq cline 0 5))
              (setf (hdr-from new-hdr)
                    (string-trim (format nil " ~C" #\")
                                 (subseq cline 6 (1- (or (position #\< cline)
                                                         (length cline))))))
	      (setf (hdr-from-address new-hdr)
                    (or (ignore-errors (string-trim "<> " (subseq cline (position #\< cline)
                                                                  (position #\> cline))))
                        (hdr-from new-hdr))))
             ((string-equal "Date:" (subseq cline 0 5))
              (setf (hdr-date new-hdr)
                    (string-trim " " (subseq cline 6 (1- (or (position #\- cline)
                                                             (length cline)))))))
             ((string-equal "subject:" (subseq cline 0 8))
              (setf (hdr-subject new-hdr)
                    (string-trim " " (subseq cline 8 (length cline)))))
             ((string-equal "Reply-To:" (subseq cline 0 9))
              (setf (hdr-reply-to new-hdr)
                    (string-trim " " (subseq cline 9 (length cline)))))
             )))
    new-hdr))

(defun read-mailbox (&optional (path mail-path**))
  (let ((mb (make-instance 'mailbox :last-edited (file-write-date path)))
        line (cmsg NIL) (state :body))
    (with-open-file (strm path :direction :input :if-does-not-exist NIL)
      (loop while (not (null (setf line (read-line strm Nil Nil)))) do
            (case state
              (:body
               (if (and (> (length line) 5)
                        (string-equal "From " (subseq line 0 5)))  ;; new message
                   (progn
                     (push (setq cmsg (make-instance 'mail-message)) (mb-messages mb))
	             (push line (msg-header-text cmsg))
		     (setq state :header))
		   (push line (msg-body cmsg))))
	      (:header
	       (if (string-equal "" line)
		   (setq state :body)
		   (push line (msg-header-text cmsg)))))))
    ;; final touching up
    (loop for cmsg in (mb-messages mb) do
          (setf (msg-header-text cmsg) (nreverse (msg-header-text cmsg)))
          (setf (msg-header cmsg) (parse-header (msg-header-text cmsg)))
          (setf (msg-body cmsg) (nreverse (msg-body cmsg))))
    (setq *mrmb* mb)))

(defun spitc ()
  (let ((foo 'ass)
        (pooh (make-hash-table)))
    (with-open-file (strm "/usr/spool/mail/schmill" :direction :input :if-does-not-exist NIL)
      (loop while foo do
	    (setq foo (read-char strm))
	    (unless (alphanumericp foo)
              (unless (gethash foo pooh)
                (progn
                  (setf (gethash foo pooh) 'there)
                  (prin1 foo)
                  (fresh-line))))))))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; message manipulation commands

(defun sendmail (to message-text)
  (let (buff)
    (uocase (command-style** :command-style)
	    (:unix (with-open-file (strm (setq buff (format nil "~D~A" (get-universal-time)
                                                            (symbol-name (gensym "TMP.MAIL."))))
                                         :direction :output :if-does-not-exist :create)
		     (format strm message-text))
             (syscall (format nil "cat ~A | mail ~A" buff
                              (subseq to 3 (length to))))
             (delete-file** buff)))))

;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; message manipulation commands

(defmethod dump-mailbox ((mb mailbox) path &optional (if-exists :supersede))
  (with-open-file (strm path :direction :output :if-does-not-exit :create :if-exists if-exists)
    (loop for m in (mb-messages mb) do
          (when (not (member :delete (msg-markings m)))
            (format strm "~{~A~%~}~%" (msg-header-text m))
            (format strm "~{~A~%~}~%" (msg-body m))))
    (protect** path '(:user-read :user-write))))
