;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: Titanic: /usr/users/eksl/systems/utils/dev/resources.lisp *-*
;;;; *-* Last-edit: Thursday, July 20, 1995  01:25:30; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                           Resource Functions                           *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  05-09-95 File Created.  (anderson)
;;;  06-28-95 Implemented `describe-resource.'  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(allocate-resource 
	   deallocate-resource 
	   with-resource
	   with-resources
	   using-resource
	   using-resources
	   clear-resource
	   describe-resource
	   defresource))

;;; --*--
;;; ***************************************************************************
;;; Resources have a lot of the functionality of `make-allocatable,' but
;;; `make-allocatable' assumes that (1) things are structures or CLOS objects
;;; and (2) that they are all interchangeable, so the allocate function can just
;;; return the first on a stack.  Resources are used for weirder things like
;;; arrays, and they allow a ``matcher'' function that can find the right kind
;;; of thing in the resource.  Thus, resources are slightly more general and
;;; hence more expensive than `make-allocatable.'

(defstruct resource
  name
  arglist
  contents
  initial-allocation
  constructor
  initializer
  matcher
  deallocator)

(defun allocate-resource* (resource args)
  "Either get a matching element from `resource' or construct a new object.  See
`define-resource.'  Users shouldn't call this; it's an internal function."
  (let* ((matcher (resource-matcher resource))
	 (elt     (or (cond ((null matcher)
			     ;; The easy case
			     (and (resource-contents resource)
				  (pop (resource-contents resource))))
			    ((null (resource-contents resource))
			     nil)
			    ((apply matcher (car (resource-contents resource)) args)
			     ;; Slightly harder is matching the first elt
			     (pop (resource-contents resource)))
			    (t
			     ;; In general, have to edit the list
			     (do* ((prev (resource-contents resource) (cdr prev))
				   (this (cdr prev) (cdr prev)))
				  ((endp this) nil)
			       (when (apply matcher (car this) args)
				 ;; remove cons cell with matching car
				 (setf (cdr prev) (cdr this))
				 (return (car this))))))
		      (apply (resource-constructor resource) args))))
    (apply (resource-initializer resource) elt args)
    elt))

#+obsolete
(defun allocate-resource* (resource args)
  "Either get a matching element from `resource' or construct a new object.  See
`define-resource.'  Users shouldn't call this; it's an internal function."
  (let* ((state   0)
	 (matcher (resource-matcher resource))
	 (elt     (or (cond ((null matcher)
			     ;; The easy case
			     (setf state 'null-matcher)
			     (and (resource-contents resource)
				  (pop (resource-contents resource))))
			    ((null (resource-contents resource))
			     (setf state 'contents-null)
			     nil)
			    ((apply matcher (car (resource-contents resource)) args)
			     ;; Slightly harder is matching the first elt
			     (setf state 'pop)
			     (pop (resource-contents resource)))
			    (t
			     (setf state 1)
			     ;; In general, have to edit the list
			     (do* ((prev (resource-contents resource) (cdr prev))
				   (this (cdr prev) (cdr prev)))
				  ((endp this) nil)
			       (when (apply matcher (car this) args)
				 (incf state)
				 ;; remove cons cell with matching car
				 (setf (cdr prev) (cdr this))
				 (return (car this))))))
		      (progn (setf state 'construct) nil)
		      (apply (resource-constructor resource) args))))
    (setf state 'initializing)
    (apply (resource-initializer resource) elt args)
    (setf state 'done)
    elt))

(defun allocate-resource (resource-name &rest args)
  "Either get a matching element from resource named `name' or construct a new
object.  See `define-resource.'"
  (check-type resource-name symbol)
  (let ((resource (get resource-name 'resource)))
    (if resource
	(allocate-resource* resource args)
	(error "Not a defined resource:  ~s" resource-name))))

(defun deallocate-resource* (resource elt)
  "Returns `elt' to resource named `name.' See `define-resource.' Users
shouldn't call this; it's an internal function."
  (funcall (resource-deallocator resource) elt)
  (push elt (resource-contents resource))
  nil)

#+obsolete
(defun deallocate-resource* (resource elt)
  "Returns `elt' to resource named `name.' See `define-resource.' Users
shouldn't call this; it's an internal function."
  (let ((state 0))
    (funcall (resource-deallocator resource) elt)
    (setf state 1)
    (let ((curr (resource-contents resource)))
      (setf state 2)
      (let ((new (cons elt curr)))
	(setf state 3)
	(setf (resource-contents resource) new)))
    (setf state 'done)
    nil))

(defun deallocate-resource (resource-name elt)
  "Returns `elt' to resource named `name.' See `define-resource.'"
  (check-type resource-name symbol)
  (let ((resource (get resource-name 'resource)))
    (if resource
	(deallocate-resource* resource elt)
	(error "Not a defined resource:  ~s" resource-name))))

(defmacro with-resource ((elt resource-name &rest args) &body body)
  "Binds `elt' to an object temporarily allocated resource named
`resource-name,' and deallocates it when `body' is done."
  (check-type elt symbol)
  (check-type resource-name symbol)
  (when (not (get resource-name 'resource))
    (error "Not a defined resource:  ~s" resource-name))
  (with-variables (resource)
    `(let (,elt (,resource (get ',resource-name 'resource)))
       (unwind-protect (progn (setq ,elt (allocate-resource* ,resource (list ,@args)))
			      . ,body)
	 (and ,elt (deallocate-resource* ,resource ,elt))))))

(defmacro with-resources (vars resource-name &body forms)
  "Like `with-resource' but allows multiple instances to be allocated and bound
to variables.  Doesn't allow parameters to the allocation, as `with-resource'
does, but can be convenient sometimes when you need a bunch of objects."
  (check-type resource-name symbol)
  (when (not (get resource-name 'resource))
    (error "Not a defined resource:  ~s" resource-name))
  (with-variables (resource)
    `(let ((,resource (get ',resource-name 'resource)))
       (let ,vars
	 (unwind-protect (progn ,@(mapcar #'(lambda (var)
					      `(setq ,var (allocate-resource* ,resource nil)))
					  vars)
				. ,forms)
	   . ,(mapcar #'(lambda (var)
			  `(and ,var (deallocate-resource* ,resource ,var)))
		      vars))))))

(defmacro using-resource ((elt resource-name &rest args) &body body)
  (declare (ignore elt resource-name args body))
  (error "convert to `with-resource'"))

(defmacro using-resources (vars resource-name &body forms)
  (declare (ignore vars resource-name forms))
  (error "convert to `with-resources'"))

(defun clear-resource (resource-name)
  "Removes all cached elements from `resource.'"
  (check-type resource-name symbol)
  (let ((resource (get resource-name 'resource)))
    (if resource
	(setf (resource-contents resource) nil)
	(error "Not a defined resource:  ~s" resource-name))))

(defun describe-resource (resource-name)
  (check-type resource-name symbol)
  (let ((resource (get resource-name 'resource)))
    (if resource
	(describe resource)
	(error "Not a defined resource:  ~s" resource-name))))

(defmacro defresource
	  (name arglist &key constructor initializer matcher deallocator initial-allocation)
  "Define a resource named `name.' Most of the keyword-specified forms may refer
to `arglist,' the values of which come from the arguments to
`allocate-resource.' You must have a constructor form, the other forms are
optional.  The default initializer is a no-op.  If you supply one, it must be
code that can refer to the symbol `ELT' and the elements of `arglist,' or it may
be any function of the right number of arguments.  By default, any element of
the resource is returned unless there's a matcher, in which case a matching
element is returned.  By default there is no matcher; a matcher is either code,
which may refer to the element by the symbol `ELT' and to the arglist, or it can
be any function.  A deallocator is code that can refer to `ELT' or any function
of one argument; it is called for side-effect, to clean up the elt in some way.
The default deallocator is a no-op.  If `initial-allocation' is defined, it
should be the number of objects to create and put in the resource at load time.
The objects will be created by calling the constructor with no arguments."
  ;; Note that there is a constraint on what symbol can play the role of ELT.
  ;; That symbol must necessarily be U::ELT, while the user will be referring to
  ;; FOO::ELT in the code.  The way to ensure that they are the same is to make
  ;; sure they are both CL:ELT.  In sort, the symbol must be in the CL package.
  ;; I initially was going to use OBJECT as the symbol, as in the Explorer
  ;; resources, but changed to ELT for this reason.  It is kosher to lambda-bind
  ;; a CL symbol.
  
  ;; The preceding comment isn't really true, if we form a symbol in the current
  ;; package.  For backward compatibility, I'll continue to use `elt' as the
  ;; symbol.  Anderson 5/18/95
  (let ((elt (form-symbol "ELT")))
    `(progn
       ;; resources are stored on the 'resource property of the symbol
       (setf (get ',name 'resource)
	     (make-resource
	       :name ',name
	       :arglist ',arglist
	       :initial-allocation ,initial-allocation
	       :contents nil
	       :constructor
	       ,(etypecase constructor
		  ((or symbol function) constructor)
		  (list
		   `(named-lambda ,(form-symbol (string name) "-CONSTRUCTOR-FUNCTION")
				  ,arglist
		      ;; The following avoids warnings about unused args 
		      ,@(lambda-list->args arglist)
		      ,constructor)))
	       :initializer
	       ,(etypecase initializer
		  (null #'no-op)
		  ((or symbol function) initializer)
		  (list
		   `(named-lambda ,(form-symbol (string name) "-INITIALIZER-FUNCTION")
				  (,elt ,@arglist)
		      ;; The following avoids warnings about unused args 
		      ,elt ,@(lambda-list->args arglist)
		      ,initializer)))
	       :matcher
	       ,(etypecase matcher
		  (null nil)
		  ((or symbol function) matcher)
		  (list
		   `(named-lambda ,(form-symbol (string name) "-MATCHER-FUNCTION")
				  (,elt ,@arglist)
		      ;; The following avoids warnings about unused args 
		      ,elt ,@(lambda-list->args arglist)
		      ,matcher)))
	       :deallocator
	       ,(etypecase deallocator
		  (null #'no-op)
		  ((or symbol function) deallocator)
		  (list
		   `(named-lambda ,(form-symbol (string name) "-DEALLOCATOR-FUNCTION")
				  (,elt)
		      ;; The following avoids warnings about unused args 
		      ,elt
		      ,deallocator)))))
       (let ((n ,initial-allocation))
	 (when n
	   (let ((objects nil)
		 (resource (get ',name 'resource)))
	     (dotimes (i n)
	       (push (allocate-resource* resource nil) objects))
	     (dolist (elt objects)
	       (deallocate-resource* resource elt)))))		     
       ',name)))

;;; ============================================================================

#+test
(defresource test1 ()
	 :constructor (make-array 10)
	 :initializer (dotimes (i 10) (setf (aref elt i) i))
	 :deallocator (dotimes (i 5) (setf (aref elt i) nil))
	 :initial-allocation 3)

#+test
(defun test1 ()
  (let ((*print-array* t))
    (print (get 'test1 'resource))
    (dolist (elt (resource-contents (get 'test1 'resource)))
      (print elt))
    (print (allocate-resource 'test1))))

#+test
(defresource test2 (size)
  :constructor (make-array size)
  :matcher (<= size (array-total-size elt)))

#+test2
(defun test2 ()
  (clear-resource 'test2)
  (with-resource (a1 test2 10)
    (let (evil-pointer)
      (with-resource (a2 test2 20)
	(setf evil-pointer a2))
      (with-resource (a3 test2 20)
	(spy (eq evil-pointer a3))
	(with-resource (a4 test2 30)))))
  ;; Should be 3: sizes 10, 20 and 30
  (spy (length (resource-contents (get 'test2 'resource))))
  (spy (mapcar #'length (resource-contents (get 'test2 'resource)))))

;;; ***************************************************************************
;;; EOF
