;;;; -*- Mode: Common-Lisp; Package: EKSL-UTILITIES; Base: 10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/timeit.lisp *-*
;;;; *-* Last-Edit: Saturday, May 6, 1995  12:25:56; Edited-By: anderson *-*
;;;; *-* Machine: Earhart (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*
;;;;
;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                                 TIMEIT                                 *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  05-06-95 File Created.  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

(in-package :EKSL-UTILITIES)

(proclaim '(optimize (speed 3) (safety 1)))

(export '(timeit))

;;; ---------------------------------------------------------------------------
;;; Portable timing macro modelled on the Explorer's `timeit' macro.  It can 
;;; still be improved.
;;; ---------------------------------------------------------------------------

(defmacro timeit (&body body)
  "Executes `body' in a way that allows timing to be measured.  The first form
may be a list of keywords modifying the default behavior, which is to return the
cpu seconds taken by the body.

:values T means return the values of body, NIL is the default
:report T means print the number of cpu seconds
:stream x means report to x instead of *trace-output*; :report is T

Examples:

> (compile (defun foo () (timeit (dotimes (i 100)))))
FOO
> (foo)
102
> (compile (defun foo () (timeit (:report t) (dotimes (i 100)))))
FOO
> (foo)
;;; 102 words consed
102
> (compile (defun foo () (timeit (:values t) (dotimes (i 100)))))
FOO
> (foo)
NIL
> 

This function uses the Common Lisp function `get-internal-run-time' and the
constant `internal-time-units-per-second.'  Due to limited resolution, the
timing of very short functions may be erroneous.

Known bugs:  This macro will return weird results if a GC occurs during the
body.  It seems too risky to turn off GC.  This macro is best used in the body
of a compiled function, as the examples above were."
  (let ((first (if (and (listp (car body))
			(intersection '(:values :report :stream) (car body)))
		   (pop body)
		   nil)))
    (let ((values (getf first :values))
	  (report (or (getf first :report) (getf first :stream)))
	  (stream (getf first :stream *trace-output*))
	  (itime  (gensym))
	  (ttime  (gensym)))
      `(let ((,itime (get-internal-run-time)) ,ttime)
	  (unwind-protect (progn ,@body)
	    (setf ,ttime (/ (- (get-internal-run-time) ,itime)
			   internal-time-units-per-second))
	    ,@(when report `((format ,stream "~&;;; ~S cpu seconds" ,ttime))))
	  ,@(if values nil `(,ttime))))))

#+test
(defun empty-loop (n) (timeit (dotimes (i n))))

#+test
(defun slow-fibo (n)
  "Returns the Nth Fibonacci number using a terribly slow algorithm."
  (declare (fixnum n))
  (cond ((= 0 n) 0)
	((= 1 n) 1)
	(T (+ (the fixnum (slow-fibo (- n 1)))
	      (the fixnum (slow-fibo (- n 2)))))))

#+test
(defun time-fibo (n)
  (timeit (slow-fibo n)))

;;; ***************************************************************************
;;; *                              End of File                                *
;;; ***************************************************************************
