;;;; -*- Mode: Common-Lisp; Package: EKSL-UTILITIES; Base: 10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/clim-colors.lisp *-*
;;;; *-* Last-Edit: Thursday, January 4, 1996  14:58:56; Edited-By: Westy *-*
;;;; *-* Machine: Earhart (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*
;;;;
;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                      CLIM Color Allocation Package                     *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Matthew Schmill
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;; 05-25-95 Created (schmill)
;;; 12-01-95 Changed `defcolor' to not require an index.  (Westy)
;;; 04-10-95 Added pulsing colors [defpulsing-color] (schmill)
;;;
;;; ---------------------------------------------------------------------------

(in-package :EKSL-UTILITIES)

(export '(color-named defcolor defcolor-nickname color-index
	  defpulsing-color defpulsing-group))

;;; ---------------------------------------------------------------------------
;;; CLT stuff

(defvar *default-palette*
    ;; Make sure we are running in a windowing environment
    (ignore-errors
     ;; Calling find-port at load-time is not a great idea. Maybe we need 
     ;; to have a run-time initialization of this variable.
   #+Lispworks (and tools::*display* (port-default-palette (find-port)))
   #-Lispworks (port-default-palette (find-port)))
  "The default palette of the window manager")

(defvar *dynamically-allocated-colors* NIL "Colors that need to be deallocated")
(defvar *colormap* NIL "The actual colormap. Colors are indexed by name.")

(defvar *color-index* -1 "The last color index to used.")
(defvar *max-color-index* 256 "The largest color index we will use.")
(defvar *indexed-colormap* (make-array *max-color-index*) "An array based CLT.")

(defvar *color-nicknames* NIL "A table of nicknames for the named colors.")

;;; ---------------------------------------------------------------------------
;;; color macros

(defun color-named (color-name)
  "Finds a color by name or nickname."
  (when color-name
    (or (cdr (assoc color-name *colormap*))
        (color-named (cdr (assoc color-name *color-nicknames*))))))

(proclaim '(inline color-index))

(defun color-index (color-index)
  "Finds a color by its index in the CLT."
  (aref *indexed-colormap* color-index))

(defmacro defcolor-index (color-name index)
  "Inserts the named color into proper slot of the indexed colormap."
  (once-only (color-name index)
    `(progn
       (setf (aref *indexed-colormap* ,index) (color-named ,color-name))
       (setf (get ,color-name 'color-index) ,index))))

(eval-when (:execute :compile-toplevel :load-toplevel) ; needed to expand macros below
(defun canonicalize-color-name (name)
  (etypecase name
    (symbol (intern (symbol-name name) (find-package 'keyword)))
    (string (intern (string-upcase name) (find-package 'keyword))))))

(defun ok-to-redefine-color-p (color-name x-name-or-color rgb-or-color)
  (let ((previous-args (get color-name 'defcolor-args))
	(current-args (list x-name-or-color rgb-or-color)))
    (unless (equal previous-args current-args)
      (setf (get color-name 'defcolor-args) current-args)
      (unless (null previous-args)
        (warn "Redefining ~a." color-name))
      (values t))))

(defmacro defcolor (color-name x-name-or-color rgb-or-color)
  "Defines a color. The color is acquired by first trying to locate a named color
in the default palette, and next by asking for an actual color allocation in the
host's CLT. Optionally, colors may be acquired directly and supplied in either of
the previously mentioned color args."
  (let ((xc (gensym "XC-")))
    (setf color-name (canonicalize-color-name color-name))
    `(let (,xc)
       (when (ok-to-redefine-color-p ,color-name ,x-name-or-color ,rgb-or-color)
         (if (typep ,x-name-or-color 'color)
             (setq ,xc ,x-name-or-color)
	     (setq ,xc (ignore-errors (find-named-color ,x-name-or-color *default-palette*))))
         (unless ,xc
           (if (typep ,rgb-or-color 'color)
               (setq ,xc ,rgb-or-color)
               (push (setq ,xc (apply #'make-rgb-color ,rgb-or-color)) *dynamically-allocated-colors*)))
         (if (assoc ,color-name *colormap*)
             (setf (rest (assoc ,color-name *colormap*)) ,xc)
             (push (cons ,color-name ,xc) *colormap*))
         (defcolor-index ,color-name (or (get ,color-name 'color-index)
                                         (incf *color-index*)))
         ,(symbol-name color-name)))))

(defmacro defcolor-nickname (nickname realname)
  "Defines a nickname for a named color."
  `(if (assoc ,nickname *color-nicknames*)
       (setf (rest (assoc ,nickname *color-nicknames*))
	 ,realname)
       (push (cons ,nickname ,realname) *color-nicknames*)))

#+Debug
(defun reset-colors ()
  (setf *color-index* -1)
  (loop for (name . color) in *colormap* do
        (remprop name 'color-index)
        (remprop name 'defcolor-args))
  (setf *indexed-colormap* (make-array *max-color-index*)
        *color-nicknames* nil
        *colormap* nil)
  (destroy-allocated-colormap))

;;; ---------------------------------------------------------------------------
;;; color definitions
;;;
;;; You should make definitions to suit your application's needs
;;;

(defcolor :black          "black"           +black+)
(defcolor :white          "white"           +white+)

(defcolor :red            "Medium-Red"      +red+)
(defcolor :blue           "Medium-Blue"     +blue+)
(defcolor :green          "Medium-Green"    +green+)

(defcolor :cyan           "Cyan"            +cyan+)
(defcolor :magenta        "Magenta"         +magenta+)
(defcolor :yellow         "Yellow"          +yellow+)

(defcolor :12%-gray-color ""         '(.125 .125 .125))
(defcolor :25%-gray-color ""         '(.250 .250 .250))
(defcolor :33%-gray-color ""         '(.333 .333 .333))
(defcolor :50%-gray-color ""         '(.500 .500 .500))
(defcolor :66%-gray-color ""         '(.666 .666 .666))
(defcolor :75%-gray-color ""         '(.750 .750 .750))
(defcolor :88%-gray-color ""         '(.875 .875 .875))

(defcolor :dark-green     "Dark-Green"      '(0.0 0.4 0.0))
(defcolor :tan            "Dark-Goldenrod"  '(.72 .525 .041))
(defcolor :medium-green   "Forest-Green"    '(0.133 0.545 0.133))
(defcolor :light-green    "Lime-Green"      '(0.4 1.0 0.3))
(defcolor :dark-grey      "Dark-Slate-Gray" '(0.184 0.31 0.31))
(defcolor :grey           "Slate-Gray"      '(0.44 0.50 0.56))
(defcolor :gold           "gold"            '(1 0.84 0))
(defcolor :grey-green     "Dark-Sea-Green"  '(0.56 0.74 0.56))
(defcolor :light-grey     "Gray"            '(0.75 0.75 0.75))


;; nicknames for custom-references to colors
(defcolor-nickname :my-fave-color   :white)
(defcolor-nickname :special-color   :dark-green)

;;; ---------------------------------------------------------------------------
;;; color initialization & de-initialization

;; CLIM doesn't really deallocate the colors, I don't think.

;; Haven't found an implementation of CLIM that actually does this


(defun destroy-allocated-colormap ()
  "Gives the window manager back any colormap entries we had to forcefully allocate."
  #-LISPWORKS
  (mapc #'clim-utils:deallocate-color *dynamically-allocated-colors*)
  (setf *dynamically-allocated-colors* NIL))

;;; ***************************************************************************
;;; 			          PULSING COLORS
;;; ***************************************************************************

(defvar +pulsating-colors+ (make-hash-table))

(defclass pulsating-color ()
  ((color-object :initform NIL :accessor pc-color)
   (pulse-process :initform NIL :accessor pc-process)
   (pulse-rate   :initform 0.4 :accessor pc-pulse-rate)
   (pulses       :initform 1 :accessor pc-pulses)
   (color-circle :initform NIL :accessor pc-color-circle)))

(defmethod oscillate-dynamic-color (color)
  "Runs in the background, pulsating a pulsing color."
  (flet ((oscillate NIL
	   (loop while t do
		 (loop for i from 0 to (1- (pc-pulses color)) do
		       (setf (dynamic-color-color (pc-color color)) (aref (pc-color-circle color) i))
		       (sleep (pc-pulse-rate color))))))
    (make-thread "Pulsating Color" #'oscillate)))

(defmacro defpulsing-color (name colors &key (pulse-rate 0.1) (stopped NIL))
  "Defines a pulsating color named 'name'. It is inserted into the pallete as a regular color.
   'colors' should be a list of color objects or names.
   if 'stopped' is non-nil, the color will come up with pulsation halted."
  `(let ((color-obj (make-instance 'pulsating-color)))
     (setf (pc-color color-obj) (make-dynamic-color (car ,colors)))
     (setf (pc-color-circle color-obj) (apply #'vector 
					      (loop for col in ,colors append
						    (if (and (symbolp col) (color-named col))
							(list (color-named col))
						      (list col)))))
     (setf (pc-pulses color-obj) (length ,colors))
     (setf (pc-pulse-rate color-obj) ,pulse-rate)
     (defcolor ,name "" (pc-color color-obj))
     (setf (pc-process color-obj) (oscillate-dynamic-color color-obj))
     (if ,stopped (thread-halt "Stopped" (pc-process color-obj)))
     (setf (gethash ,name +pulsating-colors+) color-obj)
     color-obj))

;;; ***************************************************************************
;;; supporting functions

(defmethod set-pulse-rate ((color symbol) pulse-rate)
  "Changes the pulse rate of a pulsating color."
  (setf (pc-pulse-rate (gethash color +pulsating-colors+)) pulse-rate))

(defmethod stop-pulsing-color ((color symbol))
  "Halts a pulsing color from pulsing."
  (thread-halt "Stopped" (pc-process (gethash color +pulsating-colors+))))

(defmethod resume-pulsing-color ((color symbol))
  "Resumes pulsation for a pulsing color."
  (thread-resume-absolute (pc-process (gethash color +pulsating-colors+))))

(defmethod kill-pulsing-color ((color symbol))
  "Smashes a pulsating color. It isn't deallocated, but it kills the pulse process."
  (destroy-thread (pc-process (gethash color +pulsating-colors+)))
  (setf (gethash color +pulsating-colors+) NIL))

(defmethod change-pulsing-color ((color symbol) new-color)
  (setf (pc-color (gethash color +pulsating-colors+))
    (typecase new-color
      (symbol (color-named new-color))
      (number (aref (pc-color-circle (gethash new-color +pulsating-colors+))))
      (t new-color))))

(defun align-pulsing-colors (&rest colors)
  (mapc #'(lambda (cp) (mp::process-reset (pc-process (gethash cp +pulsating-colors+))))
	colors))
  
+TEST
(defpulsing-color :grey-pulse
    '(:12%-gray-color :25%-gray-color :33%-gray-color :50%-gray-color :66%-gray-color 
      :75%-gray-color :88%-gray-color :75%-gray-color :66%-gray-color :50%-gray-color 
      :33%-gray-color :25%-gray-color))

#+TEST
(defpulsing-color :pulse '(:dark-green :grey-green))

;;; ***************************************************************************

(defmethod oscillate-pulsing-group (&rest colors)
  "Runs in the background, pulsating a pulsing color."
  (flet ((oscillate NIL
	   (let ((cm (apply #'lcm (mapcar #'(lambda (dc) (length (pc-color-circle dc))) colors)))
		 (sd (loop for i in colors minimize (pc-pulse-rate i))))
	     (loop while t do
		   (loop for i from 0 to (1- cm) do
			 (loop for col in colors do
			       (setf (dynamic-color-color (pc-color col))
				 (aref (pc-color-circle col) (mod i (length (pc-color-circle col))))))
			 (sleep sd))))))
    (make-thread "Pulsating Group" #'oscillate)))

(defmacro defpulsing-group ((&key (pulse-rate 0.1) (stopped NIL))
			    &rest name-color-pairs)
  "Defines a pulsating color named 'name'. It is inserted into the pallete as a regular color.
   'colors' should be a list of color objects or names.
   if 'stopped' is non-nil, the color will come up with pulsation halted."
  `(labels ((recur-do (ncp)
	      (if (null ncp)
		  NIL
		(let ((color-obj (make-instance 'pulsating-color))
		      (colors (second ncp)))
		  (setf (pc-color color-obj) (make-dynamic-color (car colors)))
		  (setf (pc-color-circle color-obj) (apply #'vector 
							   (loop for col in colors append
								 (if (and (symbolp col) (color-named col))
								     (list (color-named col))
								   (list col)))))
		  (setf (pc-pulses color-obj) (length colors))
		  (setf (pc-pulse-rate color-obj) ,pulse-rate)
		  (eval `(defcolor ,(car ncp) "" ,(pc-color color-obj)))
		  (setf (gethash (car ncp) +pulsating-colors+) color-obj)
		  (cons color-obj (recur-do (cddr ncp)))))))
     (let* ((group (recur-do ',name-color-pairs))
	    (pp (apply #'oscillate-pulsing-group group)))
       (when ,stopped (thread-halt "Stopped" pp))
       (mapc #'(lambda (pc)
		 (setf (pc-process pc) pp)) group))))

(defun kill-all-pulsating-colors ()
  (mapc #'destroy-thread 
	(remove-if-not 
	 #'(lambda (pr) (or (equalp (#+ALLEGRO mp::process-name pr) "Pulsating Color")
			    (equalp (#+ALLEGRO mp::process-name pr) "Pulsating Group")))
	 (all-threads))))

#+Allegro
(defun reset-all-pulsating-colors ()
  (mapc #'mp:process-reset
	(remove-if-not 
	 #'(lambda (pr) (or (equalp (#+ALLEGRO mp::process-name pr) "Pulsating Color")
			    (equalp (#+ALLEGRO mp::process-name pr) "Pulsating Group")))
	 (all-threads))))
