;;;; -*- Mode: Common-Lisp; Package: EKSL-UTILITIES; Base: 10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/clim-utils.lisp *-*
;;;; *-* Last-Edit: Monday, March 4, 1996  14:25:50; Edited-By: Westy *-*
;;;; *-* Machine: Earhart (Alpha) *-*
;;;; *-* Software: Harlequin Common Lisp 3.2 *-*
;;;;
;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                             CLIM Utilities                             *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: David L. Westbrook

;;; Department of Computer Science
;;; University of Massachusetts
;;; Amherst MA 01003-4610

;;; Copyright (c) 1995 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  12-15-95 File Created.  (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(frame-with-process-mixin
	  with-output-to-message-dialog
	  message-dialog
	  with-output-to-y-or-n-dialog
	  y-or-n-dialog
	  with-user-notification))

;;; --*--
;;; ***************************************************************************
;;; Application frame class that can be mixed into frame definitions to provide
;;; a separate process for the frame toplevel loop

(define-application-frame frame-with-process-mixin ()
  ((process :initform nil :initarg :process :accessor frame-process)
   (package :initform *package* :initarg :package :accessor frame-package)))

(defmethod initialize-instance :after ((frame frame-with-process-mixin) &key process process-name process-function no-spawn)
  (store-foreground-streams)
  (unless no-spawn
    (setf (frame-process frame)
	  (or process
	      (make-output-safe-thread
               (or process-name (concatenate 'string (string (frame-name frame)) "-TOP-LEVEL"))
	       (or process-function
		   #'(lambda ()
		       (let ((*package* (frame-package frame)))
		         (run-frame-top-level frame)))))))))

(defmethod frame-exit :before ((frame  frame-with-process-mixin) #+Lispworks &key #+Lispworks destroy)
  (declare (ignore destroy))
  (when (frame-process frame)
    ;; Wanted to do this in a frame-exit :after method, but `frame-exit' does a
    ;; throw so :after code does not work.
    (setf (frame-process frame) nil)
    ))


;;; ---------------------------------------------------------------------------
;;; Message dialogs.

(defmacro with-output-to-message-dialog ((stream &rest keys) &body body)
  "Invokes a dialog in the style of MESSAGE-DIALOG, generating the text
   of the dialog via output to STREAM within the body."
  `(message-dialog (with-output-to-string (,stream)
                     ,@body)
                   ,@keys))

(defun message-dialog (string &key title (ok-text "  OK  ")
                                   (style ':inform) frame)
  "Displays the given message in a dialog box."
  (check-type style (member :inform :warning :error))
  (notify-user frame string
               :style style
               :title (or title
                          (ecase style
                            (:inform  "Message")
                            (:warning "Warning")
                            (:error   "Error")))
               :exit-boxes `((:exit ,ok-text))))


;;; Query dialogs.

(defmacro with-output-to-y-or-n-dialog ((stream &rest keys) &body body)
  "Invokes a dialog in the style of Y-OR-N-DIALOG, generating the text
   of the dialog via output to STREAM within the body."
  `(y-or-n-dialog (with-output-to-string (,stream)
                    ,@body)
                  ,@keys))

(defun y-or-n-dialog (string &key (title "Query") (yes-text "Yes")
                                  (no-text "No") frame)
  "Displays the given message using a dialog box whose two buttons are
   labeled by the YES-TEXT and NO-TEXT respectively.  Clicking the
   former returns T, while clicking the latter returns NIL."
  ;; Under Allegro CLIM, NOTIFY-USER behaves in a non-standard fashion
  ;; when :STYLE :QUESTION is specified.  If the user clicks the :EXIT
  ;; box, NOTIFY-USER returns T; if the :ABORT box, it returns NIL.
  #+Allegro
  (notify-user frame string
               :style ':question
               :title title
               :exit-boxes `((:exit ,yes-text) (:abort ,no-text)))
  #-Allegro
  (restart-case
      (progn
        (notify-user frame string
                     :style ':question
                     :title title
                     :exit-boxes `((:exit ,yes-text) (:abort ,no-text)))
        t)
    (abort ()
      nil)))

;;; ---------------------------------------------------------------------------
;;; `WITH-USER-NOTIFICATION'

;;; Defines a CLIM frame and supporting code to provide a simple pop-up frame
;;; where a message can be displayed for the duration of the body. Similar to
;;; `notify-user', but does not require a user gesture before it disappears.

#+Lispworks
(tools::define-resources
  (lispworks * pop-up-down-frame drop) :yes)

(define-application-frame pop-up-down-frame (frame-with-process-mixin)
  ((message :initarg :message :reader message)
   (draw-text-arguments :initarg :draw-text-arguments :reader draw-text-arguments)
   (finished? :initform nil :accessor finished-p))
  (:panes
   (display  :application :scroll-bars nil))
  (:layouts
   (basic display)))

(defmethod default-frame-top-level ((frame pop-up-down-frame) &key)
  (unless (finished-p frame)
    (enable-frame frame)
    ;; HACK TO FIX TIMING PROBLEM (enable frame signals another process to
    ;; start dealing with events from frame, but this process does not get to it
    ;; soon enough (before we draw out text below))
    (sleep .1)
    (let ((pane (get-frame-pane frame 'display)))
      (apply #'draw-text* pane
	     (message frame)
	     (floor (window-inside-width pane) 2)
	     (floor (window-inside-height pane) 2)
	     :align-x :center
	     :align-y :center
	     :allow-other-keys t
	     (draw-text-arguments frame))
      (force-output pane))
    (thread-wait "Wait for body to finish." #'(lambda ()
					        (finished-p frame)))
    (frame-exit frame)))
    
(defun make-user-notification-window (message &rest arguments &key
                                              (width 300) (height 40) &allow-other-keys)
  (let* ((frame (make-application-frame 'pop-up-down-frame
                  ;; center it on the screen
                  :left (truncate (- (/ (graft-width (find-graft)) 2) (/ width 2)))
                  :top  (truncate (- (/ (graft-height (find-graft)) 2) (/ height 2)))
		  :width width :height height
                  :message message :draw-text-arguments arguments)))
    (values frame)))

(defmacro with-user-notification ((message &rest arguments) &body body)
  (with-variables (frame)
    `(let ((,frame (make-user-notification-window ,message
                                                  ,@arguments
						  :text-size :huge
		                                  :text-face :bold
                                                  :text-family :sans-serif
                                                  )))
       (prog1
           ,@body
         (setf (finished-p ,frame) t)))))

#+test
(defun test-with-user-notification ()
  (with-user-notification ("Don't touch that mouse." :ink +red+)
    (sleep 3)))

;;; ***************************************************************************
;;; EOF