;;;; -*- Mode:Common-Lisp; Package:EKSL-UTILITIES; Fonts:(MEDFNT); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/debugging-utils.lisp *-*
;;;; *-* Last-edit: Wednesday, June 28, 1995  21:40:07; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                       Debugging Utility Functions                      *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: David L. Westbrook and Scott D. Anderson
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; 
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  03-25-91 Moved debugging-tag code out of Phoenix.  (Westy)
;;;  04-14-94 Copied from Transsim into MESS.  Removed the few dependencies on
;;;           Transsim, primarily by renaming *transsim-root* to *application-root*.
;;;           (anderson)
;;;  05-05-95 Added provide-utility wrappers.  (anderson)
;;;  05-07-95 Temporarily moved imports out of the wrapper.  (anderson)
;;;  05-25-95 Changed `debug' so it would be forgiving of quoted symbols.  (anderson)
;;;  05-31-95 Changed `define-debugging-class' to effectively use a separate
;;;           name space (it used to use the class name space).  (Westy)
;;;  06-09-95 More work making the above changes more efficient.  (Westy)
;;;  07-24-95 Moved `eksl-multiple-menu-choose' and `eksl-menu-choose' to
;;;           MENUS file.  (Westy)
;;;  07-25-95 Do some more error checking up front before execution. To do this we
;;;           added `assert-debug-class-defined' functiona and use it in
;;;           `debugging-p' and call `debugging-p' from debug.  (Westy)
;;;  01-03-96 Added `when-debugging-spy'.  (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(define-debugging-class
	  debug
	  undebug
	  when-debugging
	  *debug*
	  when-debugging-format
	  when-debugging-format-if
          when-debugging-spy
 	  debug-format ; local function in `when-debugging'
	  with-debugging
	  *debug-output*
	  current-function
	  edit-debug-status
	  *show-function-in-when-debugging-format?*
	  *show-class-in-when-debugging-format?*
	  ))

;;; --*--
;;; ***************************************************************************
;;; Debugging Classes

;;; This code gives you the ability to define heirarchical classes to control
;;; when debugging code is run. The primary routines in the user interface are
;;; `define-debugging-class', `debug', `undebug' and `when-debugging'.

;;; We define CLOS classes for each debugging-class and use let CLOS's inheritance
;;; mechanism do all the work. See 'define-debugging-class for more details.

;; We also maintain a flat list of all the defined debugging classes in
;; *defined-debugging-classes* and a hierchical list in
;; *debugging-class-classification-alist*, but the latter might be obsolete although
;; it is still usefull for the menu interface.

;; Another interesting part about this is that we convert the name of the debugging
;; class into an obscure name that no one will ever use (effectively creating a new
;; namespace for debugging classes) and we use this obscure name to define the class.

;;; ---------------------------------------------------------------------------
;;; LIST OF ENABLED CLASSES

;(provide-utilities
;  debugging
  
(defparameter *debug* nil
  "This will be non-nil if debugging is enabled. It will contain a list of
debugging classes that are enabled.")
  
;;; ---------------------------------------------------------------------------
;;; FLAT LISTS OF DEFINED CLASSES  
  
(defvar *defined-debugging-classes* ()
  
  "*DEFINED-DEBUGGING-CLASSES*

This parameter contains a list of debugging classes that 
have been defined by define-debugging-class.")

;; This is probably obsolete, but might still be good for a menu interface.
(defvar *debugging-class-classification-alist* nil
  "An alist of direct subclases of debugging classes.")
  
;;;----------------------------------------------------------------------------
;;; MAPPING FROM DEBUG CLASSES TO INTERNAL CLASSES
  
;; Defining mappings between external and internal class names.
;; We do this so that we have our own namespace of debugging classes.
  
(defconstant *internal-class-symbol-property* 'internal-debug-class)
(defconstant *debug-class-symbol-property* 'debug-class)

(defun make-debug-class-name (symbol)
  (form-symbol-in-package (symbol-package symbol) "<" symbol "-INTERNAL-DEBUG-CLASS>"))

(defmacro internal-class->debug-class (symbol)
  `(get ,symbol *debug-class-symbol-property*))

(defmacro debug-class->internal-class (symbol)
  `(get ,symbol *internal-class-symbol-property*))
  
;;; ---------------------------------------------------------------------------
  
(defclass basic-debugging-class ()
  ()
  (:documentation "A basic class upon which all debugging classes are built."))
  
;; This is the basic method for looking up whether debugging has been
;; enabled for a particular debug class.
(defmethod debug-class-enabled ((debug-class standard-class))
  (member (internal-class->debug-class (class-name debug-class)) *debug*))
  
;; This method is used by `debugging-p' which calls `debug-class-enabled'
;; with symbols representing the class.
(defmethod debug-class-enabled ((debug-instance-class-name symbol))
  (assert-debug-class-defined debug-instance-class-name)
  (debug-class-enabled (ensure-class-prototype debug-instance-class-name)))

(defmethod ensure-class-prototype ((class-name symbol))
  (ensure-class-prototype (find-class class-name)))

(defmethod ensure-class-prototype ((class standard-class))
  (unless (class-finalized-p class)
    (finalize-inheritance class))
  (values (class-prototype class)))
  
;;; ---------------------------------------------------------------------------
;;; DEFINE-DEBUGGING-CLASS

;; 'define-debugging-class' defines a class with a debug-class-enabled method. The main
;; `debug-class-enabled' for the class being defined 1) calls
;; `debug-class-enabled' with its class object as the argument. If it is not
;; enabled it 2) calls `debug-class-enabled' defined for each of its superclasses
;; with a class-prototype of the superclass.  Each of the superclasses methods
;; does both of these steps until we find one that is enabled or we bottom out.\
  
(defmacro define-debugging-class
     (class-symbol superclasses &rest class-options
	           &aux internal-debug-class-symbol documentation)
  
  "DEFINE-DEBUGGING-CLASS class-symbol ({superclass}*) ({class-option}*)
This macro defines `class-symbol' as a legal debugging class."
  
  #+Explorer
  (declare (arglist (class-symbol ({superclass}*) [documentation])))
  
  (check-type class-symbol symbol)
  
  ;; Make a hidden symbol that we can use for the class name.
  (setf internal-debug-class-symbol (make-debug-class-name class-symbol))
  
  (when (stringp (first class-options))
    (setf documentation (pop class-options)))
  
  (flet ((delete-option (name)
      ;; inefficient, but who cares, since it's at compile time and
      ;; `class-options' will be short.
     (let ((elt (or (find name class-options)
		    (find name class-options
			  :key #'(lambda (x) (and (consp x) (first x)))))))
       (deletef elt class-options)
       elt)))
    (let ((export (delete-option :EXPORT)))
      (when (consp export)
	(setf export (first (rest export))))
      
    
      `(progn
         ,@(when export
             `((export ',class-symbol)))
         ;; On the symbolic debug class name we store the internal name and...
         (setf (debug-class->internal-class ',class-symbol) ',internal-debug-class-symbol)
         ;; on the internal-name-symbol we store the debug class name.
         (setf (internal-class->debug-class ',internal-debug-class-symbol) ',class-symbol)
         
         (eval-when #+CLTL2 (:compile-toplevel :load-toplevel :execute)
	   #-CLTL2 (compile load eval)
	   (pushnew ',class-symbol *defined-debugging-classes* :test #'eq))
         
         ,@(if superclasses
	     `((progn
		 ;; Define the class with the proper superclasses.
		 (defclass ,internal-debug-class-symbol
		      ,(append (mapcar #'make-debug-class-name superclasses)
			       '(basic-debugging-class))
		   ()
		   ,@(when documentation
		       `((:documentation ,documentation)))
                   ,@class-options)
                 
                 (defmethod debug-class-enabled
		      ((debug-instance-class-name (eql ',class-symbol)))
                   (debug-class-enabled
		    (ensure-class-prototype ',internal-debug-class-symbol)))
                 
		 (defmethod debug-class-enabled
		      ((debug-instance ,internal-debug-class-symbol))
		   (or (debug-class-enabled (class-of debug-instance))
		       ,@(mapcar #'(lambda (class)
				     `(progn
					(assert-debug-class-defined ',class)
                                        (debug-class-enabled
				         (ensure-class-prototype ',(make-debug-class-name class)))))
				 superclasses)))
		 
		 ;; old stuff
		 (dolist (class ',superclasses)
		   (let ((place (assoc class *debugging-class-classification-alist*)))
		     (if place
		       (pushnew ',class-symbol (second place) :test #'eq)
		       (push-acons *debugging-class-classification-alist*
				   class
				   (list (list ',class-symbol))))))))
	     ;; otherwise
	     `((defclass ,internal-debug-class-symbol (basic-debugging-class)
		 ()
		 ,@(when documentation
		     `((:documentation ,documentation))))
               
               (defmethod debug-class-enabled
		    ((debug-instance-class-name (eql ',class-symbol)))
		 (debug-class-enabled
		  (ensure-class-prototype ',internal-debug-class-symbol)))
               
	       (defmethod debug-class-enabled
		    ((debug-instance ,internal-debug-class-symbol))
		 (debug-class-enabled (class-of debug-instance)))))
         
         #+Explorer
         ,@(when documentation
	     `((setf (documentation ',class-symbol 'user::when-debugging-class)
		     ',documentation)))
         
         #+Explorer
         (si:record-source-file-name ',class-symbol 'define-debugging-class)
         
         ',class-symbol))))
  
;;;----------------------------------------------------------------------------
;;; DEBUG
  
;;; The following is necessary because we are about to define a macro called
;;; `debug,' and `debug' is a Common Lisp symbol (a quality for compiled code,
;;; like speed or space).  Lispworks detects this as a redefinition (which is
;;; wrong), and will give us an error.  At least this downgrades it to a
;;; warning.  We have to save the current value, because we want to reset it
;;; afterwards.

(#+allegro excl::without-package-locks
 #+lispworks let #+lispworks ((lw:*handle-warn-on-redefinition* :warn))
 #-(or allegro lispworks) progn   

(defmacro debug (&rest specs)
  "Similar to trace except that it controls execution of `when-debugging' forms.
Specs should be bunch of classes \(unquoted symbols\) which have been used as the
class in a `define-debugging-class' form. Specs can also be T which means enable all
debugging forms or NIL which means disable.
Notes:
The global variable *debug* contains the latest state of debugging."
      
  (if (null specs)
    `(values *debug*)
    `(dolist (class ',specs *debug*)
       (cond ((null class)
	      (setf *debug* nil))
	     ((eq class t)
	      (when (null *debug*)
		(setf *debug* *defined-debugging-classes*)))
	     ((and (listp class)
		   (eq (car class) 'quote)
		   (symbolp (cadr class))
		   ;; This fixes the syntax
		   (setf class (cadr class))
                   (format *error-output* "~&A quote is not needed. Use (DEBUG ~a).~%"
                           class)
		   ;; This makes this COND clause fail, so the next one runs.
		   nil))
	     ((symbolp class)
	      ;; We call `debugging-p' here because it is a nice way to
	      ;; make sure that the entire heirarchy of classes is defined.
              (debugging-p class t)
	      (if (listp *debug*)
		(pushnew class *debug*)
		(setf *debug* (list class))))
	     (t
	      (error "unknown debugging class: ~a" class)))))))
  
;; Need to add some error checking in here.
(defmacro undebug (&rest specs)
  "Similare to untrace except that it undefines debugging classes.
See `debug'."
  (if (null specs)
    `(values (setf *debug* nil))
    `(progn
       (mapcar #'assert-debug-class-defined ',specs)
       (values (setf *debug* (nset-difference *debug* ',specs))))))
  
;;;----------------------------------------------------------------------------

#+OLD
(unless (or (eq class-symbol t)
	    (member class-symbol *defined-debugging-classes* :test #'eq))
  (let (#+Explorer (compiler:warn-on-errors nil)
		   )				; let them get the compiler error
    (cerror "The class will be defined."
	    "Debugging class `~S' was not defined before it was used."
	    class-symbol)))

(defun assert-debug-class-defined (class-symbol)
  (check-type class-symbol symbol)
  (unless (or (eq class-symbol t) ; T as a class in when-debugging means always.
              (member class-symbol *defined-debugging-classes* :test #'eq))
    (cerror "Assume ~s is defined and proceed. To avoid this error add a (define-debugging-class ~:*~s ()) form to your code."
	    "Debugging class `~S' was not defined before it was used."
	    class-symbol)))

#+OLD
(unless (or (eq class-symbol t)
	    (member class-symbol *defined-debugging-classes* :test #'eq))
  (let (#+Explorer (compiler:warn-on-errors nil)) ; let them get the compiler error
    (cerror "The class will be defined."
	    "Debugging class `~S' was not defined before it was used."
	    class-symbol)))
  
(defmacro debugging-p (&optional (class-symbol t) eval)
  "Returns T when the given debugging class is enabled.  If the T is given as
a class returns T if any debugging class is enabled."
  (cond (eval
	 (once-only (class-symbol)
	   `(unless (null *debug*)
	      (or (eq *debug* t)
		  (and (eq ,class-symbol t) *debug*)
		  (and (consp *debug*)
		       (debug-class-enabled ,class-symbol))))))
	(t
	 (restart-case
	     (assert-debug-class-defined class-symbol)
	   (nil () :report "Define the class and proceed."
		(eval `(define-debugging-class ,class-symbol ()))))
	 ;; At the present time these forms will always be inserted into into code and
	 ;; therefore will always test the current debugging state.  If effiency is a
	 ;; big concern we might want to conditionally insert these forms into the
	 ;; code using the check below:
	 ;;  (when (or (eq *debug* t)
	 ;;            (and (listp *debug*)
	 ;;                 (member ',class-symbol *debug*))) ...)
	 `(when
	      (or (eq *debug* t)
		  (and (eq ',class-symbol t) *debug*)
		  (and (consp *debug*)
		       (debug-class-enabled ',class-symbol)))
	    (values t)))))
  
(defmacro when-debugging (class-symbol &body body)
  "Define forms to execute when the specific debug class is enabled.
See `debug' for details. The local macro `debug-format' can be used
in the body of this form."
  `(progn
     (when (debugging-p ,class-symbol)
       (macrolet ((debug-format (format-string &rest format-args)
                    `(debugging-format-helper
                      ',',class-symbol ,format-string . ,format-args)))
         ,@body))))
  
(defmacro debug-format (&rest args)
  "Only to be used in the context of `when-debugging'; Use
`when-debugging-format' instead."
  (declare (ignore args))
  (error "Only to be used in the context of `when-debugging'; Use `when-debugging-format' instead."))
  
;;;----------------------------------------------------------------------------
  
(defvar *debug-output* *debug-io*)
  
(defmacro when-debugging-format (class-symbol format-string &rest format-args)
  "Format output to *debug-output* when the specified debugging class is
enabled."
  `(when-debugging-format-if ,class-symbol T ,format-string . ,format-args))

(defmacro when-debugging-format-if
          (class-symbol conditional format-string &rest format-args)
  "Format output to *debug-output* when the specified debugging class is
enabled and the predicate evaluates to a non-nil value."
  `(when-debugging ,class-symbol
     (when ,conditional
       (debugging-format-helper ',class-symbol ,format-string . ,format-args))))

(defmacro when-debugging-spy (class-symbol &rest spy-args)
  "Execute `spy' when the specified debugging class is enabled."
  `(when-debugging ,class-symbol
     (spy . ,spy-args)))

(defparameter *show-function-in-when-debugging-format?* nil)
(defparameter *show-class-in-when-debugging-format?* nil)
  
  
(defun debugging-format-helper (class-symbol format-string &rest format-args)
  (format *debug-output*
          "~&~:[~*~;~s; ~]~:[~*~;<~a>; ~]~?~%"
          *show-function-in-when-debugging-format?*
          (current-function)
          *show-class-in-when-debugging-format?*
          class-symbol
          format-string format-args))
  
;;; Some of these functions want to report what function they are called from.
;;; With `error' it's trivial to know what function it's in, since you end up in
;;; the debugger, but with warnings and such, it's harder.
  
  (defun current-function ()
    "Returns the name of the function for the current stack frame; useful in error
messages."
    #+Explorer
    (let* ((sg (ticl::process-stack-group w:current-process))
	   (name (eh::function-name
		   (eh::rp-function-word
		     (eh::sg-regular-pdl sg)
		     (eh::sg-next-frame sg (eh::sg-innermost-frame sg))))))
      (values (cond ((and (consp name) (eq (first name) :METHOD))
		     (third name))
		    (t name))))
    #-Explorer
    "Unknown Function")
  
  (defmacro with-debugging ((&rest tags) &body body)
    `(unwind-protect
	 (progn
	   (debug ,@tags)
	   ,@body)
       (undebug ,@tags)))
  
;  )					   ; end of provide-utilities DEBUGGING

;;;----------------------------------------------------------------------------

(defun edit-debug-status (&aux (classes *debug*))
  "This function pops up an editing menu to allow changes to the debug
status.  It might also be executed using the Edit-Debug-Status
command."
    (when (eq *debug* t)
      (setf classes *defined-debugging-classes*))
    (multiple-value-bind (choices done-it)
	(eksl-multiple-menu-choose
	  *defined-debugging-classes*
	  :label "Edit Debugging Status"
	  :original-selections classes
	  :add-item-groups     *debugging-class-classification-alist*
	  :delete-item-groups  *debugging-class-classification-alist*
          :item-print-function #'prin1-to-string)
      (cond (done-it (setf *debug* choices))
	    (t classes))))

;;; ***************************************************************************
;;; EOF











