;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/hooks.lisp *-*
;;;; *-* Last-edit: Tuesday, May 9, 1995  11:33:36; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                             Hook Functions                             *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  04-15-94 File Created.  (anderson)
;;;  05-09-95 Switched to a headed-list representation.  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(define-hook
	   list-hooks
	   run-hook-functions
	   clear-hook
	   set-hook-function
	   insert-hook-function
	   delete-hook-function))

;;; --*--
;;; ***************************************************************************
;;; A hook is essentially just a global variable that contains a headed
;;; association list.  The association is between a key and a hook function and
;;; arguments.  The key is used to add or delete hook functions.  The hook
;;; function is a list of a function and its arguments.

(defvar *all-hook-variables* nil
  "A list of all the hook variables.  The macro `define-hook' adds a variable to
this list; the function `list-hooks' lists them.")

(defmacro define-hook (name &optional (documentation ""))
  "Like defvar, except it doesn't allow an initial value, which is forced to be
NIL.  Adds the variable's name to the list of hooks.  The name of the hook is
exported from the package."
  `(eval-when #+CLTL2 (:compile-toplevel :load-toplevel :execute)
	      #-CLTL2 (compile load eval)
	  (defvar ,name (list ',name) ,documentation)
	  (export '(,name))
	  (setf *all-hook-variables*
		(cons (list ',name ,documentation)
		      (delete ',name *all-hook-variables*
			      :key #'first)))
	  ',name))

(defmacro check-hook (hook)
  "A macro for checking that `hook' is a known hook.  Used internally by the
hook functions."
  (once-only (hook)
    `(unless (find (car ,hook) *all-hook-variables* :key #'first)
       (error "Not a defined hook variable:  ~s" ,hook))))

(defun list-hooks (&optional (stream *standard-output*))
  "Lists all the known hook variables and their documentation, so that you can
find a hook variable that does what you want."
  (format stream "~&~:{~s~%~a~2%~}" *all-hook-variables*))

(defun run-hook-functions (hook)
  "Runs the functions on `hook,' which should be a variable defined with
`define-hook.' Functions can be added to a hook variable by
`insert-hook-function' and `set-hook-function,' and can be removed with
`delete-hook-function.'"
  (check-hook hook)
  (restart-case
    (loop for (key function . args) in (cdr hook) do
	  (if (symbolp function)
	      (setf function (symbol-function function)))
	  (restart-case
	    (apply function args)
	    (skip-hook ()
		       :report (lambda (stream)
				 (format stream "Skip current hook function:  ~s" key)))))
    (skip-remaining-hooks ()
			  :report "Skip remaining hook functions on ~s" (car hook))))

(defun clear-hook (hook)
  "Removes all hooks on `hook,' which must have been defined by `define-hook.'"
  (check-hook hook)
  (setf (cdr hook) nil))

(defun set-hook-function (hook key function &rest arguments)
  "Under `key,' adds `function' and `arguments' to `hook-variable,' which must
have been defined by `define-hook.' Replaces any current value under `key.'

If function is a top-level defun, you may give #'name or 'name.  If function is
not a top-level defun (say, defined by flet or labels), you must use #'."
  (check-hook hook)
  ;; This could be more efficient by going down the list only once,
  ;; but hook lists won't tend to be long and setting them won't
  ;; typically be at run time, and we're doing fairly rapid coding.
  (let ((pair (assoc key (cdr hook))))
    (if pair
	(setf (cdr pair) (cons function arguments))
	(setf (cdr hook)
	      ;; put it at the end of the hook list
	      (append (cdr hook)
		      (list (list* key function arguments)))))))

(defun insert-hook-function (hook key function &rest arguments)
  "Under `key,' adds `function' and `arguments' to `hook,' which must have been
defined by `define-hook.' Any current value under `key' still remains, so both
hooks will run, in the order they were added to the hook.  `set-hook-function'
replaces a value and `delete-hook-function' removes."
  (check-hook hook)
  (setf (cdr hook)
	;; put it at the end of the hook list
	(append (cdr hook)
		(list (list* key function arguments)))))

(defun delete-hook-function (hook key)
  "Delete all values of `key' from `hook,' which must have been defined by
`define-hook.'"
  (check-hook hook)
  (setf (cdr hook) (delete key (cdr hook) :key #'first)))

;;; ============================================================================

#+test
(progn (define-hook *test-hook* "Used for testing only.  Has no effect.")

       (defun print-args (&rest args)
	 (mapc #'print args))

       (defun test-hooks ()
	 (run-hook-functions *test-hook*)
	 (set-hook-function *test-hook* 'foo #'print-args 7)
	 (run-hook-functions *test-hook*)
	 (delete-hook-function *test-hook* 'foo)
	 (run-hook-functions *test-hook*)
	 (insert-hook-function *test-hook* 'foo 'print-args 11)
	 (run-hook-functions *test-hook*)
	 (clear-hook *test-hook*)
	 (run-hook-functions *test-hook*)
	 (list-hooks)))

;;; ***************************************************************************
;;; EOF
