;;;; **************************************************************************
;;;; *		       CLOS Structure writing utilities                       *
;;;; **************************************************************************
;;; Written by: Matthew Schmill

;;; Copyright (c) 1994 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee for non-commercial uses
;;; only (not for resale), provided that the above copyright notice of EKSL,
;;; this paragraph and the one following appear in all copies and in
;;; supporting documentation.

;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity
;;; of such damages.
;;; ---------------------------------------------------------------------------
;;; 07/26/95 MS Created
;;; ---------------------------------------------------------------------------

(unless (find-package :eksl-utilities) (defpackage eksl-utilities))

(in-package :eksl-utilities)
(export '(hardlink-mixin write-structures restore-structures))

;;; ---------------------------------------------------------------------------
;;; the home package

(defparameter *hardlink-package*  *package*)

;;; ---------------------------------------------------------------------------
;;; macro dispatch character for printing restore-forms

(defun restore-form-reader (stream subchar arg)
  "A reader macro that will restore an instance by its description."
  (declare (ignore subchar arg))
  (cons 'apply (list '(function restore-instance) (read stream t NIL t))))

(set-dispatch-macro-character #\# #\? #'restore-form-reader)

;;; ---------------------------------------------------------------------------
;;; hardlink-mixin

(defun generate-hardlink-index ()
  "Generates a (hopefully) unique hardlink index."
  (intern (gensym (format nil "RRI-~D-" (get-universal-time))) *hardlink-package*))

(defclass hardlink-mixin ()
  ((referral :initform (generate-hardlink-index) :initarg :rid :accessor rid))
  (:documentation "A class that is recursiely-recordable with this system."))

(defmethod slots-and-values ((self hardlink-mixin))
  "Returns a list of pairs where each pair is (slot-name slot-value) for 'self'."
  (let ((slots (class-slots (find-class (type-of self)))))
    (mapcar #'(lambda (slot-def) (list (slot-definition-name slot-def)
                                       (slot-value self (slot-definition-name slot-def))))
            slots)))

(defparameter *RLU* (make-hash-table) "A cache for hardlink values.")
(defparameter *cache-referential-objects* NIL)

(defmethod initialize-instance :after ((self hardlink-mixin) &key)
  "Caches the reference if it's supposed to."
  (when *cache-referential-objects*
    (setf (gethash (rid self) *RLU*) self)))

(defun flush-rri-cache () (clrhash *RLU*))

(defmethod reference? ((thing t) &optional (slot-name NIL)) NIL)
(defmethod reference? ((thing symbol) &optional (slot-name NIL))
  (and (not (eql slot-name 'referral))
       (search "RRI" (symbol-name thing) :end2 3)))

;;; ---------------------------------------------------------------------------
;;; Write Instance
;;; these are methods for writing objects in a restorable way

(defparameter *pending-writes* NIL)
(defparameter *previous-writes* NIL)
(defparameter *recursive-writing* NIL)

(defmacro with-recursion-handling (&body body)
  "Lexically notes that any further writes are recursive."
  `(let ((*recursive-writing* t))
     ,@body))

;;; define this as needed
(defgeneric write-instance (instance strm))

;;; bookeeping

(defmethod write-instance :after ((self t) stream)
  "Note that the instance has been read."
  (push self *previous-writes*))

;;; a standard-class wrapper

(defmethod write-instance :around ((self t) stream)
  "Here is where we can catch special cases and do preprocessing."
  ;; if it's not a hardlink-mixin, we may need immediate evaluation
  (format stream "#.")
  ;; if it is a hardlink and we're in a recursive write, just not the RID
  (when (and (typep self 'hardlink-mixin)
             *recursive-writing*)
    (pushnew self *pending-writes*)
    (return-from write-instance (format stream "~A" (rid self))))
  (call-next-method))

;;; basic method method
(defmethod write-instance ((self t) stream)
  (if (typep self 'standard-object)
      ;; trap instances of type standard-class 
      (let ((slot-forms (slots-and-values self)))
        (with-recursion-handling
          (format stream "#?'(~A (~:{(~A ~A)~^ ~}))~%"
                  (type-of self) (mapcar #'(lambda (sf)
                                             (list (first sf) (write-instance (second sf) NIL)))
                                         slot-forms))))
      ;; otherwise just print it out and hope for the best
      (format stream "~S" self)))

;;; hash table
(defmethod write-instance ((self hash-table) stream)
  "Writes a hash-table in restorable form."
  (format stream "#?'(hash-table (:test ~A :size ~D :rehash-size ~F :rehash-threshold ~F)~
                     ~S)~%"
	  (hash-table-test self) (hash-table-size self)
	  (hash-table-rehash-size self) (hash-table-rehash-threshold self)
                    (let (entry-table)
            (maphash #'(lambda (k e) (push (list k e) entry-table)) self)
            entry-table)))

;;; ---------------------------------------------------------------------------
;;; Restore Instance

(defparameter *pending-links* NIL)

(defgeneric restore-instance (type &rest args))

;;; this might not be so intuitive, so... for RRI instances, we will begin by
;;;  trapping them as a symbol with the 't' method. we then make an instance
;;;  of that type, and redispatch the restore-instance method to the RRI method.
(defmethod restore-instance ((type symbol) &rest args)
  (apply #'restore-instance
         (make-instance type :rid
			(when (member (find-class 'hardlink-mixin)
				      (class-precedence-list (find-class type)))
			  (let ((sf (assoc 'referral (first args))))
			    (and sf (cadr sf)))))
	 args))

(defmethod restore-instance ((obj hardlink-mixin) &rest args)
  "Restore a hardlink object."
  (destructuring-bind (slot-pairs) args
    (loop for sp in slot-pairs do
          (when (reference? (second sp) (first sp))
            (push (list obj (first sp) (second sp)) *pending-links*))
          (setf (slot-value obj (first sp)) (second sp)))
    obj))

;;; built-in class specializers

(defmethod restore-instance ((type (eql 'hash-table)) &rest args)
  "Restores a hash-table."
  (destructuring-bind (inits entries) args
    (let ((ht (make-hash-table :test (getf inits :test 'eql)
                               :size (getf inits :size 37)
                               :rehash-size (getf inits :rehash-size)
                               :rehash-threshold (getf inits :rehash-threshold))))
      (loop for ent in entries do
            (setf (gethash (first ent) ht)
                  (second ent)))
      ht)))

;;; ---------------------------------------------------------------------------
;;; Restore (finalize) Referential Links

(defmethod finalize-links ()
  "Connects all the links that were made using the hardlink-mixin."
  (loop for link in *pending-links* do
        (if (gethash (third link) *RLU*)
            (setf (slot-value (first link) (second link)) (gethash (third link) *RLU*))
            (format t "WARNING::NULL REFERENCE TO ~S IN ~A.~%" (third link) (first link)))))

;;; ---------------------------------------------------------------------------
;;; friendly macros

(defmacro with-clean-slate (&body body)
  "Sets up everything all nice and fresh."
  `(let ((*pending-writes* NIL)
         (*previous-writes* NIL)
         (*pending-links* NIL)
         (*RLU* (make-hash-table)))
     ,@body))

(defmacro caching-hardlinks (&body body)
  "Turns hardlink caching on."
  `(let ((*cache-referential-objects* t))
     ,@body))

;;; ---------------------------------------------------------------------------
;;; write a list of structures

(defmethod write-structures (flat-list filename)
  "Writes a list of structures to the named file."
  (with-clean-slate
    (let ((*pending-writes* (copy-list flat-list))
          todo)
      (with-open-file (strm filename :direction :output :if-does-not-exist :create)
        ;; while there's pending writes that haven't been written, write them
        (loop while (setq todo (set-difference *pending-writes* *previous-writes*)) do
              (loop for ds in todo do
                    (write-instance ds strm)))))))

;;; ---------------------------------------------------------------------------
;;; write self-restorable lisp-code

(defmethod structures->code (alist filename)
  "Writes out a set of structures as LisP code."
  (warn "Uh oh, can't touch this."))

;;; ---------------------------------------------------------------------------
;;; restore a structure file

(defmethod read-structures (filename)
  "Restores a list of structures written by write-structures."
  (with-clean-slate
    (caching-hardlinks
      (with-open-file (stream filename :direction :input)
        (let ((buffer NIL))
          (prog1
              (loop while (setq buffer (read stream nil nil)) collecting buffer)
            (finalize-links)))))))
     

;;; ---------------------------------------------------------------------------

(defun test-writer (filename)
  (with-open-file (strm filename :direction :output :if-does-not-exist :create)
    (test-hash-table-writer strm)))

(defclass graph-node (hardlink-mixin)
  ((brother :initarg :brother :accessor brother)
   (sister  :initarg :sister  :accessor sister)))

(defun ron (l) (ignore-errors (elt l (random (1+ (length l))))))

(defparameter *graph* NIL)

(defun fake (&optional (size 5))
  (let ((graph NIL))
    (loop for i from 1 to size do
          (push (make-instance 'graph-node :sister (ron graph) :brother (ron graph)) graph))
    (setq *graph* graph)))
