;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/io.lisp *-*
;;;; *-* Last-edit: Wednesday, June 21, 1995  16:27:53; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                              I/O Utilities                             *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  03-18-94 File Created.  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(->outstream *dev-null* print-hash-table print-array))

;;; --*--
;;; ***************************************************************************

(defparameter *dev-null*
  #-lispm
  (make-two-way-stream (make-concatenated-stream) (make-broadcast-stream))
  ;; Since Lisp Machines have a built-in /dev/null which handles
  ;; additional, non-standard operations, we'll use that instead.
  #+lispm #'system:null-stream
  "A stream like Unix's /dev/null: it absorbs all output and signals eof on any
input attempt.")

(defun ->outstream (stream-arg &rest open-args)
  "Converts \\ar{stream-arg} into something appropriate for output functions.  If it's
already a stream, it's simply returned.  If it's a string or pathname, it is merged with
the user's home directory and opened it for character output, using `open-args' to
determine any other behavior.  The default behavior is :if-exists :new-version and
:if-does-not-exist :create.  If it's the constant T, the *standard-output* stream is
returned.  If it is nil, it is returned; subsequent write operations must check that the
stream is not null."
  (check-type stream-arg (or (eql T) null stream string pathname))
  (cond ((eq stream-arg T) *standard-output*)
	((null stream-arg) nil)
	((typep stream-arg 'stream)
	 ;; Lucid forgot to implement open-stream-p.  SDA 4/14/94
         (if (and #-Lucid
                  (open-stream-p stream-arg)
                  (output-stream-p stream-arg))
             stream-arg
             ;; This may still fail, but what the heck, at least we try.  SDA 7/13/95
             (->outstream (pathname stream-arg))))
	((typep stream-arg '(or pathname string))
	 (setf stream-arg (merge-pathnames stream-arg (user-homedir-pathname)))
	 (apply #'open stream-arg
		:direction :output
		:if-exists :new-version
		:if-does-not-exist :create
		open-args))
	(t (error "Unknown type of stream-arg:  ~s" stream-arg))))

#+test
(defun test->outstream ()
  (spy (ignore-errors (->outstream 5)))
  (with-open-file (s "christa:anderson;foo.lisp" :direction :input)
    (spy (ignore-errors (->outstream s))))
  (spy (->outstream T))
  (spy (->outstream nil))
  (close (spy (->outstream "christa:anderson;foo.lisp")))
  (close (spy (->outstream "foo.lisp")))
  (close (spy (->outstream (open "christa:anderson;foo.lisp" :direction :output)))))

;;; ============================================================================

(defun print-hash-table (table &optional (stream *standard-output*))
  "Print a hash table to a stream, defaulting to *standard-output*"
  (maphash #'(lambda (key value)
	       (format stream "~&~S = ~S" key value))
	   table))

(defun print-array (array &optional long-form? (stream *standard-output*))
  "Prints array on stream, which defaults to standard output.  If long-form? is
true, the elements are printed one to a line, with the index number.  If
long-form? is false, the array is printed by simply binding *print-array* to
true and using `print.'  The long form handles multi-dimensional arrays by
treating them as one-dimensional.  The function does not check that the argument
is an array; the short form works like `print,' and the long form will get an
error."
  (if (not long-form?)
      (let ((*print-array* t))
	(print array stream))
    (let* ((n (array-total-size array))
	   (w (ceiling (log n 10))))
      ;; w is the max width needed to give the element index
      (loop for i from 0 below n do
	    (format stream "~vd:  ~s~%" w i (row-major-aref array i))))))

#+test
(defun test-print-array ()
  (print-array #(1 2 3 4 5 6 7 8 9 10 11))
  (format t "~%")
  (print-array #(1 2 3 4 5 6 7 8 9 10 11) t)
  (print-array #2A((1 2 3) (4 5 6) (7 8 9)))
  (format t "~%")
  (print-array #2A((1 2 3) (4 5 6) (7 8 9)) t))

;;; ***************************************************************************
;;; EOF
