;;;; -*- Mode: Common-Lisp; Package: EKSL-UTILITIES; Base: 10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/macros.lisp *-*
;;;; *-* Last-edit: Sunday, July 2, 1995  20:28:15; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;;
;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                    Eksl Utilities Macro Definitions                    *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: David L. Westbrook

;;; Department of Computer Science
;;; University of Massachusetts
;;; Amherst, Massachusetts 01003.

;;; Copyright (c) 1995 University of Massachusetts
;;; Department of Computer Science
;;; Experimental Knowledge Systems Laboratory
;;; Professor Paul Cohen, Director.
;;; All rights reserved.

;;; Permission to use, copy, modify and distribute this software and its
;;; documentation is hereby granted without fee, provided that the above
;;; copyright notice of EKSL, this paragraph and the one following appear
;;; in all copies and in supporting documentation.
;;; EKSL makes no representation about the suitability of this software for any
;;; purposes.  It is provided "AS IS", without express or implied warranties
;;; including (but not limited to) all implied warranties of merchantability
;;; and fitness for a particular purpose, and notwithstanding any other
;;; provision contained herein.  In no event shall EKSL be liable for any
;;; special, indirect or consequential damages whatsoever resulting from loss
;;; of use, data or profits, whether in an action of contract, negligence or
;;; other tortuous action, arising out of or in connection with the use or
;;; performance of this software, even if EKSL is advised of the possiblity of
;;; such damages.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  05-26-95 File Created.  (Westy)
;;;  06-21-95 Fixed the `push-end' macro to use `get-setf-method.'  (anderson)
;;;  08-24-95 Added some allegro conditionalizations.  (Westy)
;;;  09-29-95 ...and removed them since we now import the symbols.  (Westy)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(once-only
	   nyi
           make-obsolete
	   defsubst
	   named-lambda
	   length-1-lisp-p
	   nordered-adjoin
	   pushnew-ordered
	   push-end
	   pushnew-end
	   replace-or-push
	   replace-or-push-end
	   push-acons
	   pushnew-acons
	   deletef
	   removef
	   circular-list
	   form-symbol-in-package
	   form-symbol
	   form-keyword
	   form-uninterned-symbol
	   gen-name
	   with-keywords-bound
	   docount
	   doplist
	   doalist
	   doarray
	   do-array-maybe-deleting
	   fancy-dolist
	   do-lines
	   do-sequence
	   do-list3
	   group
	   dolist-or-atom
	   gensym0
	   newsym
	   with-variables
	   with-conditional-open-file
	   with-conditional-open-files
	   assert*
	   maxf
	   minf
	   incf-assoc
	   some*))

;;; --*--
;;; ***************************************************************************

(defmacro once-only (variable-list &body body)
  "Generate code that evaluates certain expressions only once.
This is used in macros, for computing expansions.
VARIABLE-LIST is a list of symbols, whose values are subexpressions
to be substituted into a larger expression.  BODY is what uses those
symbols' values and constructs the larger expression.

ONCE-ONLY modifies BODY so that it constructs a different expression,
which when run will evaluate the subsexpressions only once, save the
values in temporary variables, and use those from then on.
Example:
\(DEFMACRO DOUBLE (ARG) `(+ ,ARG ,ARG)) expands into code that computes ARG twice.
\(DEFMACRO DOUBLE (ARG) (ONCE-ONLY (ARG) `(+ ,ARG ,ARG))) will not."
  
  (dolist (variable variable-list)
    (if (not (symbolp variable))
	(error "~S is not a variable" variable)))
  (let ((bind-vars (gensym))
	(bind-vals (gensym))
	(tem (gensym)))
    `(let ((,bind-vars nil)
	   (,bind-vals nil))
       (let ((result ((lambda ,variable-list . ,body)
		      . ,(loop for variable in variable-list
			       collect `(if (let ((variable ,variable))
					      (loop
						(when (atom variable) (return t))
						(when (or (eq (car variable) 'quote) 
							  (eq (car variable) 'function))
						  (return t))
						(if  (eq (car variable) 'the)
						     (setf variable (cadr (cdr variable)))
						     (return nil))))
					    ,variable
					    (let ((,tem (gensym)))
					      (push ,tem ,bind-vars)
					      (push ,variable ,bind-vals)
					      ,tem))))))
	 (if (null ,bind-vars)
	     result
	     `((lambda
		,(nreverse ,bind-vars) ,result) . ,(nreverse ,bind-vals)))))))

(defmacro NYI (&rest args)
  "Signals an error saying that this function is not yet implemented.  The args
are ignored, but by supplying args from the calling function, you can get them
ignored by the compiler."
  `(error "This function is not yet implemented for ~A ~A on ~A."
	  (lisp-implementation-type)
	  (lisp-implementation-version)
	  (machine-type)
          . ,args))

(defmacro make-obsolete (obsolete-definition new-definition)
  `(define-compiler-macro ,obsolete-definition (&whole form &rest args)
     (declare (ignore args))                          
     (fresh-line *error-output*)
     (write-string ,(format nil "; ~a is an obsolete function; use ~a instead.~%"
                            obsolete-definition new-definition)
                   *error-output*)
     (terpri *error-output*)
     (values form)))

;;; ============================================================================

(defmacro defsubst (function-spec lambda-list &body body)
  "Identical to `defun,' except that the function is proclaimed to be
inlineable, but isn't automatically inlined.  That's your choice."
  `(progn (proclaim '(inline ,function-spec))
	  (defun ,function-spec ,lambda-list . ,body)
	  (proclaim '(notinline ,function-spec))))

;;; ============================================================================

(defmacro named-lambda (name arglist &body body)
  "Like `lambda,' except that the lambda has a name, which can be very useful if
the function is executing far from where it was created and if it didn't have a
name you'd have a hard time figuring out where it came from.  Returns the
function object."
  (check-type name symbol)
  `(flet ((,name ,arglist . ,body))
     #',name))

#+test
(defun test-named-lambda ()
  (let ((list-of-functions nil))
    (push (named-lambda created-by-test-named-lambda () (print "hi"))
	  list-of-functions)
    (mapcar #'funcall list-of-functions))
  ;; This is supposed to produce an error, to show that you don't get an error in an
  ;; anonymous lambda, but a named function!
  (mapcar (named-lambda not-enough-args () t) '(1 2 3)))

;;;----------------------------------------------------------------------------

;; Much better than doing (= (length x) 1).
(provide-utilities
  (defsubst length-1-list-p (x) 
    "Is x a list of length 1?"
    (and (consp x) (null (cdr x)))))


;;;----------------------------------------------------------------------------

(provide-utilities
  (defun nordered-adjoin (item list lessp-predicate &key (test #'eql) key)
    "Implements a destructive ordered-insert operation as an alternative to SORT.
Inserts ITEM into the sorted LIST using the LESSP-PREDICATE as in SORT, but only if the
item is not already present, as determined by the TEST as in MEMBER.
KEY is applied to the item and members of LIST when applying LESSP-PREDICATE."
    (declare (optimize (compilation-speed 0) (speed 3) (space 0) (safety 0) (debug 0)))
    (if (null list)
	(list item)
	(do* ((item-value (if key (funcall key item) item))
	      (previous-cons nil tail)
	      (tail list (rest tail))
	      (head (first tail) (first tail)))
	     ((endp tail)
	      (setf (rest previous-cons) (list item))	   ;item appended at end
	      list)
	  (cond ((funcall test item head)  ;item = head?
		 (return list))
		((funcall lessp-predicate item-value (if key (funcall key head) head))	   ;item < head?
		 (cond ((null previous-cons)
			(return (cons item list))) ;item inserted at head of list
		       (t
			(setf (rest previous-cons) (cons item tail))
			(return list)))))))))

(defmacro pushnew-ordered (value place lessp-predicate
			   &rest nordered-adjoin-keywords)
  "Like PUSHNEW, except it calls nordered-adjoin instead of adjoin.
Arguments are passed through to nordered-adjoin."
  (multiple-value-bind (tmpvars tmpvals storevar storeform refform)
      (get-setf-method place)
    `(let* (,@(mapcar #'list tmpvars tmpvals)
	    (,@storevar (nordered-adjoin ,value ,refform ,lessp-predicate
					 ,@nordered-adjoin-keywords)))
       ,storeform)))

;;;----------------------------------------------------------------------------

(defmacro push-end (value place &environment env)
  "Like PUSH, except that the value goes on the end of the PLACE list.  
If PLACE is (), then (value) is returned."
  #+Explorer (declare (ignore env))
  (multiple-value-bind (tmpvars tmpvals storevar storeform refform)
      (get-setf-method place #-Explorer env)
    (let ((oldval  (gensym "OLDVAL"))
	  (newcons (gensym "NEWCONS")))
    `(let* (,@(mapcar #'list tmpvars tmpvals)
	    (,oldval ,refform)
	    (,newcons (cons ,value nil)))
       (if ,oldval
	   (progn (setf (cdr (last ,oldval)) ,newcons)
		  ,oldval)
	   (let ((,@storevar ,newcons))
	     ,storeform)))))
  #+obsolete
  (let ((place-to-setf place))
    (once-only (place value)
      `(if ,place
           (progn
             (setf (cdr (last ,place)) (cons ,value nil))
             ,place)
           (setf ,place-to-setf (cons ,value nil))))))

#+test
(defun test-push-end ()
  (let ((a (make-array 5 :initial-contents '(() (a) (a b c) () (a)))))
    (push-end 'z (aref a 0))
    (print (aref a 0))
    (push-end 'z (aref a 1))
    (print (aref a 1))
    (push-end 'z (aref a 2))
    (print (aref a 2))
    (let ((i 2))
      (push-end 'z (aref a (incf i)))
      (print (aref a i)))
    (let ((i 3))
      (push-end 'z (aref a (incf i)))
      (print (aref a i)))))

(defmacro pushnew-end (value place &rest test-keywords)
  "Like PUSHNEW, except that the value goes on the end of the PLACE list.  
If PLACE is (), then (value) is returned.  The normal test keywords are
allowed: :TEST :TEST-NOT and :KEY, as in PUSHNEW."
  (let ((place-to-setf place))
    (once-only (place value)
      `(if ,place
           (progn
             (or (member ,value ,place ,@test-keywords)		 
                 (setf (cdr (last ,place)) (cons ,value nil)))
             ,place)
           (setf ,place-to-setf (cons ,value nil))))))

#+test
(defun test-pushnew-end ()
  (let ((x (vector '(a b) nil)))
    (pushnew-end "C" (aref x 0) :test #'string-equal)
    (pushnew-end "C" (aref x 0) :test #'string-equal)
    (pushnew-end "D" (aref x 0) :test #'string-equal)
    (spy (aref x 0))))

;;;----------------------------------------------------------------------------

(defmacro REPLACE-OR-PUSH (new-item sequence &key key (test #'eq))
  "NEW-ITEM is used to replace any (all) elements it matches.
If there are no matches then NEW-ITEM is PUSH'ed on the front of SEQUENCE.
SEQUENCE can be a list or an array.
TEST is a function of two args to use to compare NEW-ITEM against an element (or key).
 An element matches when the TEST function returns non-NIL. TEST must be non-nil
 and defaults to 'EQ.
KEY, if non-NIL, is a function to be applied to each element to get a key,
 which is passed to TEST.  If KEY is NIL, the element itself is used.
This macro modifies SEQUENCE (like PUSH)."
  (once-only (new-item)
  `(if (zerop (count ,(if key `(funcall ,key ,new-item) new-item)
                         ,sequence :test ,test ,@(if key `(:key ,key))))
     (push ,new-item ,sequence)
     (setf ,sequence
           ,(if key
              `(nsubstitute-if ,new-item #'(lambda (old-item)
                                             (funcall ,test (funcall ,key ,new-item) old-item))
                               ,sequence :key ,key)
              `(nsubstitute-if ,new-item #'(lambda (old-item)
                                             (funcall ,test ,new-item old-item))
                               ,sequence))))))

;; We should add this COUNT stuff... (Doc line follows)
;; If COUNT is non-NIL, it is the number of such elements to replace.
(defmacro REPLACE-OR-PUSH-END (new-item sequence &key key (test #'eq))
  "NEW-ITEM is used to replace any (all) elements it matches.
If there are no matches then NEW-ITEM is placed on the end of SEQUENCE.
SEQUENCE can be a list or an array.
TEST is a function of two args to use to compare NEW-ITEM against an element (or key).
 An element matches when the TEST function returns non-NIL. TEST must be non-nil
 and defaults to 'EQ.
KEY, if non-NIL, is a function to be applied to each element to get a key,
 which is passed to TEST.  If KEY is NIL, the element itself is used.
This macro modifies SEQUENCE (like PUSH-END)."
  (once-only (new-item)
  `(if (zerop (count ,(if key `(funcall ,key ,new-item) new-item)
                         ,sequence :test ,test ,@(if key `(:key ,key))))
         (push-end ,new-item ,sequence)
         ;; else, replace the little bugger...
         (setf ,sequence
               ,(if key
                  `(nsubstitute-if ,new-item #'(lambda (old-item)
                                                 (funcall ,test (funcall ,key ,new-item) old-item))
                                   ,sequence :key ,key)
                  `(nsubstitute-if ,new-item #'(lambda (old-item)
                                                 (funcall ,test ,new-item old-item))
                                   ,sequence))))))


;;;; --------------------------------------------------------------------------
;;;;   Association List Functions
;;;; --------------------------------------------------------------------------

(define-modify-macro PUSH-ACONS (key datum)
                     (lambda (place key datum)
                       (acons key datum place))

  "PUSH-ACONS place key datum

Pushes an ACONS of key and datum onto the place alist (whether or not 
a matching key exists in the place alist.  Returns the updated alist.")


(define-modify-macro PUSHNEW-ACONS (key datum &REST keys)
                     (lambda (place key datum &REST keys)
                       (let ((assoc-result (apply 'assoc key place keys)))
                         (cond (assoc-result
                                   (setf (rest assoc-result) datum)
                                   place)
                               (t (push-acons place key datum)))))

  "PUSHNEW-ACONS place key datum  &KEYS :TEST :TEST-NOT

Performs an PUSH-ACONS of place, key, and datum only if ASSOC of 
key and place returns NIL.  Otherwise, datum replaces the old
datum of key.  In either case, PUSHNEW-ACONS returns the modified
alist.")

;;; ============================================================================
;;; Can't use define-modify-macro for these because `place' isn't first.  Darn.

(defmacro deletef (item place &rest delete-args)
  "Same as (setf list (apply #'delete item list delete-args))."
  (multiple-value-bind (tmpvar tmpvals storevar storeform refform)
      (get-setf-method place)
    `(let* (,@(mapcar #'list tmpvar tmpvals)
	    (,@storevar (delete ,item ,refform ,@delete-args)))
       ,storeform)))

#+test
(defun test-deletef ()
  (let ((l '(1 2 3 4 5 6 7 8 9 10))
	(i 3))
    (spy (deletef 3 l))
    (spy (deletef 8 (nthcdr (incf i) l)))
    (spy l i)))

(defmacro removef (item place &rest delete-args)
  "Same as (setf list (apply #'remove item list delete-args))."
  (multiple-value-bind (tmpvar tmpvals storevar storeform refform)
      (get-setf-method place)
    `(let* (,@(mapcar #'list tmpvar tmpvals)
	    (,@storevar (remove ,item ,refform ,@delete-args)))
       ,storeform)))

#+test
(defun test-removef ()
  (let ((l '(1 2 3 4 5 6 7 8 9 10))
	(i 3))
    (spy (removef 3 l))
    (spy (removef 8 (nthcdr (incf i) l)))
    (spy l i)))

;;;; --------------------------------------------------------------------------
;;;;   Functions used in the following macros
;;;; --------------------------------------------------------------------------

;;; Sometimes it's nice to have your gensyms mean something when
;;; you're reading the macroexpansion of some form.  The problem
;;; is that if you give a prefix to GENSYM it remains the prefix
;;; until you change it.  

(provide-utilities
  newsym
  (eval-when
    #+CLTL2 (:compile-toplevel :load-toplevel :execute) #-CLTL2 (compile load eval)
    ;; the eval-when is because the newsym function is used in expanding
    ;; `with-variables' and other macros below.
    
    (defvar *newsym-counter* 0
      "Counter used by NEWSYM for generating print names.")
    
    (defun newsym (&optional (prefix "X"))
      "Create a new uninterned symbol whose print name begins with `prefix', which
may be a string or a symbol.  This differs from `gensym' in that the prefix is
not sticky."
      (unless (stringp prefix)
	(setf prefix (string prefix)))
      (make-symbol (format nil "~a~4,'0d" prefix (incf *newsym-counter*))))))

(defmacro with-variables (symbols &body body)
  "Using gensyms is necessary to prevent variables produced by macro expansions
from interfering with user variables, and naming them mnemonically helps make
macro expansions and compiled code easier to read, but it's a pain to create
them properly.  This macro creates them for you, which makes writing nice macros
easier.  For example, if you are writing a macro to iterate over an array, you
used to have to write:

 (defmacro do-2d-array ((elt array) &body body)
   (let ((row (gensym \"ROW\"))
         (col (gensym \"COL\")))
     `(dotimes (,row (array-dimension 0))
        (dotimes (,col ,(array-dimension 1))
           (let ((,elt (aref ,array ,row ,col)))
               . ,body)))))

Now you can just write the following, which eliminates the need to laboriously
create the mnemonic gensyms.

 (defmacro do-2d-array ((elt array) &body body)
   (with-variables (row col)
      `(dotimes (,row ,(array-dimension 0))
          (dotimes (,col ,(array-dimension 1))
             (let ((,elt (aref ,array ,row ,col)))
                 . ,body))))
"
  `(let ,(mapcar #'(lambda (sym)
		     `(,sym (newsym ,(symbol-name sym))))
		 symbols)
     . ,body))

;;;; --------------------------------------------------------------------------
;;;;   Creating symbols
;;;; --------------------------------------------------------------------------

(eval-when (compile load eval)
  ;; used later in a macro that is used in this file
  ;; so must be available at compile time
(defmacro form-symbol-in-package (pkg &rest names)
  "FORM-SYMBOL-IN-PACKAGE pkg &rest names
   Return a symbol interned in PKG whose print name is the concatenation
   of NAMES.  Each name must be acceptable to the string function."
  `(intern (concatenate
	     'simple-string
	     ,@(mapcar #'(lambda (name)
			   (if (stringp name) name `(string ,name)))
		       names))
	   ,pkg)))

(defmacro form-symbol (&rest names)
  "FORM-SYMBOL &rest names
   Return a symbol interned in the current package whose print name is the
   concatenation of NAMES.  Each name must be acceptable to the string function."
  `(form-symbol-in-package *package* ,@names))

(eval-when (compile load eval)
  ;; used later in a macro that is used in this file
  ;; so must be available at compile time
(defmacro form-keyword (&rest names)
  "FORM-KEYWORD &rest names
   Return a symbol interned in the keyword package whose print name is the
   concatenation of NAMES.  Each name must be acceptable to the string function."
  `(form-symbol-in-package "KEYWORD" ,@names)))

(defmacro form-uninterned-symbol (&rest names)
  "FORM-UNINTERNED-SYMBOL &rest names
   Return an uninterned symbol whose print name is the concatenation of NAMES.
   Each name must be acceptable to the string function."
  `(make-symbol (concatenate
		  'simple-string
		  ,@(mapcar #'(lambda (name)
				(if (stringp name) name `(string ,name)))
			    names))))

;;;----------------------------------------------------------------------------
;;; Name generator.  

(defun gen-name (&rest strings)
  "Generate a unique symbol in a package.  `strings' are concatenated and
numbers are appended until it's unique.  If there are any symbols among
`strings,' the symbol-package of that symbol is the package used, otherwise the
current package is used."
  (setf strings (mapcar #'string strings))
  (let* ((base-name (apply #'concatenate 'string strings))
	 (base-package (or (some #'(lambda (x) (and (symbolp x) (symbol-package x))) strings)
			   *package*))
	 (base-symbol (intern base-name base-package)))

    (do ((index (1+ (get base-symbol :symbol-count 0)) (1+ index))
	 symbol
	 indicator)
	(nil)
      (multiple-value-setq (symbol indicator) (intern (format nil "~a~d" base-name index) base-package))
      (when (not indicator)
	(setf (get base-symbol :symbol-count) index)
	(return symbol)))))

;;;----------------------------------------------------------------------------

;;; WITH-KEYWORDS-BOUND:
;;;
;;; (with-keywords-bound (((key1 default1)
;;;                        (key2 default2)
;;;                        ...)
;;;                       arglist)
;;;   <body>)

; eval-when because we use this macro below?
(eval-when #+CLTL2 (:compile-toplevel :load-toplevel :execute) #-CLTL2 (compile load eval)
(defmacro with-keywords-bound ((key-specs arg-list &rest args)
			       &body body)

  "WITH-KEYWORDS-BOUND (key-specs arg-list [error-msg | keyvar] &rest otherwise)
                 &body body)

   Look through ARG-LIST for keywords and bind them, similar to the
   processing of lambda-list keywords.  ARG-LIST is evaluated and should
   be a list of keywords and values.  KEY-SPECS describes the keywords
   to check for.  Each element should be a symbol that will be bound to
   the keyword value from ARG-LIST or a list of the symbol and a default
   value.

   If the third argument is a string it is used as the format string for
   the error message when an unknown keyword is found.

   If the third argument is a symbol then it is used as the variable
   which is used internally to hold the remaining part of the list during
   processing.  OTHERWISE is code that will be executed (in an implicit
   progn) if any keyword from ARG-LIST is not a recognized keyword.
   This code can refer to KEYVAR."

  (let* ((key-symbols (mapcar #'(lambda (key)
                                  (cond ((symbolp key) key)
                                        ((symbolp (first key)) (first key))
                                        ((symbolp (first (first key))) (first (first key)))
                                        (t (error "ill-formed key spec; ~a" key)))
                                  #+OLD
				  (if (symbolp key) key (first key))
                                  )
			      key-specs))
	 (error-msg (if (stringp (first args)) (first args) nil))
	 (keyvar (if (and (symbolp (first args))
			  (not (null (first args))))
		     (first args)
		     (newsym "KEY")))
	 (otherwise (if error-msg nil (cdr args)))
         (key-spec-bindings
           (mapcar #'(lambda (key-spec)
                       (cond ((symbolp key-spec) key-spec)
                             ((symbolp (first key-spec)) `(,(first key-spec) ,(second key-spec)))
                             ((symbolp (first (first key-spec))) `(,(first (first key-spec))
                                                               ,(second key-spec)))
                             (t (error "ill-formed key spec; ~a" key-spec)))
                       #+OLD
                       (if (consp key-spec)
                         `(,(first key-spec) ,(second key-spec))
                         key-spec)
                       )
                   key-specs))
         (supplied-p-bindings
           (mapcan #'(lambda (key-spec)
                       (when (and (consp key-spec) (third key-spec))
                         `((,(third key-spec) nil))))
                   key-specs))
         (all-bindings (append key-spec-bindings supplied-p-bindings))
         )
    (or otherwise
	error-msg
	(setf error-msg (format nil "~~s is not one of (~{:~a~^ ~})."
				key-symbols)))
    `(let ,all-bindings
       (when (consp ,arg-list)
	 (do ((,keyvar ,arg-list (cdr ,keyvar)))
	     ((null ,keyvar))
	   (case (first ,keyvar)
	     ,@(mapcar #'(lambda (key-spec)
                           (multiple-value-bind 
                             (variable key)
                             (cond ((symbolp key-spec) 
                                    (values key-spec 
                                            (form-keyword key-spec)))
                                   ((symbolp (first key-spec)) 
                                    (values (first key-spec)
                                            (form-keyword (first key-spec))))
                                   ((every #'keywordp (rest (first key-spec)))
                                    (values (first (first key-spec))
                                            (mapcar #'(lambda (k) (form-keyword k))
                                                    (rest (first key-spec))))))
                             #+OLD
                             (if (symbolp key-spec) key-spec (first key-spec))
                             (let ((supplied-p-symbol (when (consp key-spec) (third key-spec))))
                               `(,key
                                 (setf ,variable (first (setf ,keyvar (cdr ,keyvar))))
                                 ,(when supplied-p-symbol
                                    `(setf ,supplied-p-symbol t))))))
                       key-specs)
             (otherwise
              ,@(or otherwise
                    `((error ,error-msg (first ,keyvar))))))))
       ,@body))))  

;;;----------------------------------------------------------------------------

(defmacro docount ((count &optional (result nil resultp)) &body body)
  
  "DOCOUNT (countform [resultform]) {declaration}* {tag | statement}*

This macro provides a simple iteration over a sequence of integers. First the
countform is evaluated to produce an integer.  The body is then executed
countform times.  Finally, the resultform is evaluated and returned (defaulting
to NIL).  The entire construct is surrounded by an implicit block
named NIL, so that RETURN may be used to exit the loop at any point. 
 
The internal counter can be declared using the THE special form in
countform."
  
  (let ((var-sym (gensym)))
    `(dotimes (,var-sym ,count ,@(when resultp `(,result)))
       ,@(when (and (consp count)
                    (eq (first count) 'the))
          `((declare (,(second count) ,var-sym))))
       ,@body)))

(defmacro doplist ((key value plist) &body body)
  "Iterate over each key/value pair in a plist.  Key and Value are
bound to successive pairs in Plist."
  (let ((atvar (gensym "ATVAR")))
    `(do* ((,atvar ,plist (cddr ,atvar))
	   (,key (car ,atvar) (car ,atvar))
	   (,value (second ,atvar) (second ,atvar)))
	  ((null ,atvar) nil)
       ,@body)))

#+test
(defun test-doplist ()
  (doplist (k v '(a 1 b 2 c 3 d 4))
    (format t "~&key = ~s, value = ~d~%" k v)))

(defmacro doalist ((key value alist) &body body)
	  "Macro to iterate over an a-list:  Elements in the alist are bound
to key and value"
  (let ((atvar (gensym "ATVAR")))
    `(dolist (,atvar ,alist)
       (let ((,key (car ,atvar))
	     (,value (cdr ,atvar)))
	     ,@body))))

#+test
(defun test-doalist ()
  (doalist (k v (pairlis '(a b c d e) '(1 2 3 4 5)))
    (format t "~&key = ~s, value = ~d~%" k v)))

(defmacro doarray ((var array
			 &optional index 
			 &key start end index-list)
		    &body body)
  "Iterate over each element in array.  Optionally bind index to the current
index.  Iterate over elements specified in index-list or from `start,' which
defaults to zero, to `end,' which defaults to the size of the array."
  (let ((counter (or index (newsym "INDEX-"))))
    (with-variables (stop-index)
      (cond (index-list
	     `(dolist (,counter ,index-list)
		(let ((,var (aref ,array ,counter)))
		  ,@body)))
	    (t
	     `(do ((,counter (or ,start 0) (1+ ,counter))
		   (,stop-index (or ,end (length ,array))))
		  ((<= ,stop-index ,counter))
		(let ((,var (aref ,array ,counter)))
		  ,@body)))))))

#+test
(defun test-doarray ()
  (doarray (elt #(1 2 3 4))
    (format t "elt = ~s~%" elt))
  (doarray (elt #(1 2 3 4) i)
    (format t "elt = ~s i = ~s~%" elt i))
  (doarray (elt #(1 2 3 4) i :start 1)
    (format t "elt = ~s i = ~s~%" elt i))
  (doarray (elt #(1 2 3 4) i :end 2)
    (format t "elt = ~s i = ~s~%" elt i))
  (doarray (elt #(1 2 3 4) i :start 1 :end 2)
    (format t "elt = ~s i = ~s~%" elt i))
  (doarray (elt #(1 2 3 4) i :index-list '(0 2))
    (format t "elt = ~s i = ~s~%" elt i)))

(defmacro do-array-maybe-deleting ((elt array) &body body)
  "Iterate over each element in `array.' If `body' returns nil, delete the
element by replacing it with the last element of the array; consequently, this
can rearrange the elements of the array.  The array must have a fill-pointer."
  (with-variables (last index)
    `(do ((,last (1- (fill-pointer ,array)))
	  (,index 0)
	  ,elt)
	 ((< ,last ,index))
       (setf ,elt (aref ,array ,index))
       (if (progn ,@body)
	   (incf ,index)
	   (setf (aref ,array ,index) (aref ,array ,last)
		 ,last (1- ,last)
		 (fill-pointer ,array) (1- (fill-pointer ,array)))))))

#+test
(defun test-do-array-maybe-deleting ()
  (let ((a (make-array 10 :fill-pointer 0)))
    (dotimes (i 10) (vector-push i a))
    ;; All ten are printed out
    (do-array-maybe-deleting (e a)
      (print e)
      (evenp e))
    ;; Only the evens remain, but in a weird order
    (print-array a)))

(defmacro fancy-dolist ((var list &optional (index (gensym "INDEX")) &rest options) &body body)
  "Just like `dolist,' but allows `index,' `start' and `end' arguments.  The
`index' argument is a variable that is bound to the index of the current list
element, as with `doarray,' so that you can simultaneously step through an
array, for example.  As with all sequence arguments, `start' defaults to 0, the
beginning of the sequence, and `end' defaults to nil, meaning the end of the
sequence.  The subsequence includes `start' but excludes `end,' so the length of
the subsequence is `end' - `start.'"
  #+Explorer
  (declare (arglist (var list &optional index &key start end) &body body))
  (when (keywordp index)
    (format t "~&Index = ~s, which is not a valid variable name.~%~
I will assume you are using the old syntax that didn't have index variables by inserting a gensym for this argument.~%" index)
    (push index options)
    (setf index (newsym "INDEX-")))
  (let ((sublist (newsym "SUBLIST-")))
    (once-only (list)
      (with-keywords-bound ((start end) options)
	`(do* ((,index   ,(or start 0)                           (1+ ,index))
	       (,sublist ,(if start `(nthcdr ,index ,list) list) (cdr ,sublist))
	       (,var     (car ,sublist)                          (car ,sublist)))
	      (,(if end
		    `(or (>= ,index ,end)
			 (null ,sublist))
		    `(endp ,sublist)))
	   ,@body)))))

#+test
(defun test-fancy-dolist ()
  (fancy-dolist (elt '(1 2 3))
    (print elt))
  (format t "~%")
  (fancy-dolist (elt '(1 2 3) i :start 1)
    (print elt))
  (format t "~%")
  (fancy-dolist (elt '(1 2 3) i :end 2)
    (print elt))
  (format t "~%")
  (fancy-dolist (elt '(1 2 3) i :start 1 :end 2)
    (print elt))
  (format t "~%")
  (fancy-dolist (elt '(1 2 3) i)
    (print i))
  ;; test syntax-checking for old-style syntax
  (format t "~%")
  (fancy-dolist (elt '(1 2 3) :start 0)
    (print elt)))

#+test
(defun test-fancy-dolist ()
  ;; test to see if we can delete the current elt.
  (let ((l '(0 1 2 3 4 5 6 7 8)))
    (fancy-dolist (x l)
      (if (evenp x) (setf l (delete x l))))
    (spy l)))
    
(defmacro do-lines ((line multi-line-string &optional (delimiters '(#\newline))) &body forms)
  "Like `dolist,' except that `line' is a symbol bound to each line of
`multi-line-string.' In other words, the macro takes care of parsing
`multi-line-string' into lines, delimited by `delimiters,' which default to just
the newline character.  Comparison is done using `char=,' so case is
significant.  If you want both #\A and #\a to be delimiters, specify both.

The only consing that this does is for each line of the string, and so the total
is roughly equal to the storage of the string.  As a special case, if the string
is only one line, the macro does no consing."
  (with-variables (s e)
    (let ((parse-code
	    (case (length delimiters)
	      (0 (error "Must have at least one delimiter in order to parse."))
	      (1 `(position ,(car delimiters)
			    ,multi-line-string
			    :test #'char=
			    :start ,s))
	      (t `(position-if #'(lambda (char)
				   (member char ,delimiters))
			       ,multi-line-string
			       :start ,s)))))
      `(loop for ,s = 0 then (1+ ,e)
	     for ,e = ,parse-code
	     for ,line = (if (and (zerop ,s) (null ,e))
			     ,multi-line-string
			     (subseq ,multi-line-string ,s ,e))
	     do ,@forms
	     until (null ,e)))))

#+test
(defun test-do-lines ()
  (format t "~2%Test the common case, parsing by newlines:~%")
  (do-lines (l (format nil "first~%second~%third"))
    (spy l))
  (format t "~2%Test the case of parsing by multiple characters:~%")
  (do-lines (l (format nil "first~%secondathirdaafourth") '(#\newline #\a))
    (spy l))
  (format t "~2%Test that the only consing is due to the `subseq' function:~%")
  (let ((multi-line-string (concatenate 'string
					(make-string 10 :initial-element #\s)
					(string #\newline)
					(make-string 10 :initial-element #\t)
					(string #\newline)
					(make-string 10 :initial-element #\u))))
    (time:timeit (:cons)
      (do-lines (l multi-line-string)
	(setf * l)))
    (time:timeit (:cons)
      (setf * (subseq multi-line-string 0 10))
      (setf * (subseq multi-line-string 11 21))
      (setf * (subseq multi-line-string 22 nil))))
  (format t "~2%Test that if the string is only a single line, no consing is done.~%")
  (let ((multi-line-string "aaaaaaaaaaaaa"))
    (time:timeit (:cons)
      (do-lines (l multi-line-string)
	(setf * l)))
    (time:timeit (:cons)
      (setf * (subseq multi-line-string 0 nil)))))

(defmacro do-sequence ((elt sequence &rest args) &body body)
  "Just like `doarray' and `fancy-dolist,' but works for general sequences.  As
with all sequence arguments, `start' defaults to 0, the beginning of the
sequence, and `end' defaults to nil, meaning the end of the sequence.  The
subsequence includes `start' but excludes `end,' so the length of the
subsequence is `end' - `start.' Be warned that if this macro is used in a
context where the sequence is either always an array or always a list, you pay a
space penalty to no purpose."
  #+Explorer
  (declare (arglist (elt sequence &optional index &key start end) &body body))
  `(etypecase ,sequence
     (array (doarray      (,elt ,sequence ,@args) ,@body))
     (list  (fancy-dolist (,elt ,sequence ,@args) ,@body))))

(defmacro do-list3 ((var1 var2 var3 list &key (circlep nil)) &body body)
  "Iterates over `list' like `dolist,' except that `var2' is the previous value
of `var1' and `var3' is the previous value of `var2.' Another difference is that
we begin with `var1' eq to the third element, `var2' eq to the second element
\(the ``previous'' value of `var1'\), and `var3' eq to the first element \(the
``previous'' value of `var2'\).  If `circlep' is true, the list is treated as if
it were circular, after `var1' hits the end of the list, it starts over as the
first element and then the second element, whereupon the looping stops.  A block
named NIL is created, so you can `return' from `body.'"
  ;; When list-ptr becomes nil, we start again at the front of the list, but we
  ;; only do the first two elements, so we count.
  (with-variables (list-ptr count)
    (once-only (list)
      `(when (>= (length ,list) 3)
	 ;; I'd prefer to implement the macro with `labels,' but the TI compiler
	 ;; doesn't do a good enough job with the tail recursion.  The
	 ;; tail-recursive form is:
	 #+ignore
	 (block nil
	   (labels ((lp (,list-ptr ,var1 ,var2 ,var3 ,count)
		      ,@body
		      (if (null ,list-ptr)
			  (if ,circlep
			      (case ,count
				(0 (lp nil (first  ,list) ,var1 ,var2 (1+ ,count)))
				(1 (lp nil (second ,list) ,var1 ,var2 (1+ ,count)))
				(2)))
			  (lp (cdr ,list-ptr) (car ,list-ptr) ,var1 ,var2 0))))
	     (lp (cdddr ,list) (third ,list) (second ,list) (first ,list) 0)))
	 ;; Instead of that elegant solution, we have the following ugliness.  Essentially, this is the compiled
	 ;; form of the preceding labels.
	 (block nil
	   (let ((,list-ptr (cdddr ,list))
		 (,var1 (third ,list))
		 (,var2 (second ,list))
		 (,var3 (first ,list))
		 (,count 0))
	     (tagbody LP 
		 ,@body
		 (setf ,var3 ,var2
		       ,var2 ,var1)
		 (if (null ,list-ptr)
		     (if ,circlep
			 (progn (case ,count
				  (0 (setf ,var1 (first ,list)))
				  (1 (setf ,var1 (second ,list)))
				  (2 (return)))
				(incf ,count))
			 (return))
		     (setf ,var1 (pop ,list-ptr)))
		 (go lp))))))))

#+test
(defun test-do-list3 ()
  (format t "~2&")
  (do-list3 (a b c '(1 2 3 4 5 6 7 8))
    (format t "~s ~s ~s~%" a b c))
  (format t "~2&")
  (do-list3 (a b c '(1 2 3 4 5 6 7 8) :circlep t)
    (format t "~s ~s ~s~%" a b c))
  (format t "~2&")
  (do-list3 (a b c '(a b c))
    (format t "~s ~s ~s~%" a b c))
  (format t "~2&")
  (do-list3 (a b c '(a b c) :circlep t)
    (format t "~s ~s ~s~%" a b c)))


(defmacro dolist-or-atom ((var possible-list &optional test-form) &body body)

  "DOLIST-OR-ATOM (var possible-list &optional test-form) &body body

   This macro handles those cases where a variable may be a single object or
   a list of objects.  If `Possible-List' is a list then DOLIST-OR-ATOM will
   act like DOLIST; if `Possible-List' is not a list then this will act like
   LET.  The optional `Test-Form' is a form that should return true if you
   want the DOLIST behavior.  It defaults to the form (LISTP Possible-List).

   Note that `Possible-List' may be evaluated twice so it should not be a
   form with side-effects or one that is expensive to compute."

  ;; This could be coded as follows but if body is large the
  ;; space penalty is significant.
  ;; 
  ;; (if ,test-form
  ;;     (dolist (,var ,possible-list) ,@body)
  ;;     (let ((,var ,possible-list)) ,@body))

  (let ((test-result (gensym))
	(remaining (gensym)))
    (when (null test-form)
      (setf test-form `(listp ,possible-list)))
    `(do* ((,test-result ,test-form)
	   (,remaining ,possible-list (cdr ,remaining))
	   (,var (if ,test-result (first ,remaining) ,remaining)
	         (first ,remaining)))
	  ((when ,test-result
	     (endp ,remaining)))
       ,@body
       (unless ,test-result (return nil)))))

;;;----------------------------------------------------------------------------

(defmacro with-conditional-open-file ((stream filename &rest options)
                                         &body body)
  
  "A slightly modified version of the Common Lisp `with-open-file.'

Executes `body' with the `stream' variable bound to a stream for the
corresponding file `filename'.  `Filename' is opened using `options', which are
the same as for the OPEN function.  If `filename' is nil, the OPEN is bypassed
and nil is bound to `stream'.  If `filename' is a stream, the stream variable is
bound to it.  If `filename' is T, the stream variable is bound to *standard-output*."
  #+Explorer
  (declare (arglist stream filename [{option}*] {form}*))
  (once-only (filename)
    (let ((close? (newsym 'close?)))
      `(let* ((,close? nil)
	      (,stream (cond
                        ((eq ,filename T)
                         *standard-output*)
			;; Nil simply return nil ::
			((null ,filename) nil)
			;; Streams return themselves ::
			((streamp ,filename) ,filename)
			;; Anything else attempts an open ::
			(t (setf ,close? t)
			   (cl:open ,filename ,@options)))))
	 (unwind-protect
	     (progn ,@body)
	   (when (and ,close? ,stream)
	     (cl:close ,stream)))))))

(defmacro with-conditional-open-files (streams &body body)
  
  "A handy version of WITH-CONDITIONAL-OPEN-FILE for nested opens.

Executes `body' with the all `stream' variables bound to a stream for the
corresponding file `filename'.  `Filename' is opened using `options', which
are the same as for the OPEN function.  If `filename' is nil, the OPEN is
bypassed and nil is bound to `stream'."
  #+Explorer
  (declare (arglist {(stream filename [{option}*])}* {form}*))
  (cond ((endp streams) `(progn ,@body))
        (t `(with-conditional-open-file ,(first streams)
              (with-conditional-open-files ,(rest streams) ,@body)))))

;;;----------------------------------------------------------------------------

(defmacro assert* (test-form &optional place error-message report-message &rest arguments)
  "A variation on the `assert' macro in which the restart handler is bound
during the execution of the test-form, so that uses of this macro can be nested.
As with `assert,' a setf-able `place' may be given, but unlike `assert,' only
one may be given.  As with `assert,' you can give an error message, but you can
also give a message to go in the restart handler.  That report message may also
use the arguments.  It will usually say something like ``supply a new value for
foo.'' The restart is named `continue,' and takes one argument, the value to
setf `place' to.  Like `assert,' NIL is returned if test-form is true."
  (let ((arg-name    (newsym "ARG-"))
	(stream-name (newsym "STREAM-")))
    `(loop (restart-case (return (unless ,test-form
				   (error ,error-message . ,arguments)))
	     (continue
	       (,arg-name)
	       :report
	       #+ignore
	       (lambda (,stream-name)
		 (format ,stream-name ,report-message . ,arguments))
	       ;; The following is stupid looking, but seems to be necessary,
	       ;; since the preceding doesn't work on the Explorer
	       (lambda (,stream-name)
		 (write-string (format nil ,report-message . ,arguments)
			       ,stream-name))
	       :interactive (lambda ()
			      (format *query-io* "Enter a form to eval:  ")
			      (list (eval (read))))
	       (setf ,place ,arg-name))))))

#+test
(defun test-assert* ()
  (let ((list-of-numbers '(a b)))
    (assert* (and (listp list-of-numbers)
		  (do ((rest list-of-numbers (cdr rest)))
		      ((null rest) t)
		    (assert* (numberp (car rest))
			     (car rest)
			     "Must be a number:  ~s"
			     "Supply a new element."
			     (car rest))))
	     list-of-numbers
	     "Must be a list of numbers:  ~s"
	     "Supply a whole new list"
	     list-of-numbers)
    list-of-numbers))



;;;----------------------------------------------------------------------------

(define-modify-macro maxf (value) max "Sets `place' to max of its old value and `value'")
(define-modify-macro minf (value) min "Sets `place' to min of its old value and `value'")

#+test
(defun test-maxf ()
  (let ((x 2)
	(a #(1 2 3 4 5 6 7 8)))
    (maxf x 4)
    (spy x)
    (maxf (aref a (incf x)) 99)
    (print-array a)
    (spy (aref a x))))

;;;----------------------------------------------------------------------------

#+BUGGY
(defmacro incf-assoc (place key &optional (delta 1))
  (let ((assoc-result (newsym "ASSOC-RESULT")))
    `(let* ((,assoc-result (assoc ,key ,place)))
       (if ,assoc-result
         (setf (rest ,assoc-result) (incf (rest ,assoc-result) ,delta))
         (push-acons ,place ,key ,delta)))))

(defmacro incf-assoc (place key &optional (delta 1))
  "Constructs an association list in `place' whose value is a count of the
number of occurences of the `key'.  Returns the association list."
  (once-only (key delta)
    (with-variables (alist pair)
      (multiple-value-bind (tmpvars tmpvals storevars storeform refform)
	  (get-setf-method place)
	`(let* (,@(mapcar #'list tmpvars tmpvals)
		(,alist  ,refform)
		(,pair   (assoc ,key ,alist)))
	   (if ,pair
	       (progn
		 (incf (rest ,pair) ,delta)
		 ,alist)
	       (let ((,@storevars  (acons ,key ,delta ,alist)))
		 ,storeform)))))))

#+(and test Lispworks)
(defsetf nthcdr (n list) (value)
  `(setf (cdr (nthcdr (1- ,n) ,list)) ,value))

#+test
(defun test-incf-assoc ()
  (let ((al (pairlis '(a b c d e f g) '(1 2 3 4 5 6 7))))
    #+Lispworks (setf al (nreverse al))
    (spy al)
    (incf-assoc al 'c)
    (spy al)
    (let ((i 3))
      (incf-assoc (nthcdr (incf i) al) 'a)
      (spy al))))

;;;----------------------------------------------------------------------------

(defmacro some* ((element list) &body body)
  "An iterative version of the function `some'.  Bind each element of list in
turn.  If the body returns non-nil, return the result.  This can be faster than
the regular function, especially when lexical variables are used in the
predicate."
  (let ((result (newsym 'result)))
    (once-only (list)
      `(when ,list
         (let (,result)
           (dolist (,element ,list)
             (when (setf ,result (progn ,@body))
               (return ,result))))))))


;;; ***************************************************************************
;;; *                              End of File                                *
;;; ***************************************************************************