;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: Titanic: /usr/users/eksl/systems/utils/provide-require.lisp *-*
;;;; *-* Last-edit: Sunday, May 7, 1995  20:14:48; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *        Macros to support Providing and Requiring of Utilities          *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  05-05-95 File Created.  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(provide-utilities require-utilities))

;;; --*--
;;; ***************************************************************************
;;; The idea of our provide/require facility is that (1) during development, the
;;; user has available all of the EKSL utilities, so that they are readily
;;; available for any purpose, but (2) at delivery time, only those functions
;;; and macros that are truly required will be delivered, thereby keeping the
;;; footprint smaller.  For example, the `spy' macro should never been
;;; delivered, since it is only for debugging, but the user shouldn't have to
;;; ask for it during development.  Our approach is to put each utility
;;; definition inside `provide-utilities.' A user names the utilities they use
;;; via `require-utilities' in their own system and that will cause the utility
;;; to be put in the delivered binary file.  Note that most utility macros don't
;;; need to be delivered because they only affect compilation.

(defvar *known-utilities* (make-hash-table)
  "A mapping from symbols to definition forms, so that the binary file can act
like those forms were defined at the site of the `require' and not at the site
of the `provide.'")

(defmacro provide-utilities (&body forms)
  "Defines `forms' but also squirrels away the definition forms, so that they
can be used at the site of `require-utilities,' depending on the value of
*deliver-utilities*.  As an added convenience, this exports the forms."
  ;; The main thing here is extracting the name being defined.
  ;; Almost every CL definition form has the name in the cadr,
  ;; so it should be easy.
  (if (symbolp (car forms))
      ;; All forms are stored as a group.
      `(progn (setf (gethash ',(car forms) *known-utilities*) 
		    '(progn . ,(cdr forms)))
	      . ,(mapcan #'(lambda (form)
			     (case (car form)
			       ((defun defmacro)
				`(,form (export ',(cadr form))))
			       (t (list form))))
			 (cdr forms)))
      ;; Each form stored individually.
      `(progn . ,(mapcan #'(lambda (form)
			 (case (car form)
			   ((defun defmacro defsubst)
			    `((setf (gethash ',(cadr form) *known-utilities*)
				    ',form)
			      ,form
			      (export ',(cadr form))))
			   (t (error "Don't know how to provide ~s" form))))
		     forms))))

#+test
(provide-utilities
  (defun foo (x y) (+ x y)))

#+test
(provide-utilities
  debugging
  (defvar *tmp* nil)
  (export '*tmp*)
  (defun tmp (tag) (pushnew tag *tmp*)))

(defvar *deliver-utilities* nil
  "When false, `require-utilities' is a no-op.  When true, `require-utilities'
acts like the definition forms for those utilities, using the definition forms
squirreled away by `provide-utilities.'")

(defmacro require-utilities (&body names)
  "When *deliver-utilities* is true, this macro acts as if `name' were defined here, rather
than in the EKSL-UTILITIES file.  Note that most utility macros don't need to be
delivered, because they only affect the compilation of the system and don't exist in the
binary files."
  (when *deliver-utilities*
    `(progn (in-package :eksl-utilities)
	    ,@(mapcar #'(lambda (name)
			  (or (gethash name *known-utilities*)
			      (error "No known definition for ~s" name)))
		      names)
	    (export ',names :eksl-utilities))))

#+test
(require-utilities foo)

#+test
(require-utilities debugging)

;;; ============================================================================
;;; Remaining problems:

;;; How do we get other forms to come along for the ride?  Example is `newsym'
;;; and `*newsym-counter.' The latter will never be explicitly asked for, and
;;; the former is needed by `with-variables.' (Actually, it would be easy to
;;; eliminate `newsym' and friends, but the problem would crop up with something
;;; else, such as define-debugging-class and *debug*.) In short, we have more
;;; complex dependencies than we thought originally.  How do we handle a
;;; `defsetf' form?  What if the getter is required, shouldn't the setter and
;;; `defsetf' come along too?  I've made a stab at a solution already by letting
;;; the car of the `provide-utilities' be a name for the set of forms.  I don't
;;; know if it will cover everything.

;;; Delivery is not worked out yet.  We want to do a compilation pass for
;;; delivery, but there's a contradiction: it has to load U in order to (a) have
;;; the macros available and (b) provide all the definition forms so that
;;; `require-utilities' works, but it has to *not* load U so that we can see the
;;; warnings about any undefined functions, the ones that might have been
;;; omitted by `require-utilities.' Possibly the compilation for delivery
;;; involves two passes.  The first loads U and compiles the `require-utilites,'
;;; the second (in a fresh lisp) doesn't load U and instead loads the
;;; `require-utilities.' Since one stands for the other, that seems right, but
;;; it would be nice to make it less complicated.  How do we get the
;;; non-required macros loaded for the compilation?  Maybe every macro is
;;; implicitly loaded into the compiler environment by the `require-utilities?'

;;; I ran into the following problem, which seems hard to solve.  I had the
;;; following form, because if you're going to require the debugging stuff, you
;;; want all of it:

#+comment
(provide-utilities
  debugging
  #+Lispworks
  (import '(clos::class-prototype))

  (defmethod debug-class-enabled ((debug-instance-class-name symbol))
    (debug-class-enabled (class-prototype (find-class debug-instance-class-name))))

  ;; lots of other functions
  )

;;; Okay, the problem is that when `provide-utilities' reads the entire form,
;;; the symbol `class-prototype' gets read and interned in the U package.  Then,
;;; when the `import' gets executed, there's a package conflict.  Seems like the
;;; `import' needs to moved out, before the `provide-utilities,' but then it
;;; doesn't get copied over to the user's system at delivery time.

;;; ***************************************************************************
;;; EOF
