;;;; -*- Mode:Common-Lisp; Package:eksl-utilities; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/threads.lisp *-*
;;;; *-* Last-edit: Friday, May 26, 1995  15:23:49; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                            Thread Functions                            *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Scott D. Anderson
;;;             Experimental Knowledge Systems Laboratory
;;;             Paul R. Cohen, Principal Investigator
;;;             David L. Westbrook, Systems Manager
;;;             David M. Hart, Laboratory Manager
;;;             Department of Computer Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  04-08-95 File Created.  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :EKSL-UTILITIES)

(export '(make-thread destroy-thread current-thread all-threads thread-wait
		      thread-wait-with-timeout thread-yield thread-interrupt
		      without-interrupts
		      store-foreground-streams with-foreground-io make-output-safe-thread))

;;; --*--
;;; ***************************************************************************

(defun make-thread (name function)
  "Creates a new thread named `name' that executes `function,' which can start
executing immediately.  `name' can be pretty much anything, but strings or
symbols are more informative."
  #FEATURE-CASE
  ((Explorer (ticl:process-run-function name function))
   (Clim     (clim-sys:make-process function :name name))
   (MCL      (ccl:process-run-function name function))))

(defun destroy-thread (thread)
  "Terminates the thread, which is an object as returned by `make-thread.'"
  #FEATURE-CASE
  ((Explorer (ticl:send thread :kill))
   (Clim     (clim-sys:destroy-process thread))
   (MCL      (ccl:process-kill thread))))

(defun current-thread ()
  "Returns the currently running thread, which will be the same kind of object
returned by `make-thread.'"
  #FEATURE-CASE
  ((Explorer ticl:current-process)
   (Clim     (clim-sys:current-process))
   (MCL      ccl:*current-process*)))

(defun all-threads ()
  "Returns a sequence of all the threads."
  #FEATURE-CASE
  ((Explorer sys:all-processes)
   (Clim     (clim-sys:all-processes))
   (MCL      ccl:*all-processes*)))

(defun thread-wait (reason-string wait-predicate)
  "This thread is suspended until `wait-predicate' returns true.
`Reason-string' is primarily for debugging/observation purposes."
  ;; Call it once so that the trivial errors happen in the user's process
  ;; (funcall wait-predicate)
  #FEATURE-CASE
  ((Explorer (ticl:process-wait reason-string wait-predicate))
   (Clim     (clim-sys:process-wait reason-string wait-predicate))
   (MCL      (ccl:process-wait reason-string wait-predicate))))

(defun thread-halt (reason-string thread)
  "Stops a thread in it's tracks by giving it an arrest reason."
  #FEATURE-CASE
  ((allegro (mp:process-add-arrest-reason thread reason-string))))

(defun thread-resume (reason-string thread)
  "Possibly resumes a thread by taking away an arrest reason."
  #FEATURE-CASE
  ((allegro (mp:process-revoke-arrest-reason thread reason-string))))

(defun thread-resume-absolute (thread)
  "Resumes a thread by taking away all arrest reasons."
  #FEATURE-CASE
  ((allegro (loop for reason-string in (mp:process-arrest-reasons thread) do
		  (mp:process-revoke-arrest-reason thread reason-string)))))

(defun thread-wait-with-timeout (reason-string seconds wait-predicate)
  "This thread is suspended until `wait-predicate' returns true or `seconds'
pass.  `Reason-string' is primarily for debugging/observation purposes."
  #FEATURE-CASE
  ((Explorer (ticl:process-wait-with-timeout reason-string (* 60 seconds) wait-predicate))
   (Clim     (clim-sys:process-wait-with-timeout reason-string seconds wait-predicate))
   (MCL      (ccl:process-wait-with-timeout reason-string (* 60 seconds) wait-predicate))))

(defun thread-yield ()
  "Allows other processes to run.  As soon as others have had a chance, this
thread continues."
  #FEATURE-CASE
  ((Explorer (ticl:process-allow-schedule))
   (Clim     (clim-sys:process-yield))
   (MCL      (ccl:process-allow-schedule))))

(defun thread-interrupt (thread function)
  "Interrupts `thread,' throwing out of its computation, and forces it to call
`function.'"
  #FEATURE-CASE
  ((Explorer (ticl:process-preset thread function))
   (Clim     (clim-sys:process-interrupt thread function))
   (MCL      (ccl:process-preset thread function))))

(defmacro without-interrupts (&body forms)
  "Executes `forms' as a critical section; no other threads can get in."
  ;; Because this macro will appear in user code (eg, the priority queue
  ;; code), whether there is multi-threading or not, we have to expand to
  ;; something, so we expand to progn.  The :eksl-generic-load-utils feature
  ;; is used as an otherwise case, since it will be defined as long as this
  ;; code is being loaded via our loading procedure.
  #FEATURE-CASE
  ((Explorer    `(ticl:without-interrupts . ,forms))
   (Clim        `(clim-sys:without-scheduling . ,forms))
   (MCL         `(ccl::without-interrupts . ,forms))
   (eksl-generic-load-utils `(progn . ,forms))))

;;; ============================================================================

(defmacro change-thread (place me you)
  "For two threads that want to `co-routine.' The `place' will be set to either
`me' or `you,' which are typically symbols but can be any marker.  When one
thread calls this, it sets the `place' to be `you,' indicating that it is now
the other thread's turn to run, and waits for `place' to be eq to `me.' Consider
the following example:
   
 (defun test-change-thread ()
   (let ((turn 'a)
         (stream *terminal-io*))
     (let ((a (make-thread \"a\" #'(lambda ()
                                     (dotimes (i 10)
                                       (print 'aaaa stream)
                                       (change-thread turn 'a 'b)))))
           (b (make-thread \"b\" #'(lambda ()
                                     (thread-wait 'first-turn #'(lambda () (eq turn 'b)))
                                     (dotimes (i 10)
                                       (print 'bbbb stream)
                                       (change-thread turn 'b 'a))))))
      (thread-yield)
      (sleep 5)
      (destroy-thread a)
      (destroy-thread b))))

`place' has to be a simple, setfable place, such as a variable or a slot in a
structure or an object.  The `place' form should have no side effects or do any
pointer chasing.  In other words, the macro doesn't use `get-setf-method.'"
  `(progn (setf ,place ,you)
	  (thread-wait ,(format nil "Wait ~s eq ~s" place me)
		       #'(lambda () (eq ,place ,me)))))

#+test
(defun test-change-thread ()
  (let ((turn 'a)
        (stream #FEATURE-CASE ((Explorer *terminal-io*)
                               (Lispworks *standard-output*))))
    (let ((a (make-thread "a" #'(lambda ()
                                 (dotimes (i 10)
                                   (print 'aaaa stream)
                                   (change-thread turn 'a 'b)))))
          (b (make-thread "b" #'(lambda ()
                                 (thread-wait 'first-turn #'(lambda () (eq turn 'b)))
                                 (dotimes (i 10)
                                   (print 'bbbb stream)
                                   (change-thread turn 'b 'a))))))
      (thread-yield)
      (sleep 5)
      (destroy-thread a)
      (destroy-thread b))))

;;; ============================================================================
;;; Experiments in Thread I/O

(defvar *foreground-streams* :unbound)

(defun store-foreground-streams ()
  "A function that stores streams so that they can be used by
`with-foreground-io.'"
  (setf *foreground-streams*
        (list *terminal-io* *debug-io* *error-output* *query-io*
	      *standard-output* *standard-input* *trace-output*)))

(defmacro with-foreground-io (&body body)
  "Binds the CL I/O streams so that I/O goes through the foreground window.  It
does not bind *terminal-io*, since CLtL2 says explicitly that ``no program
should ever change the value of *terminal-io*.  This macro binds the other I/O
streams: *debug-io* *error-output* *query-io* *standard-output* *standard-input*
and *trace-output*.  Note that this will only work if the function
`store-foreground streams' has already been run in the foreground, that is, it
was run before starting up a thread."
  #FEATURE-CASE
  (((or Lispworks Allegro Lucid MCL) 
    `(destructuring-bind
	 (*debug-io* *error-output* *query-io* *standard-output*
		     *standard-input* *trace-output*)
	 (cdr *foreground-streams*)
       . ,body))
   (Explorer `(let ((tio (car *foreground-streams*)))
		;; tio is *terminal-io*
		(let ((*debug-io* tio)
		      (*error-output* tio)
		      (*query-io* tio)
		      (*standard-output* tio)
		      (*standard-input* tio)
		      (*trace-output* tio))
		  . ,body)))))

(defun make-output-safe-thread (name function)
  "Similar to make thread, but uses `with-foreground-io' to redirect standard
output streams to reasonable places. See `with-foreground-io' for details."
  #+DEBUG
  (print-output-streams "~a: foreground:" name)  
  (let ((thread (make-thread
		 name
		 #'(lambda ()
		     #+DEBUG
		     (print-output-streams "~a: before with-foreground-io" name)
		     (with-foreground-io
		      #+DEBUG
		      (print-output-streams "~a: after with-foreground-io" name)
		      (funcall function))))))
    (values thread)))

#+test
(defun print-output-streams (&optional format-string &rest format-args)
  (dolist (s '(*debug-io* *error-output* *query-io* *standard-output* *standard-input*
			  *trace-output*))
    (let ((stream (eval s)))
      (when (output-stream-p stream)
        (fresh-line stream)
        (when format-string
          (apply #'format stream format-string format-args))
        (write-char #\space stream)
        (princ s stream)
	(write-string " is here" stream)
        (terpri stream)))))

#+test
(defun test-io ()
  (print-output-streams "foreground")
  (make-thread "io test" #'(lambda () (print-output-streams "background")))
  (sleep 1)
  (store-foreground-streams)
  (make-thread "io test" #'(lambda () (with-foreground-io (print-output-streams "fore/background"))))
  (values))

;;; ***************************************************************************
;;; EOF
